// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


/**This class is responsible to run a Servlet in the ServletCapsule. This class has variety of functionality
*and is intended to run on a seperate thread. This class is more algorithm oriented. The main features and
*functions of this class is listed below:<br/><br/>
*
*1) Responsible to retrieve a specific ServletCapsule from the ServletStore.<br/>
*2) If the ServletCapsule is not found - a proper ServletCapsule instance is created and stored in the 
*ServletStore.<br/>
*3) Responsible to procure all the related information that is to be passed to the Servlet inside the ServletCapsule
*in the form of ServletConfig, ServletRequest, ServletResponse etc.<br/>
*4) If the ServletCapsule is newly instantiated, this shall require a call to <b>init()</b>.<br/>
*5) If the ServletCapsule already exists - this class calls the <b>service()</b> method on the encapsulated servlet.<br/>
*6) All the Servlet deployed on Crusader - shall have to have a visible default constructor, which this.
*class shall use to instantiate the Servlet.<br/>
*7) This class is also responsible to handle the Exceptions the Servlet generates.<br/><br/>
*
*What this class does not do is set the HTTP response code. This is left to the main server implementation.<br/>
*
*Author - Sanjay"Ironluca"Debnath <a href="mailto:Ironluca@yahoo.com">Ironluca@yahoo.com</a><br/>
*Last Modified - 21/04/2001
*/

package jason.service.crusader;

import jason.service.requestparser.datahandler.*;
import jason.service.crusader.javaximpl.*;

import java.util.*;
import java.io.*;
import javax.servlet.SingleThreadModel;
import javax.servlet.*;

import jason.framework.protocol.*;
import jason.framework.protocol.http.*;

public class CrusaderCore implements Runnable {

    /**Stores the reference to the ServletStore from where ServletCapsules are to be retrieved - there exists
       only one ServletStore globally.*/
    protected static ServletStoreInt servletStore;

    /*Contains the mapping between the ServletContext objects and the context they belong to - there is one
      "DefaultContext".
      protected static ServletContextStoreInt servletContextStore;
    */

    /**Conatins the ServletConfig inplementation - there exists as many ServletConfig implementation as there
       are servlets.*/
    protected ServletConfigImpl servletConfig;

    /**Stores the {@link jason.service.crusader.ServletCapsule ServletCapsule} embedded in this Core.
       @see jason.service.crusader.ServletCapsuleInt*/
    protected ServletCapsuleInt servletCapsule;

    /**The ServletRequest implementation. See {@link jason.service.crusader.javaximpl.ServletRequestImpl ServletRequestImpl}
       & {@link jason.service.crusader.javaximpl.HttpServletRequestImpl HttpServletRequestImpl} for more information.*/
    protected HttpServletRequestImpl servletRequest;

    /**The ServletResponse implementation. See {@link jason.service.crusader.javaximpl.ServletResponseImpl ServletResponseImpl}
       & {@link jason.service.crusader.javaximpl.HttpServletResponseImpl HttpServletResponseImpl} for more information.*/
    protected HttpServletResponseImpl servletResponse;


    /**The class name has to be the fully qualified class-name of the Servlet class, the object of which shall be 
       encapsulated in the {@link #servletCapsule servletCapsule}.*/
    protected String className;

    /**Stores the file name of the Servlet class which shall be encapsulated in the 
       {@link #servletCapsule servletCapsule}.*/
    protected String fileName;

    /**Contains the OutputStream to the client/or wherever the output from the Servlet is supposed to go.*/
    protected OutputStream clientOStream;

    /**This static section initializes the static variable {@link #servletStore servletStore}.*/
    static{
	servletStore=(ServletStoreInt)(new ServletStore());//Initialize the ServletStore.

    }

    /**
     *This method returns the reference to the {@link jason.service.crusader.ServletStore ServletStore} - the repository for ServletCapsule objects.<br/>
     *This is the only method to get reference to the store.
     *@return ServletStoreInt An instance of a class which implements {@link jason.service.crusader.ServletStoreInt ServletStoreInt} interface. */
    public static ServletStoreInt getServletStore(){
	return servletStore;
    }

    /**
     *This static method returns the ServletConfig implementation which is attached with this thread - redundent
     *method, should be removed.  */
    public ServletConfigImpl getServletConfig(){
	return servletConfig;
    }

    private HttpRequestParser parser;

    /**
     *Default constructor initializes all the non static variable to null. Required by the {@link jason.service.pool.ThreadPool ThreadPool} class. */
    public CrusaderCore( HttpRequestParser parser ){
	this.parser = parser;
	servletCapsule = null;
	className = null;
	fileName = null;
	clientOStream = null;

	servletRequest = new HttpServletRequestImpl( parser );
	servletResponse = new HttpServletResponseImpl();
	servletConfig = new ServletConfigImpl();
    }

    /**
     *This method is used exclusively by the {@link jason.service.crusader.Crusader Crusader} Engine to set the OutputStream through
     *which the output from the Servlet shall be channelised out.
     *@param <b>clientOStream</b> The java.io.OutputStream to the client. */
    public void setClientOStream(OutputStream clientOStream){
	this.clientOStream=clientOStream;
    }

    /**
     *This method is called by the run() of this same class to set the {@link #className className} parameter.
     @param <b>className</b> The fully qualified class name of the Servlet class to be encapsulated in the {@link #servletCapsule servletCapsule}. */
    public void setClassName(String className){
	this.className = className;
    }

    /**
     *This method is responsible to decide what kind of initialization shall be performed for the {@link #servletCapsule servletCapsule}.<br/>
     *This method is called from the <b>run()</b> of the same class. */
    public void initializeService() throws Exception {

	fileName = CrusaderUtil.classToFileName( className, CrusaderUtil.DEFAULT_SERVLET_SERVICE );

	File servletClassFile=new File( fileName );
	if(!servletClassFile.exists()){
	    throw new ClassNotFoundException( "Servlet class file not found : " + fileName );
        }

        long fileLastModified = servletClassFile.lastModified();

        if( servletStore.containsKey( className ) ) {
	    servletCapsule = ( ServletCapsuleInt )servletStore.get( className );

	    if( fileLastModified > servletCapsule.getLastModified() ) {
		if ( servletCapsule.isSingleThreadModel() ) 
		    initializeServletCapsule( true, fileLastModified, servletClassFile );
		else 
		    initializeServletCapsule( false, fileLastModified, servletClassFile );
	    }
	    else{

		if ( servletCapsule.isSingleThreadModel() ) { 
		} else 
		    servletCapsule=( ServletCapsuleInt )servletStore.get( className );
	    }
        } else 
	    initializeServletCapsule( true, fileLastModified, servletClassFile );
    }

    /**
     *This method is responsible to initialize a new ServletCapsule and register it with the ServletStore.
     *It also can perform partial initialization if required. True - full initialization, False - partial initialization.
     *This method shall be called in only two situations - 1) If the Servlet is a new Servlet, 2) The Servlet
     *.class file has undergone some modification. <b>THIS METHOD SHALL NOT BE CALLED IN ANY OTHER SITUATION.</b>
     *This method is called only by the {@link #initializeService() <b>initializeService()</b>} of the same class.
     @param newCapsuleInitialization A <b>boolean</b> indicating wether there has to be a new instantiation and initialization of a {@link jason.service.crusader.ServletCapsule ServletCapsule} or not.
     @param fileLastModified The <b>long</b> value which indicates the last modified time for the <b>.class</b> file of theServlet.
     @param servletClassFile The <b>java .io.File</b> object representing the Servlet .class file.
    */
    protected void initializeServletCapsule(boolean newCapsuleInitialization, long fileLastModified, File servletClassFile) throws Exception {

	CrusaderLoader classLoader = new CrusaderLoader();
	classLoader.setServletClassFile( servletClassFile );
	Class servletClass = classLoader.loadClass( className );

	if ( newCapsuleInitialization ){
	    servletCapsule=new ServletCapsule();
	    servletCapsule.setSingleThreadModel( javax.servlet.SingleThreadModel.class.isAssignableFrom( servletClass ) );
	    servletCapsule.setServletName( className );
	    servletCapsule.setFileName( fileName );
            servletStore.put( className, servletCapsule );
	}

	servletCapsule.setReadyForService( false );

	servletCapsule.setLastModified( fileLastModified );
	servletCapsule.setServlet( null );
	servletCapsule.setServlet( initializeServlet(servletClass));

	servletCapsule.setReadyForService(true);
    }


    /**
     *This method is responsible to create and call the <b>init()</b> on the Servlet instance
     *This method is exclusively called by the {@link #initializeServletCapsule(boolean,long,File) initializeServletCapsule(boolean,long,File)} of the same class.
     @param servletClass The java.lang.Class object representing the Servlet class.
    */
    protected javax.servlet.Servlet initializeServlet( Class servletClass ) throws InstantiationException, IllegalAccessException{
	javax.servlet.Servlet tempServlet = ( javax.servlet.Servlet )servletClass.newInstance();

	try{

	    servletConfig=new ServletConfigImpl();
	    servletConfig.setServletContext( (ServletContext )( new ServletContextImpl("DefaultContext") ) );
	    tempServlet.init(servletConfig);

	}catch(Exception e){
	    e.printStackTrace();
	}

	return tempServlet;
    }

    /**
     *This method shall actually contain the code to initialize the services and 
     *run the Servlet embeded in the ServletCapsule - this is where all the work gets done. */
    public void run(){

	try{
	    this.initializeService();
	    servletResponse.reinitializeServletResponse();
	    Servlet servlet=(Servlet)servletCapsule;
	    servlet.service(servletRequest,servletResponse);

	    Hashtable tempHeaderHash = servletResponse.getHeaderHash();
	    byte [] tempMessageBody = servletResponse.getMessageBody();

	    HttpResponseBuilder response = ProtocolFactory.getHttpResponseBuilder( parser );

	    for ( Enumeration enum = tempHeaderHash.keys(); enum.hasMoreElements(); ) {
		String _key = ( String )enum.nextElement();
		String _value = ( String )tempHeaderHash.get( _key );
		response.setMessageHeader( _key, _value );
	    }

	    response.setContent( tempMessageBody );
	    response.buildResponse( clientOStream );

	} catch( Exception exc ) {
	    exc.printStackTrace();
	}

	servletCapsule = null;
    }
}
