// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


/**This class is designed to be the class loader for servlet classes that are requested by the client.<br/>
*This class is necessary because servlets whose definitions (.class files) have changed needs to be
*reloaded.<br/>
*All class loaders cache the class definitions that they load, which may not be totally acceptable for this
*situation.<br/>
*If the .class file of a Servlet is modified, the class definition (the Class object) also needs to be updated.<br/>
*However, due to the caching mechanism of the default class loader, the loader shall always return the
*earlier loaded Class object which is out dated. Thus Class.forName() cannot be used in this case.
*The alternative is a custom class loader, thus the motivation for a custom class loader.
*
*However, just a custom class loader is also not the answer. Classes loaded by different class loaders
*are considered to be different (even if the class is the same - the Class objects are not). If the situation
*so arise that there exists two class definitions for the same class, one prior to modification and one
*after modification, then the effect may be the generation of a ClassCastException if the modified
*Servlet object is type casted (in case of Servlet-Servlet communication through ServletContext). So,
*it is not only necessary to eleminate the pre-modification Servlet object, but also the pre-modification
*class loader that loaded the class (in effect destroying the Class definition). Thus different instances
*of this class shall be used to load the modified classes.
*
*This class is predominantly used by {@link jason.service.crusader.CrusaderCore#initializeServletCapsule(boolean, long, File) initializeServletCapsule()}
*method of the {@link jason.service.crusader.CrusaderCore CrusaderCore} class.
*
*Author - Sanjay"Ironluca"Debnath <a href="mailto:Ironluca@yahoo.com">Ironluca@yahoo.com</a></br>
*Last Modified - 01/05/2001
*/

package jason.service.crusader;

import java.io.*;

public class CrusaderLoader extends ClassLoader{

    protected Class servletClass=null;
    protected File servletClassFile=null;
    protected byte [] byteCode=null;

    /**
     *This method shall be used by {@link jason.service.crusader.CrusaderCore CrusaderCore} to pass the File object representing the .class file
     *of the Servlet class to be loaded. */
    public void setServletClassFile(File servletClassFile){ this.servletClassFile = servletClassFile; }

    /**
     *This method is being overriden to introduce a custom behavior - though it is strictly not required
     *in Java2. <b>NOTE: THIS METHOD DOES NOT ATTEMPT TO INTRODUCE A SecurityManager
     *WHICH MAY BE INTRODUCED IN AT A LATER TIME.</b>
     */
    public Class loadClass( String className, boolean resolve ) {
        //this method does not check wether the class has been previously loaded or not. It assumes that the
        //class must be loaded freshly.

        //It does not check with the SecurityManager wether it is safe to load the class or not.
	byteCode = null;
	servletClass = findLoadedClass( className );
	if (servletClass != null) {
	    return servletClass;
	}

	if(!(className.startsWith("java") || className.startsWith("javax"))) byteCode=this.getClassData();

	if(byteCode == null){
	    try{
		servletClass = findSystemClass( className );
		if ( servletClass != null ) return servletClass;
	    } catch(ClassNotFoundException cnfe){System.out.println("Not a System class");}
	}
	else 
	    servletClass = defineClass( className, byteCode, 0, byteCode.length );

	if( resolve )resolveClass(servletClass);

	return servletClass;
    }

    /**
     *This method is responsible to return the byte code read from the .class file in an array of bytes.
     *It returns null if there is some problem in reading the byte code. */
    public byte[] getClassData(){
	byte[] byteCode = null;
	try{
	    FileInputStream fromClassFile = new FileInputStream( servletClassFile );
	    ByteArrayOutputStream baos = new ByteArrayOutputStream();
	    int available = fromClassFile.available();
	    for(int i = 0;i < available; i++ )
		baos.write( fromClassFile.read() );
	    byteCode = baos.toByteArray();
	}catch( IOException ioe ){
	    byteCode = null;
	    return byteCode;
	}
	return byteCode;
    }

}

