// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.service.http1_0;

import jason.core.*;

import jason.service.http1_0.cache.*;
import jason.service.http1_0.framework.*;
import jason.core.log.Log;
import jason.core.log.EmptyLog;

import java.io.*;
import java.util.*;

/**
 * <p>
 * Abstract factory for the Server component. Each main element of the AWS server is provided by this class. It
 * hides how the element are found. Singleton mecanism is used for minimized element allocation. 
 * To work correctly, the ComponentFactory need to know how to instantiate each element. The Server Config is a 
 * configuration element that provides information for this factory.
 * </p>
 * @author "djefer" <djefer@club-internet.fr>
 * @version 1.0
 * @since 1.0
 */
public class ComponentFactory {

    private Service service;

    /**
     * @param service the Service
     * @exception ConfigException if an error occurs */
    public ComponentFactory( Service service ) throws ConfigException {
	this.service = service;
    }

    // Instantiate all major elements
    protected void loadComponents() throws ConfigException {
	htMime = new Hashtable();
	htMapping = new Properties();

	int cacheCapacity = 100;
	int cacheType = Cache.FREQUENCY;
	int cacheItemSize = 10;
	int cacheRefresh = 60;
	boolean cacheEnabled = true;
	String logFile = null;
	int logSize = 0;

	// Socket access
	port = service.getIntProperty( "port" );

	// Document access
	source = service.getStringProperty( "source" );
	mainPage = service.getStringProperty( "main-page" );
	errorPage = service.getStringProperty( "404" );

	// Cache 
	cacheEnabled = service.getBooleanProperty( "cache.enabled" );
	cacheCapacity = service.getIntProperty( "cache.capacity" );
	cacheItemSize = service.getIntProperty( "cache.size" );
	cacheItemSize = service.getIntProperty( "cache.item-size" );
	cacheRefresh = service.getIntProperty( "cache.refresh" );

	// Info
	String serverName = service.getStringProperty( "server.name" );
	if ( serverName == null ) serverName = "JASon plugin";

	if ( cacheEnabled ) {
	    cache = new Cache( cacheCapacity, cacheRefresh, cacheType );
	}

	String mimeType = service.getStringProperty( "mime" );
	initMime( mimeType );

	String mapping = service.getStringProperty( "mapping" );
	initMapping( mapping );
    }

    // Initialize mapping between relatif path and a path
    private void initMapping( String mapping ) {
	StringTokenizer st = new StringTokenizer( mapping, ";,\n", false );
	while ( st.hasMoreTokens() ) {
		String token = st.nextToken();
		int i = token.indexOf( "=" );
		if ( i > 0 ) {
		    String uri = token.substring( 0, i );
		    String path = token.substring( i + 1, token.length() );
		    htMapping.setProperty( uri, path );
		}
	}
    }

    // Syntax for MIME type :  " .ext=MIME TYPE, .ext2=MIME TYPE ... "
    private void initMime( String mimeType ) {
	    StringTokenizer st = new StringTokenizer( mimeType, ";,", false );
	    while ( st.hasMoreTokens() ) {
		String token = st.nextToken();
		int i = token.indexOf( "=" );
		if ( i > 0 ) {
		    String ext = token.substring( 0, i );
		    String mimeValue = token.substring( i + 1, token.length() );
		    htMime.put( ext, mimeValue );
		}
	    }
    }

    public String getMimeType( String ext ) {
	if ( htMime.containsKey( ext ) ) 
	    return (String)htMime.get( ext );
	return "text/html";
    }

    public IntDocFactory getDocFactory() { return docFactory; }

    public Cache getCache() { return cache; }

    public Log getLog() { return log; }

    public Hashtable getHtMime() { return htMime; }

    public int getPort() { return port; }

    public String getSource() { return source; }

    public String getMainPage() { return mainPage; }

    public String getErrorPage() { return errorPage; }

    public Vector getBuilding() { return building; }

    public Vector getDisabled() { return disabled; }

    public Vector getCgiJava() { return cgiJava; }

    public Vector getCgiBin() {	return cgiBin; }

    public String getMapping( String path ) {
	return htMapping.getProperty( path, path );
    }

    IntDocFactory docFactory = null;

    final String confMime = "Mime.cfg";
    final String confServeur = "ServerHTTP.cfg";

    // Document cache
    Cache cache = null;

    Log log = new EmptyLog();
    // MIME Type
    Hashtable htMime;
    // Mapping
    Properties htMapping;

    // Service Port
    int port = 80;

    // Main path
    String source = "";
    // Default path
    String mainPage = "index.html";
    String errorPage = "??";

    // Building page
    Vector building = new Vector();
    // Disabled page
    Vector disabled = new Vector();
    // Cgi Java
    Vector cgiJava = new Vector();
    // Cgi Bin
    Vector cgiBin = new Vector();
    
}


