// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.service.http1_0;

import jason.core.*;
import jason.core.log.*;
import jason.core.pool.SocketTask;

import jason.framework.protocol.*;
import jason.framework.protocol.http.*;

import jason.service.http1_0.cache.*;
import jason.service.http1_0.framework.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * <p>
 * Core for the HTTP handling.
 * When a client send a request, this is received by the HTTPService
 * that root it in this class that sends a valid response depending on the
 * HTTP 1.0 protocol.
 * </p>
 * <p>
 * A dispatch depends on the protocol.
 * If HTTP is found, a part try to send the appropriate document using
 * cache if enabled.
 * </p>
 * <p>
 * Version 1.1 : This version is adapted to the new framework 
 * </p>
 * @author Alexandre Brillant <abrillant@wanadoo.fr>
 * @version 1.1
 * @since 1.0
 * @see Runnable
 */
public class HostHandling extends SocketTask implements IntHttpCode {

    private boolean traceRequest = false;

    public HostHandling( Service service, 
			 Socket sck, 
			 ComponentFactory config ) throws ServiceException {
	super( sck );
	this.service = service;
	this.config = config;
	if ( config == null ) throw new ServiceException( "Invalid config parameter" );
	traceRequest = service.getBooleanProperty( "trace" );
    }

    public void setConfig( ComponentFactory config ) {
	this.config = config;
    }

    // Return a page
    private void sendPage( byte[] content, String source ) {
	try {
	    HttpResponseBuilder builder = ProtocolFactory.getHttpResponseBuilder( parser );
	    builder.setServer( "Jason/Http" );
	    builder.setContent( content );

	    if ( source != null ) {
		int i = source.lastIndexOf( "." );
		if ( i > -1 ) {
		    String ext = source.substring( i + 1, source.length() );
		    String mimeType = config.getMimeType( ext );
		    builder.setContentType( mimeType );
		}
	    }

	    builder.buildResponse( out );

	} catch( IOException ex ) {
	    ex.printStackTrace();
	}
    }

    private void sendPage( byte[] content ) {
	sendPage( content, null );
    }

    private void sendInformationMessage( String message ) {
	if ( !traceRequest ) return;

	if ( service != null ) {
	    ( ( (AbstractService)service ).getServiceContext() ).sendInformationMessage( service.getName(), message );
	} else
	    System.out.println( message );
    }

    private void sendErrorMessage( String message ) {
	if ( service != null ) {
	    ( ( (AbstractService)service ).getServiceContext() ).sendErrorMessage( service.getName(), message );
	} else 
	    System.err.println( message );
    }

    // Send a document
    private void procDoc( String path ) {
	Cache cache = config.getCache();
	String source = config.getSource();
	Log log = config.getLog();
	boolean cacheOk = ( cache != null );
	byte[] doc1 = null;

	String error = null;

	// building pages
	Vector building = config.getBuilding();

	for ( Enumeration enum = building.elements(); enum.hasMoreElements(); ) {
	    String s = (String)enum.nextElement();
	    if ( s.endsWith( path ) ) {
		sendError( "Building " + path );
		return;
	    }
	}

	// unauthorized pages
	Vector disabled = config.getDisabled();
	for ( Enumeration enum = disabled.elements(); enum.hasMoreElements(); ) {
	    String s = (String)enum.nextElement();
	    if ( s.endsWith( path ) ) {
		sendError( "Disabled " + path );
		return;
	    }
	}

	if ( cacheOk ) {
	    // Document in the cache
	    doc1 = cache.getDoc( new File( source, path ).toString() );

	    if ( doc1 != null ) {
		sendInformationMessage(  "--> (x) " + path );
	    } else // Miss
		sendInformationMessage( "--> ( ) " + path );
	} else
	    sendInformationMessage( "--> " + path );

	// We don't know this document
	if ( doc1 == null ) {
	    try {
		// Make a document
		File fDoc = new File( source, path );

		if ( !fDoc.exists() ) {
		    sendError( "Document not found" );
		    return;
		}

		StringBuffer sbDoc = Cache.loadDoc( fDoc );
		doc1 = sbDoc.toString().getBytes();

		if ( sbDoc.length() > 0 ) {
		    // Add it in the cache

		    if ( cache != null )
			cache.addDoc( sbDoc.toString().getBytes(), fDoc.toString() );
		} else {
		    sendError( CO_NOT_FOUND );
		    return;
		}
	    } catch( IOException ex ) {
		ex.printStackTrace();
		// Loading error
		sendError( CO_INTERNAL_SERVER_ERROR );
		return;
	    }
	}

	// Page sending
	sendPage( doc1, path );
    }

    // Process the request
    public void run() {
	try {
	    if ( socket == null || 
		 config == null ) {
		System.err.println( "** Inner error : The HostHandling is runned without setting the appropriate configuration" );
		return;
	    }

	    InputStream in = null;

	    try {

		in = socket.getInputStream();
		out = socket.getOutputStream();
		parser = ProtocolFactory.getHttpRequestParser();
		parser.parse( in );
		dispatchRequest();

	    } finally {

		try {
		    // We shut the connection
		    in.close();
		    out.flush();
		    out.close();
		} catch( Exception exc ) {
		    exc.printStackTrace();
		}

	    }

	} catch( Exception ex ) {
	    System.err.println( ex.toString() );
	    ex.printStackTrace();
	    return;
	}
    }

    // Dispatch UserRequest depending on the protocol
    private void dispatchRequest() throws IOException {
	// Check the protocol
	if ( parser.getProtocol().startsWith( "HTTP" ) )
	    processHTTPRequest();
	else
	    sendError( CO_NOT_IMPLEMENTED );
    }

    private void processHTTPGetRequest() {
	String page = parser.getRequestURI();
	if ( page.trim().equals( "/" ) ) // Default page
	    page = config.getMainPage();
	page = config.getMapping( page ); // Mapping
	procDoc( page );
    }

    // Process HTTP request. Old code, To update later
    private void processHTTPRequest() throws IOException {
	String method = parser.getMethod();
	if ( "GET".equals( method ) || "POST".equals( method ) ) {
	    processHTTPGetRequest();
	} else
	    sendError( CO_NOT_IMPLEMENTED );
    }

    // Returning an error page
    private void sendError( String mess ) {
	StringBuffer res = new StringBuffer();

	File f = new File( config.getErrorPage() );
	if ( f.exists() ) {
	    try {
		BufferedReader br = new BufferedReader( new FileReader( f ) );
		String line;
		while ( ( line = ( br.readLine() ) ) != null ) {
		    res.append( line ).append( "\n" );
		}
	    } catch( Exception exc ) {}
	} else {
	    // Default response
	    res.append( "<HTML>" );
	    res.append( " <HEAD>" );
	    res.append( "   <TITLE>ERROR</TITLE>" );
	    res.append( " </HEAD>" );
	    res.append( "<BODY>" );
	    res.append( "<H1>" + mess + "</H1>" );
	    res.append( "<HR>" );
	    res.append( "Contact ->" );
	    res.append( "<a href=\"mailto:master@ww3c.com\">" );
	    res.append( "master@ww3c.com" );
	    res.append( "</a>" );
	    res.append( "</BODY>" );
	    res.append( "</HTML>" );
	}

	sendPage( res.toString().getBytes() );
    }

    private OutputStream out;
    private HttpRequestParser parser;
    private ComponentFactory config;
    private Service service;
}
