// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.service.http1_0.cache;

import jason.service.http1_0.framework.*;

import java.util.*;
import java.io.*;

/**
 * Cache item. 
 * <p>
 * It contains a page and more data to update the cache if the tied filed is updated
 * </p>
 * TODO : Remove this cache and insert it in the core level with a better design
 *
 * @author Alexandre Brillant <abrillant@wanadoo.fr>
 * @version 1.1
 * @since 1.0
 */
class CacheItem  {

    private CacheItem() {
	freq = 0;
    }

    /**
     * @param doc the page content
     * @param pathDoc the absolute path for the content */
    public CacheItem( byte[] doc, String pathDoc ) {
	this();
	File fic = new File( pathDoc );
	this.lastModified = fic.lastModified();
	this.doc = doc;
	this.pathDoc = pathDoc;
    }

    public CacheItem( byte[] doc, 
		      String pathDoc, 
		      long lastModified ) {
	this();
	this.lastModified = lastModified;
	this.doc = doc;
	this.pathDoc = pathDoc;
    }

    public void incFrequency() { freq++; }

    public int getFrequency() { return freq; }

    public String getPath() { return pathDoc; }

    public byte[] getDoc() { return doc; }

    public long getLastModified() { return lastModified; }

    public String toString() { return "[" + getPath() + ":" + freq + "]"; }

    long lastModified;
    int freq;
    byte[] doc;
    String pathDoc;
}

/**
 * Cache for HTTP Request
 *
 * @author Alexandre Brillant <abrillant@wanadoo.fr>
 * @version 1.1
 * @since 1.0
 * @see Thread */
public class Cache extends Thread {
    public final static int FREQUENCY = 1;
    public final static int FIFO = 2;

    public Cache( int capacity, 
		  int refresh, 
		  int type ) {
	setPriority( MIN_PRIORITY );
	this.capacity = capacity;
	this.type = type;
	this.refresh = ( 1000 * refresh );
	start();
    }

    public void setCapacity( int capacity ) { 
	this.capacity = capacity; 
    }

    /**
     * @param type cache type (FIFO or FREQUENCY) */
    public void setType( int type ) { 
	this.type = type; 
    }

    public void setRefresh( int refresh ) { 
	this.refresh = ( 1000 * refresh );
    }

    /**
     * Add an item in the cache */
    public synchronized void addDoc( byte[] doc, String path ) {
	addDoc( new CacheItem( doc, path ) );
    }

    // Store an item
    public synchronized void addDoc( CacheItem item ) {
	items.addElement( item );
	htItems.put( item.getPath(), item );

	// Choose a political for updating the cache
	if ( items.size() >= capacity ) {
	    // Remove the older one
	    if ( type == FIFO ) {
		CacheItem item1 = (CacheItem)items.elementAt( 0 );
		htItems.remove( item1.getPath() );
		items.removeElementAt( 0 );
	    } else {
		// Remote the last used one
		CacheItem minItem = (CacheItem)items.elementAt( 0 );
		for ( Enumeration enum = items.elements(); enum.hasMoreElements(); ) {
		    CacheItem item1 = (CacheItem)enum.nextElement();
		    if ( item1.getFrequency() < minItem.getFrequency() ) {
			minItem = item1;
		    }
		}
		htItems.remove( minItem.getPath() );
		items.removeElement( minItem );
	    }
	}
    }

    /** @return the document for a relative path */
    public synchronized byte[] getDoc( String path ) {
	CacheItem item = (CacheItem)htItems.get( path );
	if( item != null ) {
	    item.incFrequency();
	    return item.getDoc();
	}
	return null;
    }

    public synchronized void stat() {
	for ( Enumeration enum = items.elements(); enum.hasMoreElements(); ) {
	    CacheItem item1 = (CacheItem)enum.nextElement();
	    System.out.println( item1.getPath() + "  " + item1.getFrequency() );
	}
    }

    /**
     * Load a page */
    public static StringBuffer loadDoc( File doc ) throws IOException {
	StringBuffer sbRes = new StringBuffer();
	FileInputStream fIn = new FileInputStream( doc );
	try {
	    int n = 0;
	    while ( ( n = fIn.read() ) != -1 ) {
		sbRes.append( (char)n );
	    }
	} finally {
	    fIn.close();
	}
	return sbRes;
    }

    // Update the cache in a background process
    // To be removed later
    public void run() {
	try {
	    for ( ;; ) {
		sleep( refresh );
		for ( Enumeration enum = items.elements(); enum.hasMoreElements(); ) {
		    CacheItem item1 = (CacheItem)enum.nextElement();
		    File fic = new File( item1.getPath() );
		    if ( fic.lastModified() > item1.getLastModified() ) {

			try {
			    StringBuffer sbTmp = loadDoc( fic );
			    CacheItem itemTmp = new CacheItem( sbTmp.toString().getBytes(), 
							       item1.getPath(), 
							       fic.lastModified() );
			    htItems.put( item1.getPath(), itemTmp );
			} catch( IOException ex ) {
			    System.err.println( ex.toString() );
			}
		    }
		}
	    }
	} catch( InterruptedException ex ) {
	    return;
	}
    }

    private int capacity;
    private int type;
    private int refresh = ( 1000 * 60 );
    private Vector items = new Vector();
    private Hashtable htItems = new Hashtable();
}


