// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core;

import java.io.Serializable;
import java.util.Enumeration;
import jason.core.ServerConfig;

/**
 * <p>This class is the common class for services implementation. 
 * Its role is to managed the service state (START/STOP). It
 * takes into consideration the XML service configuration file
 * by using the <code>ServiceContext</code>. So user can retreive
 * properties for handling correctly the service. The container
 * calls the <code>init</code> method and the <code>start</code> method
 * automatically. The start method will call by default the run method
 * that should contains the user core function.
 * <p>
 * The AbstractService doesn't behave excatly like a <code>Thread</code>. 
 * For the Thread support, please look at the <code>ThreadableService</code> class
 * </p>
 * <p>Developer can made its service persistent thanks to the <code>setPersistent</code> method.
 * This method must be changed before the <code>start</code> method. Then during the <code>init</code>
 * the <code>setPersistent</code> wish is delegated to the <code>ServiceContext</code>. User can change the behavior in
 * the <code>notifyInit</code> method. Then, if the service is set persistent, the container will call the
 * <code>getServiceState</code> method during the <code>stop</code> method for storing the service state. On the contrary, 
 * when starting a service, the container will reload the old service state. A current implementation tried to get all 
 * value for each <code>get</code> starting method name, writing the state is writing for all tied <code>set</code> method.
 * </p>
 * @author "Alexandre Brillant" <abrillant@wanadoo.fr>
 * @version 1.4
 * @since 1.0
 * @see Service
 * @see ThreadableService
 * @see SocketService
 * @see ServiceContext
 * @see ServiceVersion */
public abstract class AbstractService implements 
					  Service, 
					  Serializable,
					  ServiceDataListener {

    public static final int NOSTATE = -1;
    public static final int INITIALIZED = 0;
    public static final int STOPPED = 1;
    public static final int STARTED = 2;

    /**
     * Default constructor, nothing to do */
    public AbstractService() {
	super();
    }

    /**
     * Initialize the current service. You shouldn't override this method,
     * if you want to be notified when this method if done, please use
     * the notifyServiceInitReady rather */
    public void init( ServiceContext sc ) {
	sc.setPersistent( isPersistent() );
	setServiceContext( sc );
	setVerbose( sc.isVerbose() );
	setTest( sc.isTest() );
	setName( sc.getName() );
	// Receive all data change by default
	ServiceDataSpace.addServiceDataListener( "*", this );
	notifyInitReady();
	sc.notifyInit();
	state = INITIALIZED;
    }

    private boolean test = true;

    /**
     * This method shows for the first run call a message
     * that control your service is really running */
    protected void setTest( boolean test ) {
	this.test = test;
    }

    /**
     * @return <code>true> is the test mode is supported */
    protected boolean isTest() { return test; }

    /**
     * @return an enumeration of all property name */
    public Enumeration getPropertiesName() {
	checkForContext();
	return getServiceContext().getPropertiesName();
    }

    /** Reset the starting property. it can be only called when the
     * <code>init</code> method has been called. So, be careful and called it rather
     * in notifiyInitReadt called. If you invoke it before a <code>RuntimeException</code> will be
     * called */
    protected void setProperty( String name, Object value ) throws ServiceException {
	checkForContext();
	getServiceContext().setProperty( name, value );
    }

    /**
     * @return a Service property as a string */
    public String getStringProperty( String name ) {
	return "" + getProperty( name );
    }

    /**
     * @return a Service property as an integer */
    public int getIntProperty( String name ) {
	try {
	    return Integer.parseInt( "" + getProperty( name ) );
	} catch( NumberFormatException exc ) {
	    return 0;
	}
    }

    /**
     * @return a Service property as a boolean */
    public boolean getBooleanProperty( String name ) {
	return "true".equals( getProperty( name ) );
    }

    /**
     * @return a property service */
    public Object getProperty( String name ) {
	checkForContext();
	return getServiceContext().getProperty( name );
    }

    /**
     * Check that the context is not empty, else
     * it throws a <code>RuntimeException</code>, 
     * because this is an illegal access */
    protected void checkForContext() {
	if ( getServiceContext() == null ) 
	    throw new RuntimeException( "Invalid method access, please call it after when the context is ready" );
    }

    /**
     * This method is called when the <code>init</code> method is 
     * runned */
    protected void notifyInitReady() {}

    /**
     * This method is called at the end of the <code>start</code> method */
    protected void notifyStartReady() {}

    /**
     * This method is called at the end of the <code>stop</code> method */
    protected void notifyStopReady() {}

    // Can be overrided for information
    public void notifyData( String key, Object value ) {}

    public void restart() {}

    public void stop() {
	postStop();
	state = STOPPED;
	checkForContext();
	getServiceContext().notifyStop();
	notifyStopReady();
    }

    private boolean cAutoStop = false;

    /** Decide to stop the service automatically after the run invocakation, by default the value is <code>false</code> */
    protected void setAutoStop( boolean autoStop ) {
	this.cAutoStop = autoStop;
    }

    /** @return <code>true</code> if the service is automatically stop after the <code>run</code> method */
    protected boolean isAutoStop() {
	return cAutoStop;
    }

    private boolean cPersistent = false;

    /** Make the current service persistent through the <code>getServiceState</code> */
    public void setPersistent( boolean persistent ) {
	this.cPersistent = persistent;
    }

    /** @return <code>true</code> if the service state can be store */
    public boolean isPersistent() {
	return cPersistent;
    }

    /** Action before stopping the service */
    protected void postStop() {
	if ( getServiceState() != null )
	    getServiceState();
    }

    /** @return the current service state */
    public ServiceState getServiceState() {
	return new ServiceState( this );
    }

    /**
     * Start the service. By default this method will call the
     * <code>run</code> method */
    public void start() {
	state = STARTED;

	checkForContext();
	getServiceContext().notifyStart();
	notifyStartReady();	

	actionForStart();
    }

    /**
     * This method is called at the end of the <code>start</code> method. 
     * By default it tries to call the <code>run</code> method */
    protected void actionForStart() {
	try {
	    run();
	    if ( isAutoStop() )
		stop();
	} catch( RuntimeException exc ) {
	    fireCriticalErrorDetected( exc );
	}
    }

    // Service state information
    public boolean isStarted() { return state == STARTED; }
    public boolean isStopped() { return state == STOPPED; }

    /**
     * this method return <code>true</code> when the
     * service has a <code>ServiceContext</code> */
    public boolean isInitialized() { return state == INITIALIZED; }

    public int getState() { return state; }

    private ServiceVersion version = null;

    /**
     * Reset the version of the service */
    public void setVersion( ServiceVersion version ) { this.version = version; }
    public ServiceVersion getVersion() { 
	if ( version == null )
	    version = new ServiceVersion( 1 );
	return version; 
    }

    /**
     * By default it returns <code>false</code>*/
    public boolean isOverride() { return false; }

    /**
     * @return Information about the service */
    public ServiceInfo getInfo() { return null; }

    /**
     * Reset the name of the service */
    public void setName( String name ) { this.name = name; }
    public String getName() { return name; }

    /**
     * User should override this method for running a task. This is the
     * core of the service running state. Note that you shouldn't use
     * a main loop like for standard thread. You just have to notify
     * that this service use a main loop for the <code>setThreadLoopMode</code> method */
    protected void run() {}

    /**
     * @param ServiceContext. Should be do once before starting the service */
    public void setServiceContext( ServiceContext serviceContext ) { this.serviceContext = serviceContext; }

    /**
     * @param ServiceContext share context between service */
    public ServiceContext getServiceContext() { return serviceContext; }

    /**
     * @return <code>true</code> if the serviceContext is available */
    public boolean hasServiceContext() { return serviceContext != null; }

    /**
     * Decide for the verbosity */
    protected void setVerbose( boolean verbose ) { 
	this.verbose = verbose;
    }

    /**
     * This method will avoid to show the Information message type
     * Look at the <code>ServiceContect</code> for more information */
    public boolean isVerbose() {
	return verbose;
    }

    /**
     * Check for the thread support from a thread pooling system */
    public boolean isThreadable() { return serviceContext.isThreadable(); }

    /**
     * Check for a loadbalancing for each new process */
    public boolean isProcessable() { return serviceContext.isProcessable(); }

    /**
     * Send an error message */
    public void sendErrorMessage( String message ) {
	checkForContext();
	getServiceContext().sendErrorMessage( getName(), message );
    }

    /**
     * Send an inner error exception */
    public void sendInnerError( Exception exc ) {
	checkForContext();
	getServiceContext().sendInnerError( exc );
    }

    /**
     * Send a warning message */
    public void sendWarningMessage( String message ) {
	checkForContext();
	getServiceContext().sendWarningMessage( getName(), message );
    }

    /**
     * Send an information message, this message will not be taken into
     * account if the verbose value is <code>false</code> */
    public void sendInformationMessage( String message ) {
	checkForContext();
	getServiceContext().sendInformationMessage( getName(), message );
    }

    /**
     * Send a custom message with a tied type */
    public void sendCustomMessage( int type, String message ) {
	checkForContext();
	getServiceContext().sendCustomMessage( getName(), type, message );
    }

    private boolean privateState = false;

    /**
     * Design the accessibility level for this service */
    public boolean isPrivate() { return privateState; }

    /**
     * This method lets the user change the accessibility for service */
    protected void setPrivate( boolean privateState ) {
	this.privateState = privateState;
    }

    private ServiceType type;

    /**
     * @return the Service type */
    public ServiceType getType() {
	if ( type == null ) {
	    try {
		type = new ServiceType( ServiceType.CUSTOM_SERVICE );
	    } catch( ServiceException exc ) {}
	}
	return type;
    }

    /**
     * This method must be called for critical error, it will avoid the service
     * to start */
    protected void fireCriticalErrorDetected( Exception exc ) {
	enabled = false;
	if ( hasServiceContext() )
	    getServiceContext().notifyCriticalError( exc );
    }

    /**
     * @param type Reset the service type */
    protected void setType( ServiceType type ) {
	this.type = type;
    }

    /**
     * Manage a request 
     * @return the request's response */
    public ServiceResponse manageRequest( ServiceRequest request ) throws ServiceException {
	// Default behavior
	throw new ServiceException( "Can't manage the request " + request );
    }

    protected ServiceResponse manageRequestForType( ServiceType type, 
						    ServiceRequest request ) throws ServiceException {
	checkForContext();
	return getServiceContext().fireRequestForType( type, request );
    }

    private boolean enabled = true;

    /**
     * This method is used only when starting the first time this
     * service
     * @return <code>true</code> is the service can be run */
    public boolean isEnabled() { return enabled; }

    /**
     * Change the enabled mode. This method must be called
     * before the start method such as in the constructor 
     * @param enabled can start or not this service */
    protected void setEnabled( boolean enabled ) {
	this.enabled = enabled;
    }

    protected transient ServiceContext serviceContext;
    protected String name;

    private int state = NOSTATE;
    private boolean verbose = false;
}








