// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.io.File;

import jason.core.pool.ThreadCapsule;
import jason.core.pool.ThreadPool;
import jason.core.pool.ThreadPoolImpl;

import org.w3c.dom.*;
import org.xml.sax.*;

/**
 * <b>Created Wed Mar 27 10:34:40 2002</b>
 * <p>
 * This class describes a domain content.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0 */
public class Domain {

    private String name;
    private boolean verbose = true;
    private JasonEnv env;
    private ServiceWorkingDirectoryManager manager = null;

    /**
     * @param name Domain name */
    public Domain( String name, boolean verbose, JasonEnv env ) {
	super();
	this.name = name;
	this.env = env;
	this.verbose = verbose;
	sm = new ServicesManager( this );
	setModel( new ServiceContainerDataModel() );
	setDoorModel( new DoorDataModel() );
	sm.setModel( scdm );

	manager = new ServiceWorkingDirectoryManager();
	scdm.addServiceContainerDataModelListener( manager );
    }

    private int adminPort = 8085;
    private String adminUser = "jason01";
    private String adminPassword = "z86";

    /** @return admin name */
    public String getAdminUser() { return adminUser; }
    /** @return admin password */
    public String getAdminPassword() { return adminPassword; }

    /** set the port for administration of JAson, by default 8085 */
    public void setAdminPort( int adminPort ) {
	this.adminPort = adminPort;
    }

    /** @return the administration port of JASon, by default 8085 */
    public int getAdminPort() {
	return adminPort;
    }

    /** Set the dataModel for available <code>ServiceContainer</code> object */
    public void setModel( ServiceContainerDataModel model ) {
	this.scdm = model;
    }

    /** @return the dataModel with available <code>ServiceContainer</code> object */
    public ServiceContainerDataModel getModel() {
	return scdm;
    }

    /** Reset the model for available door */
    public void setDoorModel( DoorDataModel model ) {
	this.ddm = model;
    }

    /** @return the model for available door */
    public DoorDataModel getDoorModel() {
	return ddm;
    }

    /**
     * Add default service as a part of the Domain */
    public void initDefaultService() {
	try {
	    addService( new jason.core.service.Scheduler() );
	    addService( new jason.core.service.LogService() );
	    addService( new jason.core.service.TaskService() );
	    addService( new jason.core.service.UpdatableService() );
	    addService( new jason.core.service.admin.AdminService( getAdminPort() ) );
	} catch( JasonException exc ) {
	    System.err.println( "Error while initializing the Domain" );
	    exc.printStackTrace();
	}
    }

    // Location of the log
    private String logLocation = "log";
    // AutoClean for the log
    private boolean logAutoClean = true;

    /**
     * @return true if the log is deleted when starting JASon */
    public boolean isLogAutoClean() { return logAutoClean; }

    /**
     * @return the log location */
    public String getLogLocation() { return logLocation; }

    /**
     * Set the log strategy. This is for service that supports the
     * Log.TYPE type
     * @param autoClean if true, it will delete when starting the log */
    public void setLogAutoClean( boolean autoClean ) {
	logAutoClean = autoClean;
    }

    /**
     * Set the log location. This is for service that supports the Log.TYPE type 
     * @param location a path for the log location */
    public void setLogLocation( String location ) {
	logLocation = location;
	System.setProperty( "jason.log.location", location );
    }

    // Location of the configuration file
    private String location;

    /** 
     *	Set the location of the configuration file */
    public void setLocation( String location ) {
	this.location = location;
    }

    /**
     * @return the location of the file */
    public String getLocation() {
	return location;
    }

    /** @return the path containing the configuration file */
    public String getConfigurationFilePath() {
	File f = new File( location );
	return "" + f.getParentFile();
    }

    public boolean isVerbose() { return verbose; }

    /**
     * Return the current environnement */
    JasonEnv getEnv() { return env; }

    public void init() {
	if ( mh == null )
	    mh = new ConsoleMessageHandler( this );
	tp = new ThreadPoolImpl( 32 );
	
    }

    private ThreadPool tp;

    /**
     * Return a thread capsule */
    public ThreadCapsule getThreadCapsule() { return tp.getThreadCapsule(); }

    /**
     * Start the domain and all tied services */
    public void run() {
	DomainRunner dr = new DomainRunner( this );
	dr.run();
    }

    /**
     * Save the current domain state for the next running */
    public void stop() {
	sm.notifyStop();
	JasonSecurityManager.enabledExit( true );
	System.exit( 0 );
    }

    /**
     * Set the message handler. This class name must implement the
     * <code>MessageHandler</code> interface. 
     * @param cl a Class name
     * @throw JasonException for invalid class */
    public void setMessageHandler( String cl ) throws JasonException {
	if ( cl == null ) 
	    throw new JasonException( "Invalid messageHandler" );
	try {
	    Class _cl = Class.forName( cl );
	    setMessageHandler( ( MessageHandler )_cl.newInstance() );
	} catch( Exception exc ) {
	    exc.printStackTrace();
	    throw new JasonException( "Invalid messageHandler " + cl );
	}
    }

    /**
     * @deprecated
     * Add a services location. This is typically another xml file */
    public void addServices( String location ) throws JasonException {
	throw new JasonException( "Method not supported" );
    }

    /**
     * Add a single service location. This is another xml file */
    public void addService( File location ) throws JasonException {
	sm.addService( location );
    }

    /** Add a service in its binary form */
    public void addService( String fileName, byte[] service ) throws JasonException {
	// Write it to the services directory
	DoorDataModel mModel = getDoorModel();
	boolean mWriteOk = false;
	for ( int i = 0; i < mModel.doorCount(); i++ ) {
	    Door mDoor = mModel.getDoorAt( i );
	    if ( mDoor.writeService( fileName, service ) ) {
		mWriteOk = true;
		break;
	    }
	}
	if ( !mWriteOk ) {
	    throw new JasonException( "Can't store the service " + fileName );
	}
    }

    /**
     * Add a single service location. This is another xml file */
    public ServiceContainer addService( String location ) throws JasonException {
	return sm.addService( location );
    }

    public void addService( String location, boolean start ) throws JasonException {
	ServiceContainer mContainer = addService( location );
	if ( start )
	    mContainer.start();

	// Invalidate each service that point to the location
	for ( int i = 0; i < scdm.serviceContainerCount(); i++ ) {
	    ServiceContainer _mContainer = scdm.getServiceContainerAt( i );
	    if ( location.equals( "" + _mContainer.getFilePath() ) ) {
		_mContainer.setFilePath( new File( location ) );
	    }
	}
    }

    /**
     * This method shouldn't be called by user, there's for the moment
     * no reason that could explain why the user should call this method.
     * @param s a service instance */
    private void addService( Service s ) throws JasonException {
	ClassServiceBuilder csb = new ClassServiceBuilder( sm, this );
	ServiceContainer _sc = new ServiceContainer( sm, 
						     s, 
						     s.getName(), 
						     s.getClass().getName(), 
						     false );
	sm.addServiceContainer( _sc );
    }

    /** Replace the old service container with the new one */
    public void replaceServiceContainer( ServiceContainer oldSc, ServiceContainer newSc ) {
	sm.removeServiceContainer( oldSc );
	// Start the new service
	sm.addServiceContainer( newSc );
	// Start it
	if ( newSc.isStopped() ) {
	    try {
		newSc.start();
	    } catch( JasonException exc ) {
		exc.printStackTrace();
	    }
	}
    }

    /**
     * Add a door for further services intergration.
     * @param name name of the door
     * @param location location of the services
     * @param _class can be equals to "", contains a Door class implementation */
    public void addDoor( String name, 
			 String location, 
			 String _class ) throws JasonException {
	if ( "".equals( _class ) ) 
	    _class = "jason.core.PathDoor";
	try {
	    Door _door = ( Door )( Class.forName( _class ) ).newInstance();
	    _door.init( new DoorContext( this, location, env.getConfigurationFile().getParentFile().toString() ) );
	    ddm.addDoor( _door );

	    Service[] s = _door.getServices();
	    if ( s != null ) {
		for ( int i = 0; i < s.length; i++ ) {
		    addService( s[ i ] );
		}
	    }
	} catch( Exception exc ) {
	    exc.printStackTrace();
	    throw new JasonException( "Can't add door " + name + " > " + exc.getMessage() );
	}
    }

    /**
     * This method is used for overriding a set of property defined both in 
     * an external xml file and in the domain definition. This properties
     * help to redefine default property
     * @param service this is a service node with a location from a domain definition */
    public void overrideServiceProperty( String location, Element service ) throws JasonException {
	NodeList l = service.getElementsByTagName( "property" );
	for ( int i = 0; i < l.getLength(); i++ ) {
	    Element e = (Element)l.item( i );
	    String name = e.getAttribute( "name" );
	    String value = e.getAttribute( "value" );
	    if ( "".equals( value ) )
		value = e.getFirstChild().getNodeValue();
	    sm.setPropertyForServiceLocation( location, name, value );
	}
    }

    public ServicesManager getServicesManager() { return sm; }

    private ServicesManager sm;
    private ServiceContainerDataModel scdm;
    private DoorDataModel ddm;
    private MessageHandler mh;

    public void setMessageHandler( MessageHandler mh ) {
	this.mh = mh;
	mh.setDomain( this );
    }

    /**
     * @return available service */
    public Service[] getServices() {
	return sm.getServices();
    }

    /**
     * @return available service container */
    public ServiceContainer[] getServiceContainer() {
	return scdm.getState();
    }

    /** @return the number of service container */
    public int serviceContainerCount() {
	return scdm.serviceContainerCount();
    }

    /**
     * @return the current domain name */
    public String getName() { return name; }

    /**
     * @return the current messageHandler */
    public MessageHandler getMessageHandler() { return mh; }

    // Proxy for the MessageHandler

    public void sendErrorMessage( String content ) {
	getMessageHandler().sendErrorMessage( "domain/" + getName(), content );
    }

    public void sendWarningMessage( String content ) {
	getMessageHandler().sendWarningMessage( "domain/" + getName(), content );
    }

    public void sendInformationMessage( String content ) {
	getMessageHandler().sendInformationMessage( "domain/" + getName(), content );
    }
}

// Domain ends here






