// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core;

import java.io.File;
import java.io.FileInputStream;

import org.apache.xerces.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * <b>Created Sun Sep 01 18:32:18 2002</b>
 * <p>
 * This is builder for Service loaded and managed by an XM file
 * </p>
 * @author Alexandre Brillant
 * @version 1.0
 * @see ServiceContainer
 * @see ServicesManager
 */
public class FileServiceBuilder extends AbstractServiceBuilder {
    private Domain d;
    private ServicesManager sm;

    public FileServiceBuilder( ServicesManager sm, Domain d ) {
	this.sm = sm;
	this.d = d;
    }

    /** @return the service container for the service at this file location */
    public ServiceContainer buildService( String location ) throws JasonException {
	super.buildService( location );
	Element root = parse( null,
			      location );

	if ( !"service".equals( root.getTagName() ) )
	    throw new JasonException( "Invalid service xml file for location " + location );

	return buildServiceContainer( location, root );
    }

    private ServiceContainer buildServiceContainer( String location, Element e ) throws JasonException {
	String name = e.getAttribute( "name" );
	String _class = e.getAttribute( "class" );
	String classpath = e.getAttribute( "classpath" );
	boolean test = (! "false".equals( e.getAttribute( "test" ) ) );
	ClassLoader cl = null;
	File fLocation = new File( location );

	if ( !"".equals( classpath ) )
	    cl = new JasonClassLoader( fLocation.getParentFile(), classpath );

	ServiceContainer sc = new ServiceContainer( sm, cl, name, _class, test );
	sc.setServiceContainerBuilder( this );

	sc.setFilePath( fLocation );
	sc.setThreadable( !( "false".equals( e.getAttribute( "thread" ) ) ) );
	sc.setProcessable( !( "false".equals( e.getAttribute( "process" ) ) ) );

	// Add info
	NodeList l = e.getElementsByTagName( "info" );
	if ( l != null ) {
	    if ( l.getLength() > 0 ) {
		Element i = (Element)l.item( 0 );
		sc.setAuthor( i.getAttribute( "author" ) );
		sc.setVersion( i.getAttribute( "version" ) );
	    }
	}

	// Verbose
	l = e.getElementsByTagName( "verbose" );
	if ( l != null ) {
	    for ( int i = 0; i < l.getLength(); i++ ) {
		Element v = ( Element )l.item( i );
		sc.setVerbose( v.getAttribute( "type" ),
			       "true".equals( v.getAttribute( "enabled" ) ) );
	    }
	}

	// Add properties
	l = e.getElementsByTagName( "property" );

	for ( int i = 0; i < l.getLength(); i++ ) {
	    Element _e = (Element)l.item( i );
	    String _name = _e.getAttribute( "name" );
	    Object _value = _e.getAttribute( "value" );
	    if ( "".equals( _value ) ) { // XML Child  
		NodeList _l = _e.getChildNodes();
		for ( int j = 0; j < _l.getLength(); j++ )
		    if ( _l.item( j ).getNodeType() != Node.TEXT_NODE ) {
			_value = ServicePropertyManager.getPropertyValue( _l.item( j ) );
			break;
		    }
	    }
	    if ( _value instanceof String )
		_value = ( (String)_value ).trim();
	    sc.setLegalProperty( _name, _value );
	}

	sc.init();

	String[] fields = { "start-period", 
			    "start-minute",
			    "start-hour",
			    "start-day",
			    "start-month" };

	for ( int i = 0; i < fields.length; i++ ) {
	    sc.setSchedulingField( fields[ i ], e.getAttribute( fields[ i ] ) );
	}

	return sc;
    }

    /**
     * Parse the services xml file */
    private Element parse( File refFile, String file ) throws JasonException {
	try {
	    DOMParser parser = new DOMParser();
	    try {
		parser.setFeature( "http://xml.org/sax/features/validation", true);
	    } catch( SAXException e ) {
	    }
	    
	    XMLErrorHandler _handler;
	    parser.setErrorHandler( _handler = new XMLErrorHandler( file ) );

	    File validFile;
	    if ( refFile == null ) 
		validFile = new File( file );
	    else
		validFile = new File( refFile.getParentFile(), file );

	    parser.parse( new InputSource( new FileInputStream( validFile ) ) );

	    if ( _handler.hasError() ) 
		throw new JasonException( "Invalid file " + file );

	    Document doc = parser.getDocument();
	    return doc.getDocumentElement();
	} catch( Exception exc ) {
	    exc.printStackTrace();
	    throw new JasonException( exc.getMessage() );
	}
    }

}

// FileServiceBuilder ends here




