// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;

import java.util.Enumeration;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/**
 * <b>Created Mon Jun 10 15:11:51 2002</b>
 * <p>
 * This is a manager for the .jas format. This format is 
 * a compact way for integrating services in JASon. This
 * format is always temporary and uncompress in the specified
 * directory each time for performance usage.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 */
public class JasManager {

    /**
     * @param f File */
    public JasManager( File f ) {
	super();   
	this.f = f;
    }

    /**
     * Load the '.jas' file */
    public void init( Domain d ) {
	this.d = d;
	processFile();
    }

    // Manage the '.jas' file
    private void processFile() {
	File p = f.getParentFile();

	// Create a directory in the parent directory 
	String fileName = getServicePath( f.toString() );
	File jasDirectory = new File( p, fileName );
	jasDirectory.mkdir();

	try {
	    // Unjar the .jas file in this new directory
	    uncompress( jasDirectory );
	} catch( Exception exc ) {
	    System.err.println( "Can't read " + f );
	}

	// Delete the .jas file
	f.delete();
    }

    private String getServicePath( String fullPath ) {
	int i = 0;
	if ( ( i = fullPath.lastIndexOf( "/" ) ) < 0 ) {
	    i = fullPath.lastIndexOf( "\\" );
	    if ( i < 0 ) 
		i = 0;
	}
	int j = fullPath.lastIndexOf( "." );
	if ( j > 0 ) 
	    return fullPath.substring( i + 1, j );
	return "temp";
    }

    // Uncompress the .jas in the target directory
    private void uncompress( File target ) throws IOException {
	JarFile jf = new JarFile( f );
	try {
	    for ( Enumeration enum = jf.entries(); enum.hasMoreElements(); ) {
		ZipEntry _entry = ( ZipEntry )enum.nextElement();
		String _name = _entry.getName();
		if ( _entry.isDirectory() ) {
		    File _d = new File( target, _name );
		    _d.mkdirs();
		} else {
		    InputStream _input = jf.getInputStream( _entry );
		    writeEntry( new File( target, _name ), _input );
		}
	    } 
	} finally {
	    jf.close();
	}
    }

    private JasManagerListener cListener = null;

    public void setJasManagerListener( JasManagerListener listener ) {
	this.cListener = listener;
    }

    // Write the entry
    private void writeEntry( File target, 
			     InputStream input ) throws IOException {

	if ( target.exists() )
	    target.delete();

	byte[] buffer = new byte[ 1024 ];
	int read = 0;
	FileOutputStream fOut = new FileOutputStream( target );
	while ( ( read = input.read( buffer ) ) > 0 ) {
	    fOut.write( buffer, 0, read );
	}
	fOut.close();
	input.close();

	if ( cListener != null )
	    cListener.notifyFile( target );
    }

    private File f;
    private Domain d;
}

// JasManager ends here






