// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.ByteArrayOutputStream;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

/**
 * <b>Created Thu Mar 28 15:44:07 2002</b>
 * <p>
 * This is a classloader for user services. This classLoader is 
 * able to find a class from a set of directories or jar/zip file
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 */
public class JasonClassLoader extends ClassLoader {

    public JasonClassLoader( ClassLoader parent, 
			     File refPath,
			     String classpath ) {
	super( parent );   
	init( refPath, classpath );
    }

    /**
     * Here a classpath definition that can contains .jar or .zip or a directory 
     * @param refPath the path the contains the xml service file
     * @param classpath the classpath for the service */
    public JasonClassLoader( File refPath, String classpath ) {
	super();
	init( refPath, classpath );
    }

    /**
     * Initialize a memory footprint from this classpath definition */
    private void init( File refPath, String classpath ) {
	StringTokenizer st = new StringTokenizer( classpath, ":;" );
	while ( st.hasMoreTokens() ) {
	    String item = st.nextToken();
	    File f = new File( refPath, item );

	    if ( item.endsWith( ".jar" ) || item.endsWith( ".zip" ) ) 
		addZip( f ); 
	    else {
		if ( f.isDirectory() )
		    addDirectory( f );
	    }
	}
    }

    // Store a list of class directories
    private Vector vd = new Vector();

    private void addDirectory( File directory ) {
	vd.addElement( directory );
    }

    private Hashtable htClassFromZip = new Hashtable();

    // Convert an inputstream to a data byte array
    private byte[] getBytesFromInputStream( InputStream in ) throws IOException {
	byte[] buffer = new byte[ 1024 ];
	ByteArrayOutputStream out = new ByteArrayOutputStream();
	while ( true ) {
	    int n = in.read( buffer );
	    if ( n == -1 ) 
		break;
	    out.write( buffer, 0, n );
	}
	in.close();
	return out.toByteArray();
    }

    // Store a list of zip/jar
    private void addZip( File zipFile ) {
	try {
	    ZipFile input = new ZipFile( zipFile );

	    try {
		for ( Enumeration enum = input.entries(); enum.hasMoreElements(); ) {
		    ZipEntry entry = (ZipEntry)enum.nextElement();
		    String name = entry.getName();
		    if ( !entry.isDirectory() && name.endsWith( ".class" ) ) {
			// Store it
			InputStream in = input.getInputStream( entry );
			byte[] buffer = getBytesFromInputStream( in );
			htClassFromZip.put( name, buffer );
		    }
		}
	    } finally {
		input.close();
	    }
	} catch( Exception exc ) {
	    // For the moment ignore any problem
	}
    }

    // Return a byte array for a class name
    private byte[] getClassBytesForName( String name ) throws IOException {
	// Look at the htClassFromZip
	byte[] array = ( byte[] )htClassFromZip.get( name );
	if ( array != null ) 
	    return array;

	// Look at directories
	for ( int i = 0; i < vd.size(); i++ ) {
	    File d = ( File )vd.elementAt( i );
	    File f = new File( d, name );

	    if ( f.exists() ) 
		return getBytesFromInputStream( new FileInputStream( f ) );
	}
	return null;
    }

    // Cache for class
    private Hashtable htClass = new Hashtable();

    public Class loadClass( String name ) throws ClassNotFoundException {
	return findClass( name );
    }

    /**
     *  This is the method where the task of class loading
     *  is delegated to our custom loader.
     * @param  name the name of the class
     * @return the resulting <code>Class</code> object
     * @exception ClassNotFoundException if the class could not be found */
    protected Class findClass(String name) throws ClassNotFoundException {

	try {
	    Class cl = (Class)htClass.get( name );
	    if ( cl != null )
		return cl;

	    String path = name.replace( '.', '/' );
	    byte[] classBytes = getClassBytesForName( path + ".class" );

	    if ( classBytes == null ) {
		// Try to find it from the System classLoader
		Class _cl = getSystemClassLoader().loadClass( name );
		if ( _cl == null )
		    throw new ClassNotFoundException( name );
		htClass.put( name, _cl );
		return _cl;
	    }
	    Class ncl = defineClass( name, classBytes, 0, classBytes.length );
	    // Store it in the cache
	    htClass.put( name, ncl );
	    return ncl;
	} catch ( Exception e ) {
	    throw new ClassNotFoundException( name );
	}
    }

}


// JasonClassLoader ends here
