// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.io.File;
import java.io.FileInputStream;

import org.apache.xerces.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

/**
 * <b>Created Wed Mar 27 09:53:26 2002</b>
 * <p>
 * This is the major configuration manager. It parses
 * the 'jason.xml' file and build an Domain object managing
 * with all available services.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0 */
public class JasonEnv {

    public static String version = "0.8";

    private String defaultDomain;
    private Domain rootDomain;
    private File file;

    /**
     * It parses the following file with a domain jason definition */
    public JasonEnv( File file ) throws JasonException {
	super();
	this.file = file;
	System.out.println( "Starting JASon " + version + " ..." );
    }

    /**
     * Set a property for the command line */
    public void setArgumentProperty( String name, String value ) {
	if ( "version".equals( name ) )
	    versionMessage();
	else
	    if ( "help".equals( name ) )
		helpMessage();
	    else
		if ( "info".equals( name ) ) {
		    System.setProperty( "jason.services.info", "true" );
		} else
		    if ( "domain".equals( name ) ) {
			defaultDomain = value;
		    } else {
			System.out.println( "Unknown option, You must use one of the following : " );
			helpMessage();
		    }
    }

    // Write the version message
    private void versionMessage() {
	System.out.println( "JASon version " + version );
	System.out.println( "(c) 2001,2002 Alexandre Brillant" );
	System.out.println( "Official Site : http://www.djefer.com/~aws" );
	System.out.println();
	System.exit( 0 );
    }

    // Write the help message
    public void helpMessage() {
	System.out.println( "-info : Show all service information" );
	System.out.println( "-domain : A domain name" );
	System.out.println( "-help : This help message" );
	System.out.println( "-version : JAson version" );
	System.exit( 0 );
    }

    /**
     * @return the main configuration file */
    File getConfigurationFile() { return file; }

    private void setRootDomain( Domain rootDomain ) {
	this.rootDomain = rootDomain;
	rootDomain.setLocation( "" + getConfigurationFile() );
    }

    /**
     * Parse the jason.xml file */
    public void init() throws JasonException {
	parse( file );
	initEnv( (Element)jasonRoot );
    }

    public void run() {
	System.setSecurityManager( new JasonSecurityManager() );

	getRootDomain().init();
	getRootDomain().run();
    }

    /**
     * Return the domain tied to the configuration file */
    public Domain getRootDomain() {
	return rootDomain;
    }

    private Node jasonRoot;

    // Parse the major configuration file
    private void parse( File file ) throws JasonException {
	try {
	    DOMParser parser = new DOMParser();
	    parser.setErrorHandler( new XMLErrorHandler( file.toString() ) );
	    parser.parse( new InputSource( new FileInputStream( file ) ) );
	    Document doc = parser.getDocument();
	    jasonRoot = doc.getDocumentElement();
	} catch( Exception exc ) {
	    exc.printStackTrace();
	    throw new JasonException( exc.getMessage() );
	}
    }

    // Initialize the configuration from the root node
    protected void initEnv( Element root ) throws JasonException {
	if ( defaultDomain == null )
	    defaultDomain = root.getAttribute( "domain" );

	boolean found = false;

	NodeList domains = root.getElementsByTagName( "domain" );
	for ( int i = 0; i < domains.getLength(); i++ ) {
	    Element domain = (Element)domains.item( i );
	    String name = domain.getAttribute( "name" );
	    if ( defaultDomain.equals( name ) ) {
		initDomain( domain );
		found = true;
		break;
	    }
	}

	if ( !found ) {
	    throw new JasonException( "Unknown domain " + defaultDomain );
	}
    }

    // Initialize this domain
    protected void initDomain( Element domain ) throws JasonException {
	String _name = domain.getAttribute( "name" );
	String _port = domain.getAttribute( "port" );
	boolean verbose = !( "false".equals( domain.getAttribute( "verbose" ) ) );
	Domain d = new Domain( _name, verbose, this );
	try {
	  d.setAdminPort( Integer.parseInt( _port ) );
	} catch( NumberFormatException exc ) {
	  throw new JasonException( "Invalid port number : " + _port );	  	  
	}

	setRootDomain( d );


	// reset the messageHandler
	NodeList mh = domain.getElementsByTagName( "messageHandler" );
	if ( mh.getLength() > 0 ) {
	    Element e = ( Element )mh.item( 0 );
	    d.setMessageHandler( e.getAttribute( "class" ) );
	}

	// add all door services
	NodeList _l = domain.getElementsByTagName( "log" );
	if ( _l.getLength() > 0 ) {
	    Element e = ( Element )_l.item( 0 );
	    boolean autoClean = "true".equals( e.getAttribute( "auto-clean" ) );
	    String location = e.getAttribute( "location" );

	    d.setLogAutoClean( autoClean );
	    d.setLogLocation( location );
	}

	d.initDefaultService();

	// add all single service
	NodeList l = domain.getElementsByTagName( "service" );
	for ( int i = 0; i < l.getLength(); i++ ) {
	    Element e = ( Element )l.item( i );
	    String mLocation = "" + ( new File( getConfigurationFile().getParentFile(), e.getAttribute( "location" ) ) );
	    d.addService( mLocation );
	    d.overrideServiceProperty( mLocation, e );
	}

	// add all door services
	l = domain.getElementsByTagName( "door" );
	for ( int i = 0; i < l.getLength(); i++ ) {
	    Element e = ( Element )l.item( i );
	    try {
		d.addDoor( e.getAttribute( "name" ), 
			   e.getAttribute( "location" ), 
			   e.getAttribute( "class" ) );
	    } catch( JasonException exc ) {
		System.err.println( "Invalid door content " + e.getAttribute( "name" ) + " -> " + exc.getMessage() );
		exc.printStackTrace();
	    }
	}
    }

}

// JasonEnv ends here
