// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core;

import java.util.Vector;
import java.io.*;

/**
 * <b>Created Thu May  9 09:58:17 2002</b>
 * <p>
 * This is the default door for a domain. This is a way to
 * integrate easily services from a directory. So, user
 * needn't to declare its services in the jason.xml file and
 * it works like a set of plugIns.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.3
 * @since 0.6
 * @see Door
 */
public class PathDoor implements Door, JasManagerListener {
    // XML configuration file
    private static final String JAS_EXT_CONF = "jml";
    // XML configuration processed file
    private static final String JAS_EXT_CONF_OLD = "jmo";
    // JAR for JASon Plugin
    public static final String JAS_EXT_FILE = "jas";

    private String location;
    private File lastWorkingFile;

    /**
     * This is called at the beginning of the Door using by the Domain container. The
     * location is normally a value from the `jason.xml' file. This is a location for
     * service in any format.
     * @param location location for services
     * @exception DoorException exception for bad location value */
    public void init( DoorContext context ) throws DoorException {
	location = context.getLocation();

	if ( "".equals( location ) ) 
	    throw new DoorException( "Empty location" );

	File f = new File( context.getEnvLocation(), location );
	lastWorkingFile = f;
	this.context = context;

	if ( f.list() != null ) {
	    initJas( f );
	    initXml( f );
	}
    }

    // Check for new jar and processed it
    public void update() {
	vJasProcessed = new Vector();
	initJas( lastWorkingFile );
	for ( int i = 0; i < vJasProcessed.size(); i++ ) {
	    String path = "" + vJasProcessed.elementAt( i );
	    try {
		context.getDomain().addService( path, true );
	    } catch( JasonException exc ) {
		exc.printStackTrace();
	    }
	}
    }

    // Manage only '.jas' format
    private void initJas( File f ) {
	String[] files = f.list();
	for ( int i = 0; i < files.length; i++ ) {
	    if ( files[ i ].endsWith( JAS_EXT_FILE ) ) {
		context.sendInformationMessage( "Installing " + files[ i ] );

		File mFile = new File( new File( context.getEnvLocation(), location ),
				       files[ i ] );

		JasManager jm = new JasManager( mFile );
		jm.setJasManagerListener( this );
		jm.init( context.getDomain() );
	    }
	}
    }

    private Vector vJasProcessed = null;

    // From the JasManagerListener interface
    public void notifyFile( File f ) {
	if ( f.toString().endsWith( JAS_EXT_CONF ) )
	    notifyNewJmlFile( f );
    }

    private void notifyNewJmlFile( File f ) {
	if ( vJasProcessed == null )
	    vJasProcessed = new Vector();
	vJasProcessed.addElement( f );
    }

    // Manage service '.jml' file
    private void initXml( File f ) {
	String [] files = f.list();

	// Process all '.xml' file from the current f directory
	for ( int i = 0; i < files.length; i++ ) {
	    if ( files[ i ].endsWith( JAS_EXT_CONF ) ) {
		File mJml = new File( f, files[ i ] );
		try {
		    String path = mJml.toString();
		    context.getDomain().addService( mJml );
		} catch( JasonException exc ) {
		    exc.printStackTrace();
		}
	    }
	}

	// Process sub-directories too
	for ( int i = 0; i < files.length; i++ ) {	
	    File _f = new File( f, files[ i ] );
	    if ( _f.isDirectory() ) {
		initXml( _f );
	    }
	}
    }

    /**
     * @return All available services */
    public Service[] getServices() {
	return null;
    }

    /** This door can be dynamically called for updating or loading services */
    public boolean isUpdatable() {
	return true;
    }

    /** Store the service for the name */
    public boolean writeService( String name, byte[] service ) {
	try {
	    File t = new File( lastWorkingFile, name );
	    System.out.println( "Write service on " + t );
	    FileOutputStream fOut = new FileOutputStream( t );
	    fOut.write( service );
	    fOut.close();
	} catch( Throwable th ) {
	    th.printStackTrace();
	    return false;
	}
	return true;
    }

    private DoorContext context;
}
    
// PathDoor ends here



