// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * <b>Created Sat Jun  1 20:57:22 2002</b>
 * <p>
 * This service is able to work with several processes. This kind
 * of service is only for Socket service type. When is manages a socket,
 * it can send the socket request to another process rather than another
 * thread. By default this service will detect a service conflict from multiple
 * instance and merge this instances into one service. 
 *
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 */
public abstract class ProcessableService extends SocketService {

    public ProcessableService() {
	super();
    }

    int[] loadBalancingServer = null;
    int loadBalancingMax = 50;
    int loadBalancingServerCount = 0;
    int loadBalancingServerCursor = -1;

    /**
     * Reset the max value for loadbalancing support. By default, the service will accept
     * until 50 processes load balancing */
    protected void setLoadBalancingMax( int max ) {
	this.loadBalancingMax = max;
    }

    // Store a new port for loadbalancing
    private void addLoadBalancingServer( int port ) {
	if ( loadBalancingServerCursor >= loadBalancingMax ) {
	    sendErrorMessage( "Can't loadbalancing : max capacity exceeded !" );
	    return;
	}
	if ( loadBalancingServer == null )
	    loadBalancingServer = new int[ loadBalancingMax ];
	loadBalancingServer[ loadBalancingServerCount++ ] = port;
    }

    // Remove a port
    private void removeLoadBalancingServer( int port ) {
	sendInformationMessage( "Remove cooperation with the port " + port );
	boolean mode = false;
	for ( int i = 0; i < loadBalancingServerCount; i++ ) {
	    if ( loadBalancingServer[ i ] == port ) {
		mode = true;
	    }
	    if ( mode )
		loadBalancingServer[ i ] = loadBalancingServer[ i + 1 ];
	}
	loadBalancingServerCount = Math.max( 0, loadBalancingServerCount - 1 );
	loadBalancingServerCursor = -1;
    }

    /**
     * Convert the raw inputStream to an array of byte
     * @param s InputStream to read 
     * @return the InputStream content with a data byte array */
    private byte[] getBytesFromInputStream( InputStream in ) throws IOException {
	ByteArrayOutputStream out = new ByteArrayOutputStream();
	byte[] buffer = new byte[ 1024 ];
	int b = 0;
	while ( true ) {
	    b = in.read( buffer );
	    if ( b <= 0 ) 
		break;
	    out.write( buffer, 0, b );
	}
	return out.toByteArray();
    }


    // Reports the current request to another service that works with the following port
    private void reportRequestToAnotherService( Socket host, int _port ) throws IOException {
	try {

	    // Get all the inputstream request
	    byte[] b = getHostRequestFromInputStream( host.getInputStream() );

	    try {

		// Context the other service
		// We limit the scope for the moment to a local target
		Socket s = new Socket( "127.0.0.1", _port );
		OutputStream o = null;
		InputStream i = null;
		byte[] r = null;

		try {
		    o = new BufferedOutputStream( s.getOutputStream() );
		    // Send the current user request
		    o.write( b );
		    o.flush();
		    // Get the service response
		    r = getBytesFromInputStream( i = s.getInputStream() );
		} finally {
		    try {
			i.close();
			o.close();
			s.close();
		    } catch( Exception exc ) {}
		}

		host.getOutputStream().write( r );

	    } catch( Exception exc ) {
		// We can't contact this host, we remove it
		removeLoadBalancingServer( _port );
		// Try it by the thread balancing
		fireSocketAccept( new SocketProxy( new ByteArrayInputStream( b ), host ) );
	    }

	} finally {
	    try {
		host.getInputStream().close();
		host.getOutputStream().close();
		host.close();
	    } catch( Exception exc ) {}
	}
    }

    /**
     * Manage the load balancing with the PushBacckInputStream content 
     * for the host request */
    private void manageSocketForLoadBalancing( Socket host, 
					       InputStream in ) throws IOException {

	try {
	    // Cooperation mode
	    sendInformationMessage( "Contact for cooperation" );
	    // Get the port
	    BufferedReader br = new BufferedReader( new InputStreamReader( in ) );
	    String _port = br.readLine();
	    if ( _port == null ) {
		sendWarningMessage( "No contact supported, invalid port " + _port );
		br.close();
		host.close();
		return;
	    }
	    _port = _port.trim();

	    if ( !isProcessable() ) {
		sendInformationMessage( "Multiple service runned on port " + _port + " load balancing unsupported" );
	    } else {
		sendInformationMessage( "Cooperation accepted with the port " + _port );
		try {
		    int refPort = Integer.parseInt( _port );
		    addLoadBalancingServer( refPort );
		} catch( NumberFormatException exc ) {
		} 
	    }

	} finally {
	    try {
		in.close();
		host.close();
	    } catch( Exception exc ) {}
	}
    }

    /**
     * This method will dispatch any request between another processes */
    protected void fireSocketAccept( Socket host ) {

	if ( isTest() ) {
	    sendInformationMessage( "Service can work..." );
	    setTest( false );
	}

	if ( traceRequest )
	    sendInformationMessage( ">Connection " + host.toString() );

	try {

	    // Try to test if this a will to go to the LoadBalancing mode
	    InputStream in = host.getInputStream();

	    if ( in == null ) {
		// The connection is out
		sendWarningMessage( "Connection out for host " + host );
		return;
	    }

	    if ( in.markSupported() ) {
		in.mark( 0 );
	    } else {
		in = new PushbackInputStream( in, 1024 );
	    }

	    int _f = in.read();

	    // A service wants to "loadBalancing"
	    if ( _f == 1 ) {
		manageSocketForLoadBalancing( host, in );
		return;
	    }

	    if ( in.markSupported() ) {
		in.reset();
	    } else {
		( (PushbackInputStream)in ).unread( _f );
		host = new SocketProxy( in, host );
	    }

	} catch( Exception exc ) {
	    sendErrorMessage( exc.getMessage() );
	    exc.printStackTrace();
	}

	dispatchTask( host );	

    }

    /**
     * Dispatch the socket for the current process or another one */
    private void dispatchTask( Socket host ) {

	try {

	    // Load balancing in the current process
	    if ( loadBalancingServerCursor == -1 ) {
		notifyTask( getTaskForSocket( host ) );
	    } else {
		try {
		    reportRequestToAnotherService( host, loadBalancingServer[ loadBalancingServerCursor ] );
		} catch( IOException exc ) {
		    exc.printStackTrace();
		    removeLoadBalancingServer( loadBalancingServer[ loadBalancingServerCursor ] );
		    // Try again on a previous server
		    loadBalancingServerCursor = Math.max( 0, loadBalancingServerCursor-- );
		    fireSocketAccept( host );
		}
	    }

	    loadBalancingServerCursor++;

	    if ( loadBalancingServerCursor >= loadBalancingServerCount ) {
		loadBalancingServerCursor = -1;
	    }
	} catch( ServiceException ex ) {
	    ex.printStackTrace();
	}

    }

    /**
     * This method returns an array of bytes with the full user request, this is needed for load
     * balancing between several process. By default this method tries to read until two '\n' or the end of the stream.
     * If your protocol doesn't limit the scope of the request with the same method, please override this method.
     *
     * @param in InputStream from the user request 
     * @return the request if a byte array */
    protected abstract byte[] getHostRequestFromInputStream( InputStream in ) throws IOException;

    private int limitForLoadBalancingTest = 10;
    private boolean loadBalancingMode = false;

    /** We contact another service from the current service port by including a first 1 caracter and we
     * need to find a new port */
    protected void notifyBadPort() {

	if ( !isProcessable() ) {
	    sendWarningMessage( "Can't run the service" );
	    return;
	}

	limitForLoadBalancingTest--;
	if ( limitForLoadBalancingTest <= 0 ) {
	    sendWarningMessage( "Can't access to a valid port for Load Balancing, we give up" );
	    return;
	}

	loadBalancingMode = true;

	// We need to contact same service to propose the load balancing
	int refPort = ( port + 1 );
	// Search for an empty port
	while ( true ) {
	    try {
		ServerSocket ss = new ServerSocket( refPort );
		// Send a message for the original service to notify that this service
		// is a part of the load balacing
		// Communicate with the original service
		Socket s = new Socket( "127.0.0.1", port );
		port = refPort;
		// Notify the new port
		s.getOutputStream().write( 1 );
		BufferedWriter bw = new BufferedWriter( new OutputStreamWriter( s.getOutputStream() ) );
		bw.write( ( "" + refPort ) );
		bw.newLine();
		bw.flush();
		bw.close();
		sendInformationMessage( "Load Balancing process mode for service on port " + port );
		ss.close();
		// Start the service on the new port
		run();
		break;
	    } catch( Exception exc ) {
	    }
	    refPort++;
	}
    }

    /**
     * @return a task for a user socker */
    protected abstract jason.core.pool.Task getTaskForSocket( java.net.Socket host ) throws ServiceException;

}

// ProcessableService ends here
