// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

/**
 * <b>Created Wed May  8 15:50:51 2002</b>
 * <p>
 * This service type has a part of a task scheduler. This kind of 
 * service uses a clock system that awakes the service when needed. 
 * This is very like the crontab unix/linux system.
 * </p>
 * <p>
 * Note that the <code>start</code> and <code>stop<code> methods will be
 * call several time during the life cycle of this objet.
 * When the clock awake this service, the <code>start</code> is
 * called. This is of the service responsability to call the
 * <code>stop</code> method.
 * </p>
 * <p>
 * By default, if you don't reset a starting period, this service
 * is awakened every minute.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0 */
public class SchedulableService extends AbstractService {

    private int minute = -1;
    private int hour = -1;
    private int day = -1;
    private int month = -1;
    private double period = 0;

    /**
     * This service is not by default starting as a thread. 
     * This is not a good idea to change this behavior as this
     * service is managed by the container */
    public SchedulableService() {
	super();
    }

    /**
     * Overrided for avoiding an override of the 
     * <code>notifyInitReady</code> method */
    public void init( ServiceContext sc ) {
	super.init( sc );

	try {
	    manageRequestForType( jason.core.service.Scheduler.TYPE, 
				  new ServiceRequest( this ) );
	    notifyInitSchedulableReady();
	} catch( Exception exc ) {
	    fireCriticalErrorDetected( exc );
	}

	notifyInitSchedulableReady();
    }

    /**
     * This method is called at the end of the notifyInitReady method */
    protected void notifyInitSchedulableReady() {}

    /**
     * This method will awake the service every period in minute. 
     * By default a <code>0</code> value is set, this is 
     * a value for cancelling thie period too. This value can
     * be used with a starting value too 
     * @param period a period value in minutes */
    public void setStartingPeriod( double period ) {
	this.period = period;
    }

    /**
     * Set the starting minute. If you don't specify a hour,
     * it will work every hour
     * @param minute starting minute starting from 0 */
    public void setStartingMinute( int minute ) {
	this.minute = minute;
    }

    /**
     * Set the starting hour. If you don't specify a day, 
     * it will work every day
     * @param hour starting hour starting from 0 */
    public void setStartingHour( int hour ) {
	this.hour = hour;
    }

    /**
     * Set the starting day number. If you don't specify a month
     * it will work every month
     * @param a day number starting from 0 */
    public void setStartingDay( int day ) {
	this.day = day;
    }

    /**
     * Set the starting month number. 
     * @param month a month numerb starting from 0 */
    public void setStartingMonth( int month ) {
	this.month = month;
    }

    /**
     * @return a starting minute time */
    public int getStartingMinute() {
	return minute;
    }

    /**
     * @return a starting hour time */
    public int getStartingHour() {
	return hour;
    }

    /**
     * @return a starting day time */
    public int getStartingDay() {
	return day;
    }

    /**
     * @return a starting month time */
    public int getStartingMonth() {
	return month;
    }

    /**
     * @return a starting period in minute */
    public double getStartingPeriod() {
	return period;
    }

    /**
     * Overrided method for avoiding any thread support or 
     * management */
    public void run() {
	checkForContext();
	getServiceContext().notifyRun();
    }

}

// SchedulableService ends here
