// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.StringWriter;

import java.util.Enumeration;
import java.util.Hashtable;

import javax.xml.parsers.*;

import org.apache.xerces.parsers.*;
import org.apache.xml.serialize.*;

import org.w3c.dom.*;
import org.xml.sax.*;

/**
 * <b>Created Mon Jun 10 10:00:27 2002</b>
 * <p>
 * <code>ScoreManager</code>. This is a toolkit for storing/loading
 * service performance. This is used by the <code>ServicesManager</code>
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 * @see ServicesManager */
public class ScoreManager {

    public static final String FILE = "state/score.xml";

    private Hashtable htScoreInfo = null;

    public ScoreManager() {
	super();   
	htScoreInfo = new Hashtable();
    }

    private void initScore( Element e ) {
	NodeList _l = e.getElementsByTagName( "service" );
	for ( int i = 0; i < _l.getLength(); i++ ) {
	    Element _e = ( Element )_l.item( i );
	    htScoreInfo.put( _e.getAttribute( "id" ), 
			     new ScoreItem( _e.getAttribute( "host" ),
					    _e.getAttribute( "request" ) ) );
	}
    }

    // Return a valid id for service s
    private String getId( Service s ) { 
	return s.getName() + "/" + s.getVersion();
    }

    /**
     * @return the score for host management by this service */
    public long getHostScore( Service s ) {
	ScoreItem _item = ( ScoreItem )htScoreInfo.get( getId( s ) );
	if ( _item == null )
	    return 0;
	return _item.scoreHost;
    }

    /**
     * @return the score for request management by this service */
    public long getRequestScore( Service s ) {
	ScoreItem _item = ( ScoreItem )htScoreInfo.get( getId( s ) );
	if ( _item == null )
	    return 0;
	return _item.scoreRequest;
    }

    /**
     * Parse the services xml file */
    private Element parse( File file ) throws JasonException {
	try {
	    DOMParser parser = new DOMParser();
	    parser.parse( new InputSource( new FileInputStream( file ) ) );
	    Document doc = parser.getDocument();
	    return doc.getDocumentElement();
	} catch( Exception exc ) {
	    exc.printStackTrace();
	    throw new JasonException( exc.getMessage() );
	}
    }

    public void storeScore( Service s, 
			    ServiceContext sc ) {
	htScoreInfo.put( getId( s ),
			 new ScoreItem( sc.getSkillForHost(), 
					sc.getSkillForRequest() ) );
    }

    public void load() throws JasonException { 
	File f = new File( FILE );
	if ( f.exists() ) {
	    Element element = parse( f );
	    initScore( element );
	}
    }

    public void save() {
	DocumentBuilderFactory _fac = DocumentBuilderFactory.newInstance();
	try {
	    DocumentBuilder _builder = _fac.newDocumentBuilder();
	    Document _doc = _builder.newDocument();
	    Element _root = _doc.createElement( "score" );

	    for ( Enumeration enum = htScoreInfo.keys(); enum.hasMoreElements(); ) {
		String _id = ( String )enum.nextElement();
		ScoreItem _item = ( ScoreItem )htScoreInfo.get( _id );

		Element _service = _doc.createElement( "service" );
		_service.setAttribute( "id", _id );
		_service.setAttribute( "host", "" + _item.scoreHost );
		_service.setAttribute( "request", "" + _item.scoreRequest );

		_root.appendChild( _service );
	    }

	    _doc.appendChild( _root );

            OutputFormat format  = new OutputFormat( _doc, "ISO-8859-1", true );
            StringWriter stringOut = new StringWriter();
            XMLSerializer serial = new XMLSerializer( stringOut, format );
            serial.asDOMSerializer();
            serial.serialize( _doc.getDocumentElement() );

	    StringBuffer sb = new StringBuffer( stringOut.toString() );

	    BufferedOutputStream bOut = new BufferedOutputStream( new FileOutputStream( FILE ) );
	    try {
		bOut.write( sb.toString().getBytes() );
	    } finally {
		bOut.close();
	    }

	} catch( Exception exc ) {
	    System.err.println( "Can't store " + FILE );
	}
    }

    ///////////////

    // Wrapper for score
    private class ScoreItem {
	public long scoreHost = 0;
	public long scoreRequest = 0;

	public ScoreItem( long scoreHost, long scoreRequest ) {
	    this.scoreHost = scoreHost;
	    this.scoreRequest = scoreRequest;
	}

	public ScoreItem( String scoreHost, String scoreRequest ) {
	    try {
		this.scoreHost = Long.parseLong( scoreHost );
		this.scoreRequest = Long.parseLong( scoreRequest );
	    } catch( NumberFormatException exc ) {}
	}
    }

}

// ScoreManager ends here
