// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core;

/**
 * Interface that describes a service plugged into the jason server. 
 * <p>1) A Service is initialized (init)
 * <p>2) A Service is started
 * <p>3) A Service is stopped
 * <p>
 * Look at the <code>ServiceLoader</code> and <code>ServicesManager</code> to see how
 * a service is used.
 * <p>For sendind an message please, use the <code>ServiceContect</code>
 * 
 * @author Alexandre Brillant <abrillant@wanadoo.fr>
 * @version 1.3
 * @since 1.0
 * @see AbstractService 
 * @see ServiceContext
 * @see Runlevel
 * @see MessageHandler
 * @see ServiceVersion */
public interface Service {

    // Message type for the sendMessage method
    public static int INFORMATION_MESSAGE = 0;
    public static int WARNING_MESSAGE = 1;
    public static int ERROR_MESSAGE = 2;
    public static int VERBOSE_MESSAGE = 3;

    /**
     * This name should distinguish your service from another one. This is better for
     * multiple service for the same service type to not introduce generic name like 
     * HTTP service or MAIL service here.
     * @return the service name */
    public String getName();

    /**
     * @return the version of the service */
    public ServiceVersion getVersion();

    /**
     * @return information about the service (author,creation date...) */
    public ServiceInfo getInfo();

    /**
     * @return true if the service should write running information */
    public boolean isVerbose();

    /**
     * Note that this value can avoid your service to be taken into account. This is a rescue solution
     * when you want to propose a service type, but you don't want your service to start if this service
     * type is already managed. The strategy for using or not your service is based on the order, 
     * the last one is always the last supported one. It means that as sample you can replace 
     * the log service, or the scheduler service by your own service
     * @return true if the service can be replaced by another service with the same type */
    public boolean isOverride();

    /**
     * initialize the service, reading properties from the ServiceContext */
    public void init( ServiceContext sc );

    /**
     * start the service */
    public void start();

    /**
     * stop the service */
    public void stop();

    /**
     * restart the service */
    public void restart();

    /**
     * @return a Service property */
    public Object getProperty( String name );

    /**
     * @return a Service property as a string */
    public String getStringProperty( String name );

    /**
     * @return a Service property as an integer */
    public int getIntProperty( String name );

    /**
     * @return a Service property as a boolean */
    public boolean getBooleanProperty( String name );

    /**
     * The <code>AbstractService</code> code will set this value to <code>false</code> by default 
     * @return <code>true</code> if you don't want this service cooperate with other service */
    public boolean isPrivate();

    /**
     * This is an important function for cooperation, it says what it the role of a service
     * such as a Web service. Defining a service type will help the JASon container to
     * tied services */
    public ServiceType getType();

    /**
     * This method provides an indication for the JASon container for starting or not this service.
     * Normally this method should return <code>true</code> */
    public boolean isEnabled();

    /**
     * This method is useful for delegating a work to another service
     * @param request A request for the service
     * @return a response about the resut
     * @throws ServiceException when the request is not managed */
    public ServiceResponse manageRequest( ServiceRequest request ) throws ServiceException;


    /** @return the current service state */
    public ServiceState getServiceState();
}
