// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core;

import java.io.*;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;

/**
 * <b>Created Wed Mar 27 11:02:31 2002</b>
 * <p>
 * This is a ServiceContainer. This class prepares
 * a Service for running
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 * @see ServicesManager */
public class ServiceContainer implements ServiceContextListener {

    private String name;
    private String cl;
    private ServicesManager sm;
    private boolean test = true;

    /**
     * Manager a new service
     * @param sm the original ServicesManager
     * @param clLo the ClassLoader for the service
     * @param name a service name
     * @param cl a service class name, this class must implement the <code>jason.core.Service</code> interface
     * @param test Decide to active or not the test state */
    public ServiceContainer(ServicesManager sm, 
			    ClassLoader clLo,
			    String name,
			    String cl,
			    boolean test ) throws JasonException {
	super();
	this.sm = sm;
	this.clLo = clLo;
	this.name = name;
	this.cl = cl;
	this.test = test;
    }

    public ServiceContainer(ServicesManager sm, 
			    Service s,
			    String name, 
			    String cl,
			    boolean test ) throws JasonException {
	super();
	this.sm = sm;
	this.name = name;
	this.cl = cl;
	this.test = test;
	setService( s );
    }

    public static final int NOSTATE = -1;
    public static final int INITIALIZED = 0;
    public static final int STOPPED = 1;
    public static final int STARTED = 2;

    private File cServiceFile = null;
    private long cServiceFileModified = 0;

    private File cWorkingDirectory = null;

    /** Update the working directory */
    public void setWorkingDirectory( File wd ) {
	this.cWorkingDirectory = wd;
    }

    /** @return the working directory */
    public File getWorkingDirectory() {
	return cWorkingDirectory;
    }

    private boolean cTemporary = false;

    /** Make the working directory a temporary directory, so this is clean each time the 
	service is runned */
    public void setWorkingDirectoryTemporary( boolean temporary ) {
	this.cTemporary = temporary;
    }

    /** @return <code>true</code> if the working directory is temporary */
    public boolean isWorkingDirectoryTemporary() {
	return cTemporary;
    }

    /** Set the file path for the service */
    public void setFilePath( File f ) { 
	cServiceFile = f;
	cServiceFileModified = f.lastModified();
    }

    public File getFilePath() { return cServiceFile; }

    private ServiceBuilder cBuilder;

    /** Builder for this service container */
    public void setServiceContainerBuilder( ServiceBuilder builder ) {
	this.cBuilder = builder;
    }

    /** @return the Builder for this service container */
    public ServiceBuilder getServiceContainerBuilder() {
	return cBuilder;
    }

    /** Check if the service must be reloaded */
    public boolean checkForUpdate() {
	if ( cServiceFile == null )
	    return false;
	if ( cServiceFile.lastModified() != cServiceFileModified ) {
	    cServiceFileModified = cServiceFile.lastModified();
	    return true;
	} else return false;
    }

    /** Update from file */
    public void update() { }

    public MessageHandler getMessageHandler() { return sm.getMessageHandler(); }

    public boolean isVerbose() { return sm.isVerbose(); }
    public boolean isTest() { return test; }

    /**
     * @return the name of the service */
    public String getName() { return name; }
    public String toString() { return getName(); }

    // Return current services manager 
    public ServicesManager getServicesManager() { return sm; }

    // Proxy
    ServiceType getServiceType() { return service.getType(); }

    boolean isOverride() { return service.isOverride(); }

    public Object getPrivateObject() { return getServicesManager(); }

    /** Initialize the service container */
    void init() throws JasonException {
	resetService();
    }

    /** Stop the service */
    public void stop() throws JasonException {
	if ( service == null ) 
	    throw new JasonException( "Invalid service container / Can't stop : no service yet " );
	try {
	    service.stop();
	} catch( RuntimeException exc ) {
	    throw new JasonException( exc.getMessage() );
	}
	state = STOPPED;
    }

    private int state;

    /** @return the current service state : NOSTATE, INITIALIZED, STOPPED or STARTED */
    public int getState() { return state; }

    /** Start the tied service */
    void start() throws JasonException {
	if ( service == null )
	    resetService();
	if ( service == null ) 
	    throw new JasonException( "Invalid service container / Can't start : no service yet " );
	try {
	    initService();
	    service.start();
	} catch( RuntimeException exc ) {
	    throw new JasonException( exc.getMessage() );
	}
	state = STARTED;
    }

    private Hashtable htVerbose = null;
    private boolean allVerbose = false;

    /**
     * Set the verbose suport */
    void setVerbose( String type, boolean enabled ) {
	if ( htVerbose == null )
	    htVerbose = new Hashtable();
	htVerbose.put( type, new Boolean( enabled ) );
	if ( "*".equals( type ) && enabled )
	    allVerbose = true;
    }

    boolean isVerbose( String type ) {
	if ( allVerbose )
	    return true;
	Boolean state = (Boolean)htVerbose.get( type );
	if ( state == null )
	    return false;
	return state.booleanValue();
    }

    /**
     * reset the service class */
    private void resetService() throws JasonException {
	if ( service != null )
	    return;
	try {
	    Class _cl = getClassLoader().loadClass( cl );
	    setService( (Service)_cl.newInstance() );
	} catch( Exception exc ) {
	    exc.printStackTrace();
	    throw new JasonException( "Can't load service " + cl );
	}

    }

    /**
     * Field for scheduling service with period,minute,hour,day,month */
    void setSchedulingField( String field, String value  ) throws JasonException {
	if ( "".equals( value ) ) 
	    return;

	if ( !( getService() instanceof SchedulableService ) ) {
	    throw new JasonException( "Invalid service type, can't set the " + field + " value for a non schelulableService" );
	}

	int i = 0;

	try {
	    i = Integer.parseInt( value );
	} catch( NumberFormatException exc ) {
	    throw new JasonException( "Invalid " + field + " value" );
	}

	SchedulableService ss = ( SchedulableService )getService();

	if ( "start-period".equals( field ) ) {
	    ss.setStartingPeriod( i );
	} else
	      if ( "start-minute".equals( field ) ) 
		  ss.setStartingMinute( i );
	      else
		  if ( "start-hour".equals( field ) )
		      ss.setStartingMinute( i );
		  else
		      if ( "start-day".equals( field ) )
			  ss.setStartingDay( i );
		      else
			  if ( "start-month".equals( field ) )
			      ss.setStartingMonth( i );
			  else
			      throw new JasonException( "Invalid field " + field );
    }

    private ClassLoader clLo = null;

    /**
     * @return the classloader for the service */
    public ClassLoader getClassLoader() {
	if ( clLo == null ) 
	    return  ClassLoader.getSystemClassLoader();
	return clLo;
    }

    /**
     * Reset the service context */
    public ServiceContext initService() {
	context = new ServiceContext( this );
	context.addServiceContextListener( this );
	getService().init( context );
	return context;
    }

    public void addServiceContextListener( ServiceContextListener listener ) {
	if ( context == null ) 
	    throw new RuntimeException( "No context" );
	context.addServiceContextListener( listener );
    }

    public void removeServiceContextListener( ServiceContextListener listener ) {
	if ( context == null )
	    throw new RuntimeException( "No context" );
	context.removeServiceContextListener( listener );
    }

    private boolean cStop = true;
    public final static String STATE_FILE_NAME = "_state";

    public void notifyInit() {
	cStop = false;
	state = INITIALIZED;

	if ( context.isPersistent() ) {
	    File f = context.getWorkingDirectoryFile( STATE_FILE_NAME );

	    if ( f.exists() ) {
		try {
		    FileInputStream file = new FileInputStream( f );
		    ObjectInputStream in = new ObjectInputStream( file );
		    ServiceState ss = (ServiceState)in.readObject();
		    in.close();
		    ss.preRead( service );
		} catch( Exception ex ) {
		    context.sendInnerError( ex );
		}
	    }
	}
    }

    public void notifyStart() {
	cStop = false;
	state = STARTED;
    }

    public void notifyStop() {
	cStop = true;
	state = STOPPED;
	if ( context.isPersistent() ) {
	    ServiceState ss = service.getServiceState();
	    ss.postWrite( service );

	    try {
		// Write it
		FileOutputStream file = context.getWorkingDirectoryFileOutputStream( STATE_FILE_NAME );
		ObjectOutputStream out = new ObjectOutputStream( file );
		out.writeObject( ss );
		out.close();
	    } catch( Exception exc ) {
		exc.printStackTrace();
		context.sendInnerError( exc );
	    }
	}
    }

    /** @return is the sub service is stopped */
    public boolean isStopped() { 
	return cStop;
    }

    // Current service
    private Service service;
    // Current context
    private ServiceContext context;

    /**
     * Set the available service */
    void setService( Service service ) {
	this.service = service;
    }

    /**
     * @return the current service */
    public Service getService() {
	return service;
    }

    /**
     * @return the current service context */
    public ServiceContext getServiceContext() {
	return context;
    }

    // Properties for the current service. This 
    // properties are accessed through the ServiceContext
    private Hashtable htProperties = new Hashtable();

    /**
     * Set a property for the service */
    public void setProperty( String name, Object value ) {
	if ( value == null )
	    htProperties.remove( name );
	else
	    htProperties.put( name, value );
    }

    // Particular initializing access
    void setLegalProperty( String name, Object value ) {
	if ( value == null )
	    htProperties.remove( name );
	htProperties.put( name, value );
    }

    /**
     * @return a property value for the service as String */
    public String getStringProperty( String name ) {
	return (String)htProperties.get( name );
    }

    public Object getProperty( String name ) {
	return htProperties.get( name );
    }

    /**
     * @return a enumeration of all properties name */
    public Enumeration getPropertiesName() {
	return htProperties.keys();
    }

    private String author;
    private String version;

    public void setAuthor( String author ) { this.author = author; }
    public String getAuthor() { return author; }

    public void setVersion( String version ) { this.version = version; }
    public String getVersion() { return version; }

    // Is the service can be processed in a new thread
    private boolean threadable = true;

    /**
     * Determine if the current service can be handled in a new thread each time with a thread pooling mecanism */
    public void setThreadable( boolean threadable ) { this.threadable = threadable; }
    public boolean isThreadable() { return threadable; }

    // Is the service can be processed in a another process with a load balancing system
    private boolean processable = false;

    /**
     * Determine if the current process can be handled in a new process with a load balancing mecanism */
    public void setProcessable( boolean processable ) { this.processable = processable; }

    public boolean isProcessable() { return processable; }

}

// ServiceContainer ends here






