// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import java.util.Enumeration;
import java.util.Vector;

import jason.core.pool.Task;
import jason.core.pool.TaskContext;
import jason.core.pool.ThreadCapsule;
import jason.core.cache.Cache;

/**
 * <b>Created Wed Mar 27 11:33:57 2002</b>
 * <p>
 * Service context. This class is used by both
 * the <code>ServiceContainer</code> and the Service classes.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.1
 * @see ServiceContainer
 * @see Service */
public class ServiceContext implements ServiceContextListener {
    private ServiceContainer sc;
    private Vector listeners;

    public ServiceContext( ServiceContainer sc ) {
	super();
	this.sc = sc;
	listeners = new Vector();
    }

    /** Add a listener for the context state */
    public void addServiceContextListener( ServiceContextListener listener ) {
	listeners.add( listener );
    }

    /** Remove a listener for the contexte state */
    public void removeServiceContextListener( ServiceContextListener listener ) {
	listeners.remove( listener );
    }

    /** @return the service working directory */
    public File getWorkingDirectory() {
	return sc.getWorkingDirectory();
    }

    /** @return the working directory for the file f */
    public File getWorkingDirectoryFile( String f ) {
	return new File( getWorkingDirectory(), f );
    }

    /** @return an inputstream for the file f */
    public FileInputStream getWorkingDirectoryFileInputStream( String f ) throws IOException {
	return new FileInputStream( getWorkingDirectoryFile( f ) );
    }

    /** @return an outputstream for the file f */
    public FileOutputStream getWorkingDirectoryFileOutputStream( String f ) throws IOException {
	return new FileOutputStream( getWorkingDirectoryFile( f ) );
    }

    /** @return a reader for the file f */
    public FileReader getWorkingDirectoryFileReader( String f ) throws IOException {
	return new FileReader( getWorkingDirectoryFile( f ) );
    }

    /** @return a writer for the file f */
    public FileWriter getWorkingDirectoryFileWriter( String f ) throws IOException {
	return new FileWriter( getWorkingDirectoryFile( f ) );
    }

    /** Clean the working directory content.
	@param depth <code>true</code> to clean all the directory content with all sub-directory, else
	it cleans only the conctent at the first level */
    public void cleanWorkingDirectory( boolean depth ) {
	File f = getWorkingDirectory();
	cleanWorkingDirectory( f, depth );
    }

    private void cleanWorkingDirectory( File f, boolean depth ) {
	String[] ff = f.list();
	if ( ff == null )
	    return;
	for ( int i = 0; i < ff.length; i++ ) {
	    File subFile = new File( f, ff[ i ] );
	    if ( subFile.isDirectory() && depth ) {
		cleanWorkingDirectory( subFile, depth );
	    }
	    if ( !ff[ i ].startsWith( "_" ) ) {
		subFile.delete();
	    }
	}
    }

    /**
     * Check if verbose is enabled */
    public boolean isVerbose() { return sc.isVerbose(); }

    /**
     * Check if a starting test is enabled */
    public boolean isTest() { return sc.isTest(); }

    /**
     * Check if the service can be handled in a new thread */
    public boolean isThreadable() { return sc.isThreadable(); }

    /**
     * Check if the service can be handled in a new process */
    public boolean isProcessable() { return sc.isProcessable(); }

    public ServiceContainer getServiceContainer() { 
	return sc;
    }

    public MessageHandler getMessageHandler() {
	return sc.getMessageHandler();
    }

    public String getName() { return sc.getName(); }

    /**
     * @return available services */
    public Service[] getServices() { return sc.getServicesManager().getServices(); }

    /**
     * This method will return any service matching the good type even bad service
     * @return available services matching the type */
    public Service[] getServiceForType( ServiceType type ) {
	return sc.getServicesManager().getServiceForType( type );
    }

    /**
     * This method will return an array of service matching the good type
     * sorting it with the criteria.
     * This criteria are 
     * <code>SERVICE_NO_SKILL_CRITERIA</code>, <code>SERVICE_HOST_SKILL_CRITERIA</code>, <code>SERVICE_REQUEST_SKILL_CRITERIA</code> 
     * from the ServicesManager class.
     * @return available services matching the type */
    public Service[] getServiceForType( ServiceType type, int criteria ) {
	return sc.getServicesManager().getServiceForType( type, criteria );
    }

    /**
     * Get all service for the type and return the response of the first one that can answer to the response. Note that
     * this method will try any service for getting the response even the same previous bad service.
     * @param type A Service type
     * @param request a Service request 
     * @exception ServiceException for invalid service request */
    public ServiceResponse fireRequestForType( ServiceType type, ServiceRequest request ) throws ServiceException {
	return fireRequestForType( type, request, ServicesManager.SERVICE_NO_SKILL_CRITERIA );
    }

    // Cache level

    private ServiceType cacheType1, cacheType2;
    private Service cacheService1, cacheService2;

    private boolean cEnableCacheForRequestDispath = true;

    /** Enable/Disable the inner cache for dispatching a request */
    public void setEnabledCacheForRequestDispatch( boolean ok ) {
	this.cEnableCacheForRequestDispath = ok;
    }

    /** @return the state for enabling/disabling the inner cache */
    public boolean isEnabledCacheForRequestDispatch() {
	return cEnableCacheForRequestDispath;
    }

    private static final boolean debugMode = false;

    /**
     * Get all service for the type and return the response of the first one that can answer to the response. Note that
     * this method will try to find the best service depending on your criteria choice. This criteria are 
     * <code>SERVICE_NO_SKILL_CRITERIA</code>, <code>SERVICE_HOST_SKILL_CRITERIA</code>, <code>SERVICE_REQUEST_SKILL_CRITERIA</code> from the 
     * ServicesManager class. 
     * @param type A Service type
     * @param request a Service request 
     * @param criteria SERVICE_NO_SKILL_CRITERIA, SERVICE_HOST_SKILL_CRITERIA, SERVICE_REQUEST_SKILL_CRITERIA
     * @exception ServiceException for invalid service request */
    public ServiceResponse fireRequestForType( ServiceType type, ServiceRequest request, int criteria ) throws ServiceException {

	// First step to avoir to search the lastly found services

	Service cacheService = null;

	if ( cEnableCacheForRequestDispath ) {
	    if ( cacheType1 == type ) {
		cacheService = cacheService1;
	    } else
		if ( cacheType2 == type ) {
		    cacheService = cacheService2;
		}
	}

	if ( debugMode ) {
	    System.out.println( "Request T:" + type + " R:" + request );
	}

	if ( cacheService != null ) {
	    ServiceResponse _r = cacheService.manageRequest( request );
	    if ( _r != null ) {
		ServiceContext _sc = sc.getServicesManager().getServiceContextForService( cacheService );
		if ( _sc != null )
		    _sc.increaseSkillForRequest();		
		return _r;
	    }
	}

	// Try with all available services

	Service[] s = getServiceForType( type, criteria );

	if ( debugMode ) {
	    System.out.println( "Found " + s.length + " valid services" );
	    if ( s.length > 0 )
		System.out.println( "- " + s[ 0 ] );
	}

	for ( int i = 0; i < s.length; i++ ) {
	    try {
		try {

		    // Avoid to use it again if the cache using was bad

		    if ( s[ i ] != cacheService ) {

			ServiceResponse _r = s[ i ].manageRequest( request );

			if ( _r != null ) {

			    // Update the 2 level cache

			    if ( cacheType1 == null ) {
				cacheType1 = type;
				cacheService1 = s[ i ];
			    } else 
				if ( cacheType2 == null ) {
				    cacheType2 = type;
				    cacheService2 = s[ i ];
				}

			    // Increase the skill
			    ServiceContext _sc = sc.getServicesManager().getServiceContextForService( s[ i ] );
			    _sc.increaseSkillForRequest();
			    return _r;
			}
		    }

		} catch( RuntimeException exc ) {
		    // Two penality
		    ServiceContext _sc = sc.getServicesManager().getServiceContextForService( s[ i ] );
		    if ( _sc != null ) {
			_sc.decreaseSkillForRequest();
			_sc.decreaseSkillForRequest();
		    }
		}
	    } catch( ServiceException exc ) {
		// One penality
		ServiceContext _sc = sc.getServicesManager().getServiceContextForService( s[ i ] );
		if ( _sc != null )
		    _sc.decreaseSkillForRequest();
	    }
	}

	throw new ServiceException( "Can't find a service managing this request " + request );
    }

    /**
     * @return an enumeration of all property name */
    public Enumeration getPropertiesName() {
	return sc.getPropertiesName();
    }

    /**
     * @return a service property */
    public Object getProperty( String name ) {
	return sc.getProperty( name );
    }

    /**
     * Update a service property. This change will be illegal for private service */
    public void setProperty( String name, Object value ) throws ServiceException {
	if ( sc.getService().isPrivate() ) {
	    throw new ServiceException( "Can't change a property" );
	}
	sc.setProperty( name, value );
    }

    // Reset from the ServicesManager the tied value
    void setLegalProperty( String name, Object value ) {
	sc.setProperty( name, value );
    }

    /** @return the current service domain */
    public Domain getDomain() {
	return sc.getServicesManager().getDomain();
    }

    /**
     * @return a ThreadCapsule for Thread Pooling */
    public ThreadCapsule getThreadCapsule() {
	// Get it for the current domain
	ThreadCapsule tc = sc.getServicesManager().getDomain().getThreadCapsule();
	if ( tc != null )
	    tc.setTaskContext( new TaskContext( this ) );
	else
	    tc = new DefaultThreadCapsule();
	return tc;
    }

    private Cache cache = null;

    /**
     * @return a cache. By default this is the <code>BasicCacheImpl</code> */
    public Cache getCache() {
	if ( cache == null )
	    cache = new jason.core.cache.BasicCacheImpl();
	return cache;
    }

    // Proxy for the MessageHandler

    public void sendErrorMessage( String service, String content ) {
	getMessageHandler().sendErrorMessage( service, content );
	try {
	    fireRequestForType( jason.core.service.LogService.TYPE, new ServiceRequest( service + "(Err) :" + content ) );
	} catch( ServiceException exc ) {}
    }

    /**
     * Support for inner class exception */
    public void sendInnerError( Exception exc ) {
	try {
	    // One penality
	    decreaseSkillForHost();
	    fireRequestForType( jason.core.service.LogService.TYPE, new ServiceRequest( exc, "exception" ) );
	} catch( ServiceException exc2 ) {}	
    }

    public void sendWarningMessage( String service, String content ) {
	getMessageHandler().sendWarningMessage( service, content );
	try {
	    fireRequestForType( jason.core.service.LogService.TYPE, new ServiceRequest( service + "(War) : " + content ) );
	} catch( ServiceException exc ) {}
    }

    public void sendInformationMessage( String service, String content ) {
	getMessageHandler().sendInformationMessage( service, content );
	try {
	    fireRequestForType( jason.core.service.LogService.TYPE, new ServiceRequest( service + "(Inf) : " + content ) );
	} catch( ServiceException exc ) {}
    }

    public void sendCustomMessage( String service, int type, String content ) {
	if ( getServiceContainer().isVerbose( "" + type ) ) {
	    getMessageHandler().sendCustomMessage( service, type, content );
	    try {
		fireRequestForType( jason.core.service.LogService.TYPE, new ServiceRequest( service + "(" + type + ") : " + content ) );
	    } catch( ServiceException exc ) {}
	}
    }

    private void notifyInitToListeners() {
	for ( int i = 0; i < listeners.size(); i++ ) {
	    ServiceContextListener mListener = ( ServiceContextListener )listeners.elementAt( i );
	    mListener.notifyInit();
	}
    }

    private void notifyStartToListeners() {
	for ( int i = 0; i < listeners.size(); i++ ) {
	    ServiceContextListener mListener = ( ServiceContextListener )listeners.elementAt( i );
	    mListener.notifyStart();
	}
    }

    private void notifyStopToListeners() {
	for ( int i = 0; i < listeners.size(); i++ ) {
	    ServiceContextListener mListener = ( ServiceContextListener )listeners.elementAt( i );
	    mListener.notifyStop();
	}
    }

    /**
     * Notify the service context that the Service is initialized */
    public void notifyInit() {
	notifyInitToListeners();
    }

    /**
     * Notify the service context that the Service is started */
    public void notifyStart() {
	notifyStartToListeners();
    }

    /**
     * Notify the service context that the Service is stopped */
    public void notifyStop() {
	sendInformationMessage( getServiceContainer().getService().getName(), "Stopping " );
	notifyStopToListeners();
    }

    /**
     * Notify the service context that the Service is runned. Please
     * call this method each time the run method is called */
    public void notifyRun() {

    }

    /**
     * Notify the service context that the Service has managed a Task
     * It should be called when a work is managed by the Service, this
     * is used for benchmark */
    public void notifyTask() {

    }

    /**
     * Notify the service context that a Sub task is starting for a service */
    public void notifyStartTask() {

    }

    /**
     * Notify the service context that a Sub task is stopping for a service */
    public void notifyStopTask() {

    }

    /**
     * This method must be called for critical error */
    public void notifyCriticalError( Exception exc ) {
	// One penalities
	decreaseSkillForHost();
	sendInnerError( exc );
    }

    // This field shows the skill for managing request
    private long skillForRequest = 0;
    private long skillForHost = 0;

    void setSkillForRequest( long skillForRequest ) { this.skillForRequest = skillForRequest; }
    void setSkillForHost( long skillForHost ) { this.skillForHost = skillForHost; }

    // Decrease the skill due to a bad request management
    void decreaseSkillForRequest() { skillForRequest--; }
    // Increase the skill due to a good request management
    void increaseSkillForRequest() { skillForRequest++; }
    // Return the current skill for request management
    long getSkillForRequest() { return skillForRequest; }

    // Decrease the skill for the host management 
    void decreaseSkillForHost() { skillForHost--; }
    // Increase the skill for the host management
    void increaseSkillForHost() { skillForHost++; }
    // Return the skill for the host management
    long getSkillForHost() { return skillForHost; }

    private boolean cPersistent = false;

    /** Make the current service persistent */
    void setPersistent( boolean persistent ) {
	this.cPersistent = persistent;
    }

    /** @return <code>true</code> is the current service is able to store its working state */
    boolean isPersistent() {
	return cPersistent;
    }

    //////////////////////

    /**
     * This is a default ThreadCapsule when not Thread are available, it will send automatically the
     * task to a TaskService that will process it as possible */
    class DefaultThreadCapsule implements ThreadCapsule {
	private Task task;

	public void setDefaultPriority( int defaultPriority ) {} 
	public void setKeepAlive(boolean keepAlive) {}
	public boolean isKeepAlive() { return false; }

	public Task getTask() {
	    return task;
	}

	public boolean hasTask() {
	    return task != null;
	}

	public void setTask( Task task ) {
	    if ( task != null ) {
		task.init( new TaskContext( ServiceContext.this ) );
	    }
	    try {
		// We can't manage the task now, we send it to a task manager service that will
		// process it as possible 
		fireRequestForType( jason.core.service.TaskService.TYPE, new ServiceRequest( task ) );
	    } catch( ServiceException exc ) {
		sendInnerError( exc );
	    }
	    this.task = task;
	}

	public void setTask( Task task, int priority ) {
	    this.task = task;
	}

	public void setTaskContext( TaskContext tc ) {}

	public void activateThreadCapsule() {}
    }

}

// ServiceContext ends here
