// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.util.Hashtable;
import java.util.Vector;

import org.apache.xerces.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

/**
 * <b>Created Thu Jun 13 15:50:19 2002</b>
 * <p>
 * This class is a helper for scanning and converting a service XML property
 * into a Java Object Value. This class used a part of the XML-RPC specification for
 * working.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 * @see ServicesManager
 */
public final class ServicePropertyManager {

    /**
     * Convert a property node description into a Java object. 
     *  <code>
     *     <!ELEMENT property ( #PCDATA | value | struct | array )*>
     *     <!ATTLIST property
     *     name CDATA #REQUIRED
     *     value CDATA #IMPLIED>
     *
     *     <!--
     *     Struct definition : 
     *     This tag is for a structure building. This structure is available for service with the Hashtable object  
     *     -->
     *     <!ELEMENT struct ( member+ )>
     *
     *     <!--
     *     Member definition : 
     *     This tag is an element of the structure -->
     *     <!ELEMENT member ( name, value )>
     *
     *     <!--
     *     Value definition :
     *     This tag is an element of the property or of a composite element -->
     *     <!ELEMENT value ( string | boolean | int | i4 | array | struct )>
     *
     *     <!--
     *     Primitive type for the Value. Note that For practival reason
     *     the i4 type is equals to the int type. This is rather for 
     *     compatibilit with XML-RPC -->
     *     <!ELEMENT string ( #PCDATA )>
     *     <!ELEMENT boolean ( #PCDATA )>
     *     <!ELEMENT int ( #PCDATA )>
     *     <!ELEMENT i4 ( #PCDATA )>
     *     <!ELEMENT long ( #PCDATA )>
     *     <!ELEMENT float ( #PCDATA )>
     *     <!ELEMENT double ( #PCDATA )>
     *
     *     <!--
     *     Array definition : 
     *     An array contains one data with inside a sequence of value -->
     *     <!ELEMENT array ( data )>
     *
     *     <!--
     *     Data definition :
     *     This is a needed element for the array structure -->
     *     <!ELEMENT data ( value+ )>
     * </code>
     */
    static Object getPropertyValue( Node node ) {

	// String case
	if ( node.getNodeType() == Node.TEXT_NODE ) {
	    return node.getNodeValue();
	} else {
	    String _tag = node.getNodeName();

	    // Composite item
	    if ( "value".equals( _tag ) ) {
		return getPropertyValue( getFirstChild( node ) );
	    } else
		if ( "array".equals( _tag ) ) {

		    return getPropertyValue( getFirstChild( node ) );
		} else
		    if ( "data".equals( _tag ) ) {
			return getDataValue( node );
		    } else
			if ( "struct".equals( _tag ) ) {
			    return getStructureValue( node );
			} else
			    if ( "string".equals( _tag ) ) {
				return node.getFirstChild().getNodeValue();
			    } else 
				if ( "boolean".equals( _tag ) ) {
				    return new Boolean( "true".equals( node.getFirstChild().getNodeValue() ) );
				} else 
				    if ( "int".equals( _tag ) || "i4".equals( _tag ) ) {
					try {
					    return new Integer( node.getFirstChild().getNodeValue() );
					} catch( NumberFormatException exc ) {
					    return new Integer( 0 );
					}
				    } else
					if ( "long".equals( _tag ) ) {
					    try {
						return new Long( node.getFirstChild().getNodeValue() );
					    } catch( NumberFormatException exc ) {
						return new Long( 0 );
					    }
					} else
					    if ( "float".equals( _tag ) ) {
						try {
						    return new Float( node.getFirstChild().getNodeValue() );
						} catch( NumberFormatException exc ) {
						    return new Float( 0 );
						}
					    } else
						if ( "double".equals( _tag ) ) {
						    try {
							return new Double( node.getFirstChild().getNodeValue() );
						    } catch( NumberFormatException exc ) {
							return new Double( 0 );
						    }
						}
	    // Default
	    return node.getFirstChild().getNodeValue();
	}
    }

    // Needed for the parser DOMParser that includes PCDATA for unknown reason
    private static Node getFirstChild( Node node ) {
	NodeList l = node.getChildNodes();
	for ( int i = 0; i < l.getLength(); i++ ) {
	    Node _node = l.item( i );
	    if ( _node.getNodeType() != Node.TEXT_NODE ) {
		return _node;
	    }
	}
	return node.getFirstChild();
    }

    // Needed for the parser DOMParser that includes PCDATA for unknown reason
    private static Node getLastChild( Node node ) {
	NodeList l = node.getChildNodes();
	Node _last = null;
	for ( int i = 0; i < l.getLength(); i++ ) {
	    Node _node = l.item( i );
	    if ( _node.getNodeType() != Node.TEXT_NODE ) {
		_last = _node;
	    }
	}

	if ( _last == null ) 
	    _last = node.getLastChild();
	return _last;
    }

    // Case for array
    private static Object getDataValue( Node node ) {


	NodeList l = node.getChildNodes();
	Vector v = new Vector();

	// Process all value.
	for ( int i = 0; i < l.getLength(); i++ ) {
	    if ( l.item( i ) instanceof Element ) {
		Element e = (Element)l.item( i );
		if ( e.getNodeType() != Node.TEXT_NODE ) {
		    v.addElement( getPropertyValue( l.item( i ) ) );
		}
	    }
	}

	return v.toArray();
    }

    // Case for struct
    private static Object getStructureValue( Node node ) {
	NodeList l = node.getChildNodes();
	Hashtable hash = new Hashtable();
	// Process all member
	for ( int i = 0; i < l.getLength(); i++ ) {

	    if ( l.item( i ) instanceof Element ) {
		Element _member = ( Element )l.item( i );

		if ( _member.getNodeType() != Node.TEXT_NODE ) {

		    String __name = null;
		    if ( !( "".equals( _member.getAttribute( "name" ) ) ) )
			__name = _member.getAttribute( "name" );
		    else {
			Node _name = getFirstChild( _member );
			__name= _name.getFirstChild().getNodeValue();
		    }

		    Node _value = getLastChild( _member );


		    hash.put( __name.trim(), getPropertyValue( _value ) );

		}
	    }
	}
	return hash;
    }

}

// ServicePropertyManager ends here
