// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.util.Enumeration;
import java.util.Vector;

/**
 * <b>Created Sat Apr 20 15:44:42 2002</b>
 * <p>
 * This class represent the type tied to a service. This is an important role
 * for service cooperation. Look at <code>Service</code> and </code>AbstractService</code>
 * for usage.
 * </p>
 * <p>A ServiceType is a kind of MIME Type for Jason services. As example :
 * You can declare a service Type as Web with a subService HTTP/1.0 for a 
 * HTTP server. For helping JASon to find your service you may choose 
 * a WEB_SERVICE, CUSTOM_SERVICE or JASON_SERVICE constant.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 */
public class ServiceType {
    private String name;
    private ServiceType root;

    public static String WEB_SERVICE = "web";
    public static String CUSTOM_SERVICE = "custom";
    public static String JASON_SERVICE = "jason";

    /**
     * @param name Major service type name, you can use <code>WEB_SERVICE</code>, <code>CUSTOM_SERVICE</code>, or <code>JASON_SERVICE</code>
     * @exception ServiceException if the name is null or empty */
    public ServiceType( String name ) throws ServiceException {
	this.name = name;
	if ( name == null || "".equals( name ) ) 
	    throw new ServiceException( "Invalid ServiceType" );
    }

    /**
     * @param name Major service type name
     * @param name Sub service type, this is more a categoy */
    public ServiceType( String name, String subService ) throws ServiceException {
	this( name, new String[] { subService } );
    }

    /** return <code>true</code> for default service */
    public boolean isDefault() { return CUSTOM_SERVICE.equals( name ); }

    /**
     * @param name Major service type name
     * @param subServices list of synonymous sub service type like http/1.0, ... */
    public ServiceType( String name, String[] subServices ) throws ServiceException {
	this( name );
	if ( subServices == null ) throw new ServiceException( "Invalid subServices" );
	for ( int i = 0; i < subServices.length; i++ ) {
	    addSubServiceType( new ServiceType( subServices[ i ] ) );
	}
    }

    /**
     * @return the major name of the ServiceType */
    public String getName() { return name; }

    /**
     * The parameter object is equals to the current instance of their share the same name and
     * the same children */
    public boolean equals( Object obj ) {
	if ( obj instanceof ServiceType ) {
	    ServiceType _type = (ServiceType)obj;

	    if ( _type.getName().equals( getName() ) ) {

		if ( getSubServiceTypeCount() == _type.getSubServiceTypeCount() ) {
		    boolean ok = false;

		    for ( int i = 0; i < getSubServiceTypeCount(); i++ ) {
			ServiceType __type = getSubServiceType( i );
			// At least sub name one
			ok = ok || _type.checkForServiceTypeName( __type.getName() );
		    }
		    return ok || ( getSubServiceTypeCount() == 0 );
		}
	    }
	    return false;
	} else
	    return super.equals( obj );
    }

    /**
     * Check if the major type name is supported scanning sub service type name 
     * @param name the name to match
     * @return <code>true</code> if the name is found */
    public boolean checkForServiceTypeName( String name ) {
	if ( getName().equals( name ) ) 
	    return true;
	else {
	    if ( subServiceTypes != null ) {
		for ( Enumeration enum = subServiceTypes.elements(); enum.hasMoreElements(); ) {
		    ServiceType _serviceType = (ServiceType)enum.nextElement();
		    if ( _serviceType.checkForServiceTypeName( name ) )
			return true;
		}
	    }
	}
	return false;
    }

    /**
     * Set a service type reference 
     * @param root a service type */
    public void setRootServiceType( ServiceType root ) {
	this.root = root;
    }

    /**
     * @return a ServiceType root node */
    public ServiceType getRootServiceType() {
	if ( root == null ) 
	    return this;
	else
	    return root;
    }

    private Vector subServiceTypes = null;

    /**
     * Add a sub service type
     * @param subServiceType a sub service type */
    public void addSubServiceType( ServiceType subServiceType ) {
	if ( subServiceTypes == null ) 
	    subServiceTypes = new Vector();
	subServiceTypes.addElement( subServiceType );
	subServiceType.setRootServiceType( this );
    }

    /**
     * Remove a sub service type */
    public void removeSubServiceType( ServiceType subServiceType ) {
	if ( subServiceTypes != null )
	    subServiceTypes.removeElement( subServiceType );
    }

    /**
     * @return the sub ServiceType count */
    public int getSubServiceTypeCount() { 
	if ( subServiceTypes != null )
	    return subServiceTypes.size();
	return 0;
    }

    /**
     * @return a sub ServiceType at the index location starting by 0 */
    public ServiceType getSubServiceType( int index ) {
	if ( index < getSubServiceTypeCount() && index >= 0 ) 
	    return (ServiceType)subServiceTypes.elementAt( index );
	return null;
    }

    /**
     * This function will format the ServiceType name and sub services */
    public String toString() {
	String _res = getName();
	if ( getSubServiceTypeCount() > 0 ) {
	    StringBuffer sb = new StringBuffer( _res );
	    sb.append( "{" );
	    for ( int i = 0; i < getSubServiceTypeCount(); i++ ) {
		if ( i > 0 )
		    sb.append( "," );
		sb.append( getSubServiceType( i ) );
	    }
	    sb.append( "}" );
	    _res = sb.toString();
	}
	return _res;
    }

}

// ServiceType ends here
