// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

/**
 * <b>Created Wed May 15 16:34:08 2002</b>
 * <p>
 * This object will maintain information for a service version. 
 * User can set a major or minor version number.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 * @see Service
 * @see AbstractService
 */
public final class ServiceVersion {

    /**
     * Set the major version number like 1 or 2 ...
     * By default the minorVersion will be 0
     * @param majorVersio major version number */
    public ServiceVersion( int majorVersion ) {
	super();   
	this.majorVersion = majorVersion;
    }

    /**
     * Set the major version number and the minor versio number
     * @param majorVersion major version number 
     * @param minorVersion minor version number */
    public ServiceVersion( int majorVersion, int minorVersion ) {
	this( majorVersion );
	this.minorVersion = minorVersion;
    }

    /**
     * Set the major version number with a tagName for a code name for information version
     * @param majorVersion a major version number 
     * @param tagName a name for your version */
    public ServiceVersion( int majorVersion, String tagName ) {
	this( majorVersion );
	this.tagName = tagName;
    }

    /**
     *  Set the major version number, the minor version number and the tagName */
    public ServiceVersion( int majorVersion, int minorVersion, String tagName ) {
	this( majorVersion, tagName );
	this.minorVersion = minorVersion;
    }

    /**
     * Compare two service version number. 
     * @param version version to compare
     * @return <code>true> if the current version is greater that the parameter version */
    public boolean isGreater( ServiceVersion version ) {
	return ( majorVersion > version.getMajorVersion() ) || 
	    ( majorVersion == version.getMajorVersion() && minorVersion > version.getMinorVersion() );
    }

    private int majorVersion = 0;
    private int minorVersion = 0;
    private String tagName = null;
    private String author = null;

    public int getMajorVersion() { return majorVersion; }
    public int getMinorVersion() { return minorVersion; }
    public String getTagName() { return tagName; }

    /**
     * @return a formatted version number */
    public String toString() {
	return "v" + majorVersion + "." + minorVersion + ( tagName != null ? "{" + tagName + "}" : "" );
    }
}

// ServiceVersion ends here
