// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import java.io.File;
import java.io.FileInputStream;

import org.apache.xerces.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * <b>Created Wed Mar 27 10:46:33 2002</b>
 * <p>
 * This is a toolkit for parsing and managing services. This services are added by the domain
 * or by the user.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.1
 * @see Domain
 * @see Service
 * @see ServiceContainer
 * @see ServicePropertyManager */
public class ServicesManager {
    private Domain d;
    private ServiceBuilderFactory sbf;

    ServicesManager( Domain d ) {
	super();   
	this.d = d;
	loadServicesScore();
	sbf = new ServiceBuilderFactory( this, d );
    }

    // Notify from the DomainRunner that the service is initialized
    void notifyInitService( ServiceContext sc, Service s ) {
	htServiceContext.put( s, sc );
    }

    // Return the serviceContext for the service s
    ServiceContext getServiceContextForService( Service s ) {
	return (ServiceContext)htServiceContext.get( s );
    }

    // Notify from the DomainRunner that the service is started
    void notifyStartService( ServiceContext sc, Service s ) {
	// Initialize old score
	long _hostScore = sm.getHostScore( s );
	long _requestScore = sm.getRequestScore( s );
	sc.setSkillForRequest( _requestScore );
	sc.setSkillForHost( _hostScore );
    }

    // Notify from the Domain that all service must be stopped
    void notifyStop() {
	Service[] s = getServiceForType( jason.core.service.LogService.TYPE );
	// Notify the end by a message
	ServiceRequest _req = new ServiceRequest( "Stopping JASon" );
	for ( int i = 0; i < s.length; i++ ) {
	    try {
		s[ i ].manageRequest( _req );
	    } catch( ServiceException exc ) {}
	}

	for ( int i = 0; i < scdm.serviceContainerCount(); i++ ) {
	    ServiceContainer _container = scdm.getServiceContainerAt( i );
    	    try { 
		_container.stop();
	    } catch( JasonException exc ) {
		getMessageHandler().sendErrorMessage( _container.getService().getName(), exc.getMessage() );
	    }
	}
	storeServicesScore();
    }

    private ScoreManager sm = null;

    // Store each service score for the next running
    private void storeServicesScore() { 
	for ( Enumeration enum = htServiceContext.keys(); enum.hasMoreElements(); ) {
	    Service _s = ( Service )enum.nextElement();
	    ServiceContext _sc = ( ServiceContext )htServiceContext.get( _s );
	    sm.storeScore( _s, _sc );
	}
	sm.save();
    }

    // Load each service score for the next running
    private void loadServicesScore() { 
	if ( sm == null ) 
	    sm = new ScoreManager();
	try {
	    sm.load();
	} catch( Exception exc ) {
	}
    }

    public boolean isVerbose() { return d.isVerbose(); }
    public MessageHandler getMessageHandler() { return d.getMessageHandler(); }

    // Return the current domain
    public Domain getDomain() { return d; }
    public Object getPrivateObject() { return getDomain(); }

    /**
     * This is called by the domain for adding one service in an xml document 
     * @param location a service xml file for the service.dtd */
    public ServiceContainer addService( String location ) throws JasonException {
	ServiceContainer mContainer = null;
	scdm.addServiceContainer( mContainer = buildServiceContainerFromFile( location ) );
	htServiceContainer.put( location, mContainer );
	return mContainer;
    }

    /** This is a <code>ServiceContainer</code> builder for a file location */
    public ServiceContainer buildServiceContainerFromFile( String location ) throws JasonException {
	ServiceContainer mContainer = sbf.getFileServiceBuilder().buildService( location );
	return mContainer;
    }

    public void addService( File location ) throws JasonException {
	addService( "" + location );
    }

    /**
     * @deprecated
     * @param location a services xml file for the services.dtd */
    public void addServices( String location ) throws JasonException {
	throw new JasonException( "Method not supported" );
    }

    /** Remove the service container. The sub service is stopped if it is not already */
    public void removeServiceContainer( ServiceContainer sc ) {
	if ( !sc.isStopped() ) {
	    try {
		sc.stop();
	    } catch( JasonException exc ) {
		System.err.println( "Can't stop " + sc );
		exc.printStackTrace();
	    }
	}
	scdm.removeServiceContainer( sc );
    }

    /**
     * Add a service container. Note that this method can remove existing service that can be overrided */
    public void addServiceContainer( ServiceContainer sc ) {
	// Force property valye
	if ( ( sc.getService().getType() != null )
	     && ( sc.getService().getType().equals( jason.core.service.LogService.TYPE ) ) ) {
	    sc.setProperty( "location", d.getLogLocation() );
	    sc.setProperty( "auto-clean", "" + d.isLogAutoClean() );
	}

	// Overrided service ?
	ServiceType _type = sc.getServiceType();

	if ( _type != null )
	    for ( int i = 0; i < scdm.serviceContainerCount(); i++ ) {
		ServiceContainer _sc = scdm.getServiceContainerAt( i );
		ServiceType __type = _sc.getServiceType();
		if ( __type != null )
		    if ( __type.equals( _type ) ) {
			if ( sc.isOverride() ) {
			    d.sendInformationMessage( "Overriding service for type " + __type );
			    // Delete it
			    scdm.removeServiceContainer( _sc );
			    break;
			}
		    }
	    }
	scdm.addServiceContainer( sc );
    }

    /**
     * Overrided a property from an existing service 
     * @param location an xml file service location
     * @param property a property name
     * @param value a property value
     * @exception JasonException for invalid property overriding */
    public void setPropertyForServiceLocation( String location, 
					       String property, 
					       String value ) throws JasonException {
	if ( location == null ) 
	    throw new JasonException( "Invalid service location" );
	ServiceContainer sc = ( ServiceContainer )htServiceContainer.get( location );
	if ( sc == null ) 
	    throw new JasonException( "Invalid service location " + location );
	sc.setProperty( property, value );
    }

    /**
     * @return available service */
    public Service[] getServices() {
	// Create a snapshot for avoiding to return the current 
	// Collection
	ServiceContainer[] sc = scdm.getState();
	Service[] _services = new Service[ sc.length ];
	for ( int i = 0; i < sc.length; i++ ) {
	    _services[ i ] = sc[ i ].getService();
	}
	return _services;
    }

    /** No criteria for a skill evaluation */
    public final static int SERVICE_NO_SKILL_CRITERIA = 0;
    /** Criteria for getting a set of service for host management */
    public final static int SERVICE_HOST_SKILL_CRITERIA = 1;
    /** Criteria for getting a set of service for inner request management */
    public final static int SERVICE_REQUEST_SKILL_CRITERIA = 2;

    /**
     * @return available service matching the type */
    public Service[] getServiceForType( ServiceType type ) {
	Vector v = new Vector();

	for ( int i = 0; i < scdm.serviceContainerCount(); i++ ) {
	    ServiceContainer _container = scdm.getServiceContainerAt( i );
	    Service _service = _container.getService();

	    if ( _service.getType().equals( type ) ) 
		v.addElement( _service );
	}
	Service[] _services = new Service[ v.size() ];
	v.copyInto( _services );
	return _services;
    }

    /**
     * This method will sort the services available for a type by degree of performance from
     * the best performant service to the worst. 
     * @param type ServiceType to found
     * @param criteria use SERVICE_HOST_SKILL_CRITERIA for a service working with clients (like HTTP)
     *  or SERVICE_REQUEST_SKILL_CRITERIA for a service working with another services 
     * @return available service matching the type sorting for the criteria */
    public Service[] getServiceForType( ServiceType type, int criteria ) {
	Vector v = new Vector();

	for ( int i = 0; i < scdm.serviceContainerCount(); i++ ) {
	    ServiceContainer _container = scdm.getServiceContainerAt( i );
	    Service _service = _container.getService();

	    if ( _service.getType().equals( type ) ) { 

		boolean _added = false;
		// Current service score
		ServiceContext _sc = getServiceContextForService( _service );

		if ( criteria != SERVICE_NO_SKILL_CRITERIA ) {

		    long _score = ( criteria == SERVICE_HOST_SKILL_CRITERIA ) ? _sc.getSkillForHost() : _sc.getSkillForRequest();

		    // Insert it at the right place
		    for ( int j = 0; j < v.size(); j++ ) {
			Service __service = (Service)v.elementAt( j );
			ServiceContext __sc = getServiceContextForService( __service );
			long __score = ( criteria == SERVICE_HOST_SKILL_CRITERIA ) ? __sc.getSkillForHost() : __sc.getSkillForRequest();
			if ( __score < _score ) {
			    v.insertElementAt( _service, j );
			    _added = true;
			}
		    }
		}

		if ( !_added ) {
		    v.addElement( _service );
		}
	    }
	}

	Service[] _services = new Service[ v.size() ];
	v.copyInto( _services );
	return _services;
    }

    private ServiceContainerDataModel scdm;

    /** Set a data model for <code>ServiceContainer</code> */
    public void setModel( ServiceContainerDataModel model ) {
	this.scdm = model;
    }

    /** @return a model with all <code>ServiceContainer</code> */
    public ServiceContainerDataModel getModel() { 
	return scdm;
    }

    // Store available services
    private Vector services = new Vector();
    // Store relation between a service location and its container
    private Hashtable htServiceContainer = new Hashtable();
    // Store relation between a service and the ServiceContext
    private Hashtable htServiceContext = new Hashtable();
}

// ServicesManager ends here
