// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core;

import jason.core.pool.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * <b>Created Sat Feb 16 18:50:38 2002</b>
 * <p>
 * This service is for all socket service (like HTTP server). 
 * It simplify the socker server management by a thread pooling 
 * system. 
 * </p>
 * <p>User must implement the <code>getTaskForSocket</code> method </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 * @see AbstractService */
public abstract class SocketService extends TaskableService {
    protected int port = 0;
    protected boolean traceRequest = false;

    /**
     * Run this service in a thread mode */
    public SocketService() {
	super( true );
    }

    public void setTraceRequest( boolean trace ) {
	traceRequest = trace;
    }

    /**
     * Search for the port property */
    protected void notifyInitReady() {
	try {
	    port = Integer.parseInt( "" + getProperty( "port" ) );
	    ServiceDataSpace.setDataList( "port", "" + port );
	    setTraceRequest( "true".equals( getProperty( "trace" ) ) );
	} catch( Exception exc ) {
	    sendErrorMessage( "Can't start this service !" );
	    sendErrorMessage( "Bad port : " + getProperty( "port" ) );
	    isInvalid = true;
	}
    }

    public synchronized void start() {
	if ( !isInvalid ) 
	    super.start();
    }

    private boolean serverSocketWorking = false;
    private int backlog = 100;

    /**
     * Reset the queue size for the server socket object. By default
     * this value is set to 100 */
    public void setBacklog( int backlog ) {
	this.backlog = backlog;
    }

    /**
     * This code is called in the current thread once before the 
     * run loop. This is necessary for code that interrompt the main
     * thread. This method is the "invariant" part of the thread loop */
    protected void beforeRun() {
	sendInformationMessage( "Listening on port :" + port );
	try {
	    serverSocket = new ServerSocket( port, backlog );
	    serverSocketWorking = true;
	    beforeRunState = true;
	} catch( Exception exc ) {
	    sendErrorMessage( "Can't use the port : " + port + " / " + exc.getMessage() );

	    if ( !isProcessable() ) {
		// We stop it
		stop();
		beforeRunState = false;
	    } else {
		if ( !serverSocketWorking )
		    notifyBadPort();
	    }
	}
    }

    private int socketTimeout = 20000;

    /**
     * This method will reset the timeout for socket. By default
     * the value is 20 s
     * @param socketTimeout time out in ms */
    public void setSocketTimeout( int socketTimeout ) {
	this.socketTimeout = socketTimeout;
    }

    public int getSocketTimeout() { return socketTimeout; }

    protected ServerSocket serverSocket = null;
    private boolean beforeRunState = false;

    /**
     * Start the service waiting for a user request or another service request for 
     * loadbalancing request */
    protected void run() {

	if ( !beforeRunState ) {
	    beforeRun();
	    if ( isStopped() )
		return;
	    beforeRunState = true;
	}

	try {
	    Socket _host = serverSocket.accept();
	    _host.setSoTimeout( socketTimeout );

	    // Unknown testing needs, otherwise a socket with an empty InputStream is thrown ??
	    if ( _host.toString().indexOf( "=null" ) > -1 ) {
		try {
		    _host.close();
		} catch( IOException exc ) {}
		return;
	    }

	    // Handle the socket
	    fireSocketAccept( _host );
	    getServiceContext().notifyTask();
	} catch( IOException exc ) {}
    }

    /**
     * We find another service on the port, switch to the load Balancing mode */
    protected void notifyBadPort() {}

    /**
     * Manage a connection on the port */
    protected void fireSocketAccept( Socket host ) {
	try {
	    // Default behaviour
	    notifyTask( getTaskForSocket( host ) );
	} catch( ServiceException exc ) {
	    sendInnerError( exc );
	}
    }

    /**
     * @return a task for a user socker */
    protected abstract Task getTaskForSocket( Socket host ) throws ServiceException;

    /**
     * Test for the port value in the common service data space */
    public void notifyData( String key, Object value ) {
	if ( "port".equals( key ) ) {
	    if ( value instanceof Vector ) {
		Vector _v = (Vector)value;
		if ( _v.contains( value ) ) {
		    fireConflictDetected();
		}
	    }
	}
    }

    /**
     * A port conflict has been detected */
    protected void fireConflictDetected() {
	isInvalid = true;
	sendErrorMessage( "Can't start this service !" );
	sendErrorMessage( "Port already in use by a service : " + port );
    }

    private boolean isInvalid = false;

    /////////////

    // Proxy that override the inputStream
    class SocketProxy extends Socket {
	Socket s;
	InputStream input;

	public SocketProxy( InputStream input, Socket s ) {
	    this.s = s;
	    this.input = input;
	}

	public InetAddress getInetAddress() { return s.getInetAddress(); }
	public InputStream getInputStream() { return input; }
	// For JDK 1.3> public boolean getKeepAlive() throws SocketException { return s.getKeepAlive(); }
	public InetAddress getLocalAddress() { return s.getLocalAddress(); }
	public int getLocalPort() { return s.getLocalPort(); }
	public OutputStream getOutputStream() throws IOException { return s.getOutputStream(); }
	public int getPort() { return s.getPort(); }
	public int getReceiveBufferSize() throws SocketException { return s.getReceiveBufferSize(); }
	public int getSendBufferSize() throws SocketException { return s.getSendBufferSize(); }
	public int getSoLinger() throws SocketException { return s.getSoLinger(); }
	public int getSoTimeout() throws SocketException { return s.getSoTimeout(); }
	public boolean getTcpNoDelay() throws SocketException { return s.getTcpNoDelay(); }
	// For JDK 1.3> public void setKeepAlive(boolean on) throws SocketException { s.setKeepAlive( on ); }
	public void setSendBufferSize(int size) throws SocketException { s.setSendBufferSize( size ); }
	public void setSoLinger(boolean on, int linger) throws SocketException { s.setSoLinger( on, linger ); }
	public void setSoTimeout(int timeout) throws SocketException { s.setSoTimeout( timeout ); }
	public void setTcpNoDelay(boolean on) throws SocketException { s.setTcpNoDelay( on ); }
    }
}

// SocketService ends here
