// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core;

/**
 * <b>Created Wed May 15 16:05:33 2002</b>
 * <p>
 * This is a core class for services that needs to manage a thread.
 * This class will enable a thread to be started for the <code>Start</code>
 * method. User can specify to loop in the <code>run</code> automatically.
 * </p>
 * <p>
 * <b>You shouldn't write a main loop in the <code>run</code> method. Please
 * call <code>setThreadLoopMode</code> with the correct value for that.
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0 */
public class ThreadableService extends AbstractService {

    /**
     * By default the thread will not be started at the <code>start</code> method. 
     * For enabling it, please call the <code>enableServiceForThread</code> at the
     * right time */
    public ThreadableService() {
	super(); 
    }

    /**
     * This constructor will let the service be started as a thread automatically
     * with the <code>start</code> method.
     * @param autoStart <code>true</code> for starting the service as a thread automatically */
    public ThreadableService( boolean autoStart ) {
	this();
	if ( autoStart ) // Prepare the thread
	    try {
		enableServiceForThread();
	    } catch( ServiceException exc ) {}
    }

    /**
     * Be careful with this method, it is normally in the
     * constructor of your service you choose it
     * When starting the thread will call the <code>runThread</code> method
     * @exception if another thread is already running */
    protected void enableServiceForThread() throws ServiceException {
	if ( activeThread != null )
	    throw new ServiceException( "a thread is already running" );

	activeThread = new Thread( new Runnable() {
		public void run() {
		    runThread();
		}
	    } );
	// You may again have to run the start method	
	isThreadStarted = false;
    }

    /**
     * Don't override it, this method is called when starting the service
     * for starting the thread is needed of the service. Note that if
     * you don't specify to run the service as a thread for starting, the
     * <code>run</code> will be called once */
    protected void actionForStart() {
	if ( activeThread != null ) {
	    isThreadStarted = true;
	    if ( newPriority > -1 )
		activeThread.setPriority( newPriority );
	    activeThread.start();
	}
	else
	    super.actionForStart();
    }

    /**
     * This is an important method, because it permits to use
     * the <code>run</code> in a loop. This is useful for the
     * thread state. By default this method is to <code>true</code>.
     * so if you want a thread that is launched one time and expired
     * after the first task use the <code>false<code> value. Note
     * that is you change this value while running, it will stopped
     * the current loop if the initial value was <code>true</code>.*/
    protected void setThreadLoopMode( boolean loopMode ) {
	this.loopMode = loopMode;
    }

    /**
     * @return <code>true</code> if the code is in a loop mode for the threadable state */
    protected boolean isThreadLoopMode() { return loopMode; }

    private long delay = 100;

    /**
     * This value is only for the thread loop mode. It inserts automatically a 
     * <code>sleep</code> method with the delay value. By default this value is
     * with <code>100 ms</code>. If you use a value <= 0 then it will disable the
     * delay */
    protected void setThreadLoopDelay( long delay ) {
	this.delay = delay;
    }

    /** @return the delay in ms for the loop thread period */
    protected long getThreadLoopDelay() {
	return delay;
    }

    /**
     * This method call <code>run</code> method. This is called inside
     * a thread. This method is able to stop 
     * It could be dangerous to override it */
    protected void runThread() {
	try {

	    notifyThreadRunning();

	    if ( !isThreadLoopMode() )
		atomicRun();

	    while ( !isStopped() && isThreadLoopMode() ) {
		atomicRun();
		if ( delay > 0 ) {
		    Thread.sleep( delay );
		}
	    }

	} catch( InterruptedException exc ) {}
    }

    /**
     * This method can be overrided for a task before the
     * run method. This is called when the thread is starting. 
     * If you have specified a thread loop mode , this method
     * will not be called in the loop, it will be called only
     * once.
     */
    protected void notifyThreadRunning() {

    }

    // Method that is called at the end
    // Don't change it !
    private void atomicRun() {
	run();
	checkForContext();
	getServiceContext().notifyRun();
    }

    private int newPriority = -1;

    /**
     * Here the thread priority. You can use Thread.MAX_PRIORITY, Thread.NORM_PRIORITY or
     * Thread.MIN_PRIORITY. This method will be take into account in every case, when the
     * start will be started or when it is running */
    public void setPriority( int newPriority ) {
	this.newPriority = newPriority;
	if ( isThreadStarted ) {
	    activeThread.setPriority( newPriority );
	}
    }

    private boolean isWaiting = false;

    /**
     * This method will be called is the current thread is waiting (user
     * has call the <code>threadWait</code> method. Else by calling the
     * <code>threadNotify</code> method, the response should be <code>false</code> */
    protected boolean isWaiting() {
	return isWaiting;
    }

    public void threadWait() throws InterruptedException {
	if ( activeThread != null ) {
	    isWaiting = true;
	    activeThread.wait();
	}
    }

    public synchronized void threadNotify() {
	if ( activeThread != null ) {
	    isWaiting = false;
	    activeThread.notify();
	}
    }

    protected boolean isThreadStarted() {
	return isThreadStarted;
    }

    private boolean isThreadStarted = false;
    protected Thread activeThread = null;
    private boolean loopMode = true;
}

// ThreadableService ends here
