// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core.pool;

/*
 * This class represents a layer or a capsule over the client processing thread and is responsible for
 * managing the life-cycle of individual threads. Any interaction with the thread shall have to pass
 * through the capsule
 *
 * @author Sanjay"Ironluca"Debnath<Ironluca@yahoo.com> Initial class
 * @author Alexandre Brillant <abrillant@wanadoo.fr> Last changement
 * Last Modified - 9/03/2001 */
public class ThreadCapsuleImpl extends Thread implements ThreadCapsule {

    protected Task coreThread;
    protected boolean keepAlive = true;

    /**
     * This capsule waits until a <code>setTask</code> is called */
    public ThreadCapsuleImpl() {
	this( null );
    }

    /** 
     * Start the ThreadCapsule with a task */
    public ThreadCapsuleImpl( Task task ) {
	coreThread = task;
	setDefaultPriority( 10 );
	this.setDaemon( true );
	this.start();
    }

    private int defaultPriority;

    /**
     * Update the default thread priority. This method will update the
     * current priority unless a task has been set */
    public void setDefaultPriority( int defaultPriority ) {
	this.defaultPriority = defaultPriority;
	if ( !hasTask() )
	    setPriority( defaultPriority );
    }

    /**
     * Restore the priority */
    protected void resetPriority() {
	if ( getPriority() != defaultPriority ) 
	    setPriority( defaultPriority );
    }

    public boolean hasTask() {
	return ( coreThread != null );
    }

    public synchronized void setTask( Task task ) {
	coreThread = task;
	notify();
    }

    public synchronized void setTask( Task task, int priority ) {
	setPriority( priority );
	setTask( task );
    }

    private boolean isRunning = false;

    private TaskContext tc;

    public void setTaskContext( TaskContext tc ) {
	this.tc = tc;
    }

    public synchronized void run(){
	try {
	    isRunning = true;
	    while(keepAlive){
		if ( coreThread == null ) 
		    wait();
		if ( tc != null )
		    coreThread.init( tc );
		tc.notifyStartTask();
		try {
		    coreThread.runTask();
		} catch( Exception exc ) {
		    tc.getServiceContext().sendInnerError( exc );
		}
		tc.notifyStopTask();

		// We save no info on the last task
		if ( !loop )
		    coreThread = null;
		if ( keepAlive );
		    resetPriority();
	    }
	} catch( InterruptedException exc ) {}


	isRunning = false;
    }

    private boolean loop = false;

    /**
     * Decide to maintain the current task (<code>setTask</code>)
     * in the thread. The task will run until a new task takes its place */
    public void setLoopTask( boolean loop ) {
	this.loop = loop;
    }

    /**
     * Decide to maintain the current thread. 
     * by default the value is <code>true</code>. 
     * If you decide to set it to <code>false</code>, you
     * will not have an effective <code>run</code> method meaning
     * a task will not be run */
    public void setKeepAlive( boolean keepAlive ){ 
	this.keepAlive=keepAlive;
    }

    public boolean isKeepAlive(){
	return keepAlive;
    }

    public Task getTask(){
	return coreThread;
    }

    /**
     * This method should be called only if the keep alive state is to <code>true</code> */
    public void activateThreadCapsule(){
	if ( !isRunning )
	    this.run();
    }

}


