// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core.service;

import jason.core.*;

import java.io.*;
import java.util.Date;
import java.util.Vector;

/**
 * <b>Created Thu May  9 18:16:13 2002</b>
 * <p>
 * This is a log service that get all message from
 * other service and store it in a message file
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0 */
public class LogService extends InnerService {

    public static ServiceType TYPE = null;

    static {
	try {
	    TYPE = new ServiceType( ServiceType.JASON_SERVICE, "log" );
	} catch( Exception exc ) {}
    }

    public LogService() {
	super( true );   
	setLog( new File( "log" ) );
	setName( "log" );
	setType( TYPE );
	setThreadLoopDelay( FLUSH_TIME );
	setVersion( new ServiceVersion( 1, 1 ) );
    }

    /**
     * Define the path for storing log message 
     * by default this is the "log" path */
    public void setLog( File logLocation ) {
	this.logLocation = logLocation;
    }

    /**
     * Initialize the log service by using the Log tag attribute from the jason.xml file */
    public void notifyInitReady() {
	String logPath = null;
	if ( getProperty( "location" ) != null )
	    if ( !( "".equals( getProperty( "location" ) ) ) )
		logPath = "" + getProperty( "location" );

	if ( System.getProperty( "jason.log.location" ) != null )
	    logPath = "" + System.getProperty( "jason.log.location" );

	if ( logPath != null )
	    logLocation = new File( logPath );

	if ( getBooleanProperty( "auto-clean" ) ) {
	    cleanMessages();
	}
    }

    public File getLogLocation() { return logLocation; }

    private File logLocation;

    protected void run() {
	super.run();
	writeMessages();
    }

    protected void notifyStopReady() {
	writeMessages();
    }

    public void stop() {
	super.stop();
	// Flush messages
	writeMessages();
    }

    protected void cleanMessages() {
	File f = new File( logLocation, "messages" );
	if ( f.exists() )
	    f.delete();
    }

    /**
     * Store the file message en flush it */
    protected void writeMessages() {
	File f = new File( logLocation, "messages" );
	try {
	    RandomAccessFile raf = new RandomAccessFile( f, "rw" );
	    try {
		raf.seek( raf.length() );
		for ( ;messages.size() > 0; ) {
		    if ( messages.elementAt( 0 ) instanceof Exception ) {
			writeException( ( Exception )messages.elementAt( 0 ), raf );
		    } else {
			writeMessage( "" + messages.elementAt( 0 ), raf );
		    }
		    messages.removeElementAt( 0 );
		}
	    } finally {
		raf.close();
	    }
	} catch( FileNotFoundException exc ) {
	} catch( IOException exc ) {
	    if ( !reportError ) {
		System.err.println( "Error has occured while writing on " + f + " - > " + exc.getMessage() );
	    }
	    reportError = true;
	}
    }

    /**
     * Write an exception to the raf */
    protected void writeException( Exception exc, RandomAccessFile raf ) throws IOException {
	writeMessage( "* Exception " + exc.getMessage(), raf );
	ByteArrayOutputStream _out;
	PrintStream _ps = new PrintStream( _out = new ByteArrayOutputStream() );
	exc.printStackTrace( _ps );
	raf.writeBytes( "\n" );
	raf.writeBytes( _out.toString() );
	raf.writeBytes( "\n" );
	_ps.close();
    }

    /**
     * Write a service message on the raf */
    protected void writeMessage( String message, RandomAccessFile raf ) throws IOException {
	raf.writeBytes( "[" + new Date() + "] " );
	raf.writeBytes( message );
	raf.writeBytes( "\n" );
    }

    /**
     * Manage a message request */
    public ServiceResponse manageRequest( ServiceRequest request ) throws ServiceException {
	if ( request instanceof LogServiceRequest )
	    return manageLogServiceRequest( (LogServiceRequest)request );

	if ( request.getQuery() instanceof String ) {
	    messages.addElement( request.getQuery() );
	} else {
	    if ( request.getQuery() instanceof Exception ) 
		messages.addElement( request.getQuery() );
	    else
		throw new ServiceException( "Invalid request, waiting for a String or Exception object query" );
	}
	return ServiceResponse.EMPTY_RESPONSE;
    }

    /** @return a reponse about the log content. This is useful for sending to another client the log content */
    private ServiceResponse manageLogServiceRequest( LogServiceRequest request ) {
	int mSize = request.getSize();
	Vector mTemp = new Vector();

	File f = new File( logLocation, "messages" );
	try {
	    RandomAccessFile raf = new RandomAccessFile( f, "rw" );
	    try {
		int i = 0;
		String line;
		while ( ( line = raf.readLine() ) != null ) {
		    i++;
		    mTemp.add( line );
		    if ( i > mSize ) 
			mTemp.removeElementAt( 0 );
		}
	    } finally {
		raf.close();
	    }
	} catch( Throwable thw ) {
	    mTemp.add( "Enable to get log content :" + thw.getMessage() );
	}
	String[] mContent = new String[ mTemp.size() ];
	mTemp.copyInto( mContent );
	return new ServiceResponse( mContent );
    }

    private boolean reportError = false;
    private final int FLUSH_TIME = 30000;
    protected Vector messages = new Vector();
}

// LogService ends here




