// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.core.service;

import jason.core.*;
import jason.core.pool.Task;

import java.util.Stack;

/**
 * <b>Created Thu May 16 10:01:57 2002</b>
 * <p>
 * This service has a role for storing a task and
 * process it when possible. This is a kind of queue
 * then get task that service can't handle because
 * they have no ressource for that (like without
 * a thread)
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 */
public class TaskService extends ThreadableService {

    public static ServiceType TYPE = null;

    static {
	try {
	    TYPE = new ServiceType( ServiceType.JASON_SERVICE, "task" );
	} catch( Exception exc ) {}
    }

    public TaskService() {
	super( false );   
	setName( "task" );
	setThreadLoopMode( true );
	setThreadLoopDelay( 10 );

	setType( TYPE );
    }

    /**
     * Manage all <code>Task</code> object. If the request doesn't contain this
     * object type, an exception <code>ServiceException</code> is thrown 
     * @param request user request with a <code>Task<code> type object 
     * @throws ServiceException for bad request type */
    public ServiceResponse manageRequest( ServiceRequest request ) throws ServiceException {
	if ( ( request.getQuery() instanceof Task ) ) {
	    addTaskInQueue( (Task)request.getQuery() );
	} else {
	    throw new ServiceException( "Invalid object " + request.getQuery() );
	}
	return ServiceResponse.EMPTY_RESPONSE;
    }

    /**
     * Store the user task */
    protected void addTaskInQueue( Task task ) {
	if ( stack == null ) 
	    stack = new Stack();

	try {
	    if ( activeThread == null ) 
		enableServiceForThread();
	    if ( !isThreadStarted() ) 
		start();
	    addTaskForThread( task );
	} catch( ServiceException exc ) {
	    sendInnerError( exc );
	}
    }

    /** Store a new task. If the current thread
     * is waiting a notify call is done before */
    private void addTaskForThread( Task _task ) {
	synchronized( stack ) {
	    stack.push( _task );
	    if ( isWaiting() )
		threadNotify();
	}
    }

    /**
     * We override it not to take into account the initial running */
    protected void actionForStart() {
	if ( stack != null )
	    super.actionForStart();
    }

    protected void run() {
	if ( stack.size() == 0 ) {
	    try {
		threadWait();
	    } catch( InterruptedException exc ) {}
	}
	Task t = (Task)stack.pop();
	t.runTask();
    }

    private Stack stack = null;
}

// TaskService ends here
