// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


package jason.core.service.admin;

import jason.core.Domain;
import jason.core.JasonException;
import jason.core.SocketService;
import jason.core.Service;
import jason.core.ServiceResponse;
import jason.core.ServiceContainer;
import jason.core.ServiceContainerDataModel;
import jason.core.ServiceContext;
import jason.core.ServiceException;
import jason.core.ServiceVersion;
import jason.core.ServicesManager;
import jason.core.ServiceType;
import jason.core.pool.SocketTask;
import jason.core.pool.Task;
import jason.core.service.LogServiceRequest;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * <code>Admin</code>
 * Service that offers administration capacity for JASon.
 * It can for known user stops and starts special service.
 *
 * Next release could include information about working.
 *
 * @author Alexandre Brillant <abrillant@wanadoo.fr>
 * @version 1.2 */
public class AdminService extends SocketService {
    public static String TYPE = "admin";

    private int adminPort;
    public static boolean DEBUG_MODE = false;

    public AdminService( int port ) {
	super();
	setName( "admin" );
	setVersion( new ServiceVersion( 1, 1 ) );
	this.adminPort = port;
	try {
	    initActions();
	    setType( new ServiceType( ServiceType.JASON_SERVICE, new String[] { TYPE } ) );
	} catch( ServiceException exc ) {
	    fireCriticalErrorDetected( exc );
	}
    }

    public void notifyInitReady() {
	try {
	    setProperty( "port", "" + adminPort );
	    super.notifyInitReady();
	} catch( ServiceException exc ) {
	    fireCriticalErrorDetected( exc );
	}
    }

    public static String[] AVAILABLE_COMMANDS = { 
	"START", "STOP", "LIST", "LISTE", "PROPERTIES", "PUT", "SET", "LOG"
    };

    public static String[] HELP_COMMANDS = {
	"Start a service", 
	"Stop a service or jason with '*'",
	"List available services",
	"Extended List of available services",
	"List of properties for a service",
	"Store a service",
	"Update a service property",
	"Get last log"
    };

    // Deciding what action to do depending on the command name
    private void initActions() {
	htActions = new Hashtable();
	htActions.put( "START", new StartAction() );
	htActions.put( "STOP", new StopAction() );
	htActions.put( "LIST", new ListAction() );
	htActions.put( "LISTE", new ListDetailAction() );
	htActions.put( "PROPERTIES", new PropertiesAction() );
	htActions.put( "PUT", new PutAction() );
	htActions.put( "SET", new SetAction() );
	htActions.put( "LOG", new LogAction() );
    }

    protected Task getTaskForSocket( Socket host ) throws ServiceException {
	return new AdminProcessing( host );
    }

    private InputStream currentInputStream = null;

    /**
     * Scan the command and send a response to output */
    private void parseCommand( String line, InputStream in, OutputStream out ) throws IOException {
	try {
	    currentInputStream = in;
	    StringTokenizer stz = new StringTokenizer( line );

	    if ( DEBUG_MODE )
		System.out.println( "Receive command : " + line );

	    String name = stz.nextToken();
	    String password = stz.nextToken();

	    String command = stz.nextToken();
	    String param = stz.nextToken();

	    String adminName = getServiceContext().getDomain().getAdminUser();
	    String adminPassword = getServiceContext().getDomain().getAdminPassword();

	    sendInformationMessage( command + " " + param );

	    if ( !adminName.equals( name ) | !adminPassword.equals( password ) )
		sendResponse( out, "Unknown user" );
	    else {
		// Process command
		AdminAction action = (AdminAction)htActions.get( command );
		if ( action == null ) 
		    action = new DefaultAction();

		action.actionForUser( command, param );
	    }
	} catch( Exception exc ) {
	    exc.printStackTrace();
	    sendResponse( out, "Error" );
	}
    }

    // Send a response
    private void sendResponse( OutputStream out, String message ) throws IOException {
	out.write( message.getBytes() );
	out.write( "\n".getBytes() );
	out.flush();
    }

    private void sendResponse( String message ) throws IOException {
	sendResponse( out, message );
    }

    private OutputStream out;
    private Hashtable htActions;

    //////////////////////////////////////////////////////////////////////////////////////

    // Inner interface for admin action
    interface AdminAction {
	public void actionForUser( String command, String param ) throws IOException;
    }

    // Invalid action
    class DefaultAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    sendResponse( "Unknown command" );
	}
    }

    // Set a property
    class SetAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    try {
		StringTokenizer st = new StringTokenizer( param, "=" );
		String service = st.nextToken();
		String property = st.nextToken();
		String value = st.nextToken();

		ServiceContainer mContainer = getServiceContext().getDomain().getModel().getServiceContainerForServiceName( service );
		if ( mContainer != null ) {
		    mContainer.setProperty( property, value );
		}

	    } catch( Throwable thw ) {
		sendResponse( "Error : " + thw.getMessage() );
	    }
	}
    }

    // Put a service
    class PutAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    try {
		StringTokenizer st = new StringTokenizer( param, "=" );
		String jasName = st.nextToken();
		String length = st.nextToken();
		// Full buffer 
		byte[] content = new byte[ Integer.parseInt( length ) ];

		byte[] buffer = new byte[ 512 ];
		int c = 0;
		int i = 0;

		while( ( c = currentInputStream.read( buffer ) ) > 0 ) {
		    System.arraycopy( buffer, 0, content, i, c );
		    i += c;
		    if ( i == content.length )
			break;
		}

		if ( i == content.length ) {
		    try {
			getServiceContext().getDomain().addService( jasName, content );
			sendResponse( "Stored " + jasName );
		    } catch( JasonException exc ) {
			sendResponse( "Can't store the file " + jasName );
		    }
		} else {
		    sendResponse( "Invalid file " + jasName + " length = " + length );
		}
	    } catch( Throwable thw ) {
		sendResponse( "Invalid command " + command + " " + param );
	    }
	}
    }

    // List all property for the service
    class PropertiesAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    String mService = param;
	    if ( ( "".equals( mService ) ) || ( param == null ) ) {
		sendResponse( "Unknown service" );
		return;
	    }
	    Service[] services = getServiceContext().getServices();
	    ServiceContainerDataModel model = getServiceContext().getDomain().getModel();

	    for ( int i = 0; i < services.length; i++ ) {
		if ( services[ i ].getName().equalsIgnoreCase( mService ) ) {
		    ServiceContainer mContainer = model.getServiceContainerForService( services[ i ] );
		    if ( mContainer == null ) // Strange ?
			return;
		    // The good one
		    ServiceContext mContext = mContainer.getServiceContext();
		    for ( Enumeration enum = mContext.getPropertiesName(); enum.hasMoreElements(); ) {
			String _property = (String)enum.nextElement();
			Object _value = mContext.getProperty( _property );
			sendResponse( _property + "=" + _value );
		    }
		    return;
		}
	    }
	    sendResponse( "Unknown service" );
	}
    }

    // List all available service
    class ListAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    Service[] services = getServiceContext().getServices();
	    for ( int i = 0; i < services.length; i++ ) {
		sendResponse( services[ i ].getName() );
	    }
	}
    }

    // List all service with detail */
    class ListDetailAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    Service[] services = getServiceContext().getServices();
	    ServiceContainerDataModel mModel = getServiceContext().getDomain().getModel();

	    for ( int i = 0; i < services.length; i++ ) {
		ServiceContainer mContainer = mModel.getServiceContainerForService( services[ i ] );
		if ( mContainer != null )
		    sendResponse( services[ i ].getName() + "\t" + mContainer.getState() + "\t" + services[ i ].getVersion() + "\t" + services[ i ].getInfo() );
	    }
	}	
    }

    // Stop a service
    class StopAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    if ( "*".equals( param ) ) {
		stopAll();
	    } else {
		Service[] services = getServiceContext().getServices();
		for ( int i = 0; i < services.length; i++ )
		    if ( param.equals( services[ i ].getName() ) ) {
			services[ i ].stop();
			sendResponse( "ok" );
			return;
		    }
		sendResponse( "Don't know service " + param );
	    }
	}

	// Stop the domain
	private void stopAll() {
	    // Get the domain
	    ServiceContainer _sc = ( ServiceContainer )getServiceContext().getServiceContainer();
	    jason.core.ServicesManager _sm = ( ( ServicesManager )_sc.getPrivateObject() );
	    jason.core.Domain _d = ( Domain )_sm.getPrivateObject();
	    _d.stop();
	}
    }

    // Start a service
    class StartAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    Service[] services = getServiceContext().getServices();
	    for ( int i = 0; i < services.length; i++ )
		if ( param.equals( services[ i ].getName() ) ) {
		    services[ i ].start();
		    sendResponse( "ok" );
		    return;
		}
	    sendResponse( "Don't know service " + param );
	}
    }

    class LogAction implements AdminAction {
	public void actionForUser( String command, String param ) throws IOException {
	    try {
		ServiceResponse mResponse = getServiceContext().fireRequestForType( 
										   jason.core.service.LogService.TYPE, 
										   new LogServiceRequest() );
		if ( mResponse.getResponse() instanceof String[] ) {
		    String[] mContent = (String[])mResponse.getResponse();
		    for ( int i = 0; i < mContent.length; i++ ) {
			sendResponse( mContent[ i ] );
		    }
		} else
		      sendResponse( "Unvailable service" );
	    } catch( ServiceException th ) {
		sendResponse( "Unvailable service : " + th.getMessage() );
	    }
	}
    }

    /////////////////////////////

    class AdminProcessing extends SocketTask {

	public AdminProcessing( Socket host ) {
	    super( host );
	}

	public void run() {
	    try {
		InputStream in = socket.getInputStream();
		out = socket.getOutputStream();
		BufferedReader br = new BufferedReader( new InputStreamReader( in ) );
		String command = br.readLine();
		parseCommand( command, in, out );
		// We needn't maintain connection
		socket.close();
	    } catch( IOException ex ) {
		sendErrorMessage( ex.getMessage() );
	    }
	}
    }

}









