// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.framework.protocol.http;

import java.io.*;
import java.util.*;

/**
 * <p>
 * This class is responsible for parsing a HTTP request and obtain the following information
 * </p>
 * 
 * <p>
 * 1) Fully qualified resource name
 * 2) Header names
 * 3) Header values
 * 4) Request message body
 * It takes a InputStream to access the client request.
 * </p>
 * 
 * <p>
 * TO DO : 
 * Support Transfer-Encoding, multipart/byteranges for the Body content
 * Taking into account Keep-Alive for too long process
 * </p>
 * 
 * @author Sanjay"Ironluca"Debnath<Ironluca@yahoo.com> (Original version)
 * @author Alexandre Brillant <abrillant@wanadoo.fr> (Last version)
 * Last Modified - 02/01/2002
 *
 * @see HttpParser
 */
public class HttpRequestParserImpl implements HttpRequestParser {

    private boolean parsed = false;

    // the protocol with version number
    protected String protocol;
    // the Method of the request
    protected String method;
    // the name-value pairs for different http headers sent by the client.
    protected Hashtable headerTable;
    // the name-value pairs for different parameters sent by the client.
    protected Hashtable parameterTable;
    // the request URI from the request Line
    protected String requestURI;

    // The IStream obtained from the client containing the request.
    protected InputStream clientIStream;

    /*
     *This constructor is to be used in any other circumstances - apart from when instances of the
     *class is put in the ThreadPool */
    public HttpRequestParserImpl( InputStream clientIStream ) throws HttpParsingException {
	headerTable = new Hashtable();
	parameterTable = new Hashtable();
	this.clientIStream = clientIStream;
	if ( clientIStream != null )
	    parse( clientIStream );
    }

    /**
     * This constructor is to be used while instances of this class is put into the ThreadPool individually
     * or as a part of a different object */
    public HttpRequestParserImpl() throws HttpParsingException {
	this( null );
    }

    /**
     * Pass an inputStream for parsing the request. User must call the
     * <code>parse</code> method for activation. 
     * @param clientIStream inputStream contains an HTTP request */
    public void setClientIStream(InputStream clientIStream){
	this.clientIStream = clientIStream;
    }

    /**
     * @return the protocol used for this request with version */
    public String getProtocol(){
	return protocol;
    }

    /**
     * @return the Method of the HTTP request */
    public String getMethod(){
	return method;
    }

    /**
     * @return the page request of the HTTP request */
    public String getRequestURI() {
	return requestURI;
    }

    /**
     * @return the query tied with the URI */
    public String getRequestURIQuery() {
	if ( requestURI == null ) return null;
	int i = requestURI.indexOf( "?" );
	if ( i < 1 ) return null;
	return requestURI.substring( i + 1, requestURI.length() );
    }

    /**
     * @return the value of a header if present in the request - otherwise returns "null". */
    public String getHeader( String header ){
	return ( String )headerTable.get( header );
    }

    /**
     * @return the names of the headers of the client request as an Enumeration */
    public Enumeration getHeaderNames(){
	return headerTable.keys();
    }

    /**
     * @param parameter A parameter name
     * @return the value of the specific parameter or null if the parameter is not present in the request */
    public String getParameter( String parameter ){
	Object _obj = parameterTable.get( parameter );
	return (String)_obj;
    }

    /**
     * @return the name of the parameters in form of an Enumeration */
    public Enumeration getParameterNames(){
	return parameterTable.keys();
    }

    /**
     * @return <code>true</code> if the request had been parsed */
    public boolean isParsed() {
	return parsed;
    }

    /**
     * Parses headers following the requestLine. An empty line follow the set of headers, this
     * is the marker for terminating the processing. 
     *
     * From the RFC 2068 :
     *
     *     message-header = field-name ":" [ field-value ] CRLF
     *
     *          field-name     = token
     *          field-value    = *( field-content | LWS )
     *
     *          field-content  = <the OCTETs making up the field-value
     *                           and consisting of either *TEXT or combinations
     *                           of token, tspecials, and quoted-string>
     * @param bufr Buffer with fields
     * @param fields hashtable with the result 
     * @param resolve if <code>true</code>, replace content %AA entity with real caracter value
     */
    protected void parseHeaders( BufferedReader bufr, Hashtable fields, boolean resolve ) throws IOException {
	while ( true ) {
	    String line = bufr.readLine();
	    if ( "".equals( line ) || line == null ) {
		break;
	    } else {
		parseHeaderLine( line, fields, resolve );
	    }
	}
    }

    /**
     * Parses a single field=value line
     * @param line field=value
     * @param fields Stores the key and value from the field=value
     * @param resolve Chooses to resolve Entity %HH value */
    protected void parseHeaderLine( String line, Hashtable fields, boolean resolve ) {
	parseGenericLine( line, ":", fields, resolve, true );
    }

    /**
     * Parse a line based on Property <Separator> Value. 
     * if resolve is <code>true</code>, then all entity %HH are replaced by an ASCII value
     * It stores the result in the <code>fields</code> hashtable
     * @param line Line to parse
     * @param separator Property <Separator> Value, like ':' or '='
     * @param fields Store Property and Value in it
     * @param resolve Resolve entity in the Value */
    protected void parseGenericLine( String line, String separator, Hashtable fields, boolean resolve, boolean lowerCase ) {
	int i = line.indexOf( separator );
	if ( i > 0 ) {
	    String fieldName = line.substring( 0, i );
	    if ( fieldName.charAt( 0 ) == '\n' ) 
		fieldName = fieldName.substring( 1, fieldName.length() );
	    String fieldValue = line.substring( i + 1, line.length() );
	    if ( resolve )
		fieldValue = resolveEntity( fieldValue );
	    fields.put( lowerCase ? fieldName.toLowerCase() : fieldName, fieldValue.trim() );
	}
    }

    // Resolve %XX where XX is an hexadecimal value
    // For instance %20 for a space caracter (ASCII 32)
    private String resolveEntity( String value ) {
	int oldI = 0;
	StringBuffer sbResult = new StringBuffer();
	int length = value.length();
	int i = value.indexOf( "%" );
	value = value.replace( '+', ' ' );

	while ( i > -1 ) {

	    if ( i + 3 <= length ) {
		String _xx = value.substring( i + 1, i + 3 );
		int total = 0;

		// First Digit

		if ( _xx.charAt( 0 ) >= 'a' )
		    total = ( 10 + ( _xx.charAt( 0 ) - 'a' ) ) * 16;
		else
		    if ( _xx.charAt( 0 ) >= 'A' )
			total = ( 10 + ( _xx.charAt( 0 ) - 'A' ) ) * 16;
		    else
			if ( _xx.charAt( 0 ) >= '0' )
			    total = ( _xx.charAt( 0 ) - '0' ) * 16;

		// Second Digit

		if ( _xx.charAt( 1 ) >= 'a' )
		    total += ( 10 + ( _xx.charAt( 1 ) - 'a' ) );
		else
		    if ( _xx.charAt( 1 ) >= 'A' )
			total += ( 10 + ( _xx.charAt( 1 ) - 'A' ) );
		    else
			if ( _xx.charAt( 1 ) >= '0' )
			    total += ( _xx.charAt( 1 ) - '0' );

		char result = (char)total;
		sbResult.append( value.substring( oldI, i ) ).append( result );

		oldI = i + 3;
		i = value.indexOf( "%", oldI );

	    } else break;
	}

	// Last part

	if ( oldI < value.length() ) {
	    sbResult.append( value.substring( oldI, value.length() ) );
	}

	return sbResult.toString();
    }

    /**
     * Parses an HTTP request from the RFC 2068. It's not the role
     * of this method to decide not to support a particular method
     * @throws HttpParsingException for an error while parsing */
    public void parse( InputStream clientIStream ) throws HttpParsingException {

	String requestLine = null;
	Vector headerLineVector;
	BufferedReader bufr;

	if ( clientIStream == null ) throw new HttpParsingException( "Invalid stream reference in HttpParser" );

	try {

	    // Stores the request in a memory buffer for performance stream
	    ByteArrayOutputStream memoryStore = new ByteArrayOutputStream();
	    int c;
	    int lastC = -1;
	    int lastLastC = -1;

	    // Search for double \n or \n\r
	    while ( ( c = clientIStream.read() ) != -1 ) {
		memoryStore.write( c );
		if ( ( lastC == c && c == '\n' ) || 
		     ( lastLastC == c && lastC == 10 ) )
		    break;
		lastLastC = lastC;
		lastC = c;
	    }

	    ByteArrayInputStream memoryInput = new ByteArrayInputStream( memoryStore.toByteArray() );
	    bufr = new BufferedReader( new InputStreamReader( memoryInput ) );

	    // Processing the request line
	    requestLine = bufr.readLine();
	    parseRequest( requestLine );
	    parseHeaders( bufr, headerTable, false );
	    parseParameters( requestURI, clientIStream );
	    parsed = true;

	} catch( IOException exc ) {
	    throw new HttpParsingException( "Can't parse " + requestLine + " ( " + exc.getMessage() + " ) " );
	}
    }

    /**
     * Parse a request line 
     *
     * From the RFC 2068 :
     * request-line = Method SP Request-URI SP HTTP-Version CRLF
     * @param requestLine 
     */
    protected void parseRequest( String requestLine ) {

	char [] tempLineChars;
	int charArrayLength;
	int readIndicatorEnd,readIndicatorStart;

	// We suppose errors are not current for this line
	// We don't care at this step for an invalid protocol

	tempLineChars = requestLine.toCharArray();
	charArrayLength = tempLineChars.length;

	readIndicatorEnd = ( charArrayLength - 1 );
	int step = 0;
	int lastI = 0;

	all:for ( int i = 0; i < readIndicatorEnd; i++ ) {
	    if ( tempLineChars[ i ] == ' ' ) { 
		switch( step ) {
		case 0 : {
		    method = new String( tempLineChars, lastI, i - lastI );
		    break;
		}
		case 1 : {
		    requestURI = new String( tempLineChars, lastI, i - lastI );
		    break;
		}
		}
  	        lastI = i + 1;
		step++;
	    }
	}

	protocol = new String( tempLineChars, lastI, tempLineChars.length - lastI );
    }

    /**
     * Parses parameters from the requestLine and from the buffer content. When it exists
     * an content-length value in the header then the content can be parsed. It is possible to
     * receive parameters from the command line and from the buffer content. 
     * @throw IOException when the content can't be analysed */
    protected void parseParameters( String requestLine, InputStream content ) throws IOException {
	boolean hasContent = headerTable.containsKey( "content-length" );
	// For the moment we ignore the Transfer-Encoding field
	int i = requestLine.indexOf( "?" );

	if ( i > -1 ) {  
	    String parameter = requestLine.substring( i + 1, requestLine.length() );	    
	    parseRequestParameter( parameter );
	}

	if ( hasContent ) {
	    try {
		// Store in a buffer the Post data
  		int contentLength = Integer.parseInt( ( String )headerTable.get( "content-length" ) );
 		byte[] buffer = new byte[ contentLength ];
 		content.read( buffer );
		messageBody = buffer;
 		String requestContent = new String( buffer );
		parseRequestParameter( requestContent );
	    } catch( NumberFormatException exc ) {
	    }
	}
    }

    private byte[] messageBody = null;

    public byte[] getMessageBody() {
	return messageBody;
    }

    /**
     * Parses the request-uri from the request-line 
     * ?name=value&...
     * Stores the result in the <code>parameterTable</code>*/
    protected void parseRequestParameter( String parameter ) {
	int i = parameter.indexOf( "&" );
	int oldI = 0;

	while ( i > -1 ) {
	    String paramElement = parameter.substring( oldI, i );
	    parseRequestParameterItem( paramElement );
	    oldI = i + 1;
	    i = parameter.indexOf( "&", oldI );
	}

	// Process the last element
	String lastElement = parameter.substring( oldI, parameter.length() );
	parseRequestParameterItem( lastElement );
    }

    /**
     * Parses a single field=value from the request-line
     * @param line field=value
     * @param fields Stores the key and value from the field=value
     * @param resolve Chooses to resolve Entity %HH value */
    protected void parseRequestParameterItem( String parameterValue ) {
	parseGenericLine( parameterValue, "=", parameterTable, true, false );
    }

}
