// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.framework.protocol.http;

import java.io.*;
import java.util.*;

/**
 * <b>Created Thu Jan  3 13:11:22 2002</b>
 * <p>
 * <p>
 * This classe is responsable to build a valid HTTP response. User
 * must set parameters and call the <code>build</code> method for
 * a valid response.
 * </p>
 * <p>Here a short desciption of an HTTP response from the RFC 2068 :</p>
 * <pre>
 * generic-message = start-line
 *                   *message-header
 * CRLF
 * [ message-body ]
 *
 * start-line = Request-Line | Status-Line
 * Request-Line = Method SP Request-URI SP HTTP-Version CRLF
 * Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF
 * </pre>
 * </p>
 * @author Alexandre Brillant (abrillant@wanadoo.fr)
 * @version 1.0
 */
public class HttpResponseBuilderImpl implements HttpResponseBuilder {

    private HttpRequestParser parser;

    /**
     * Build an HttpResponseBuilder with the opponent request parser value */
    public HttpResponseBuilderImpl( HttpRequestParser parser ) {
	if ( parser == null ) throw new NullPointerException( "Invalid parser class in the class HttpResponseBuilderImpl" );
	this.parser = parser;
	// Reset a OK code by default
	setStatusCode( StatusCode.STATUS_200 );
    }

    /**
     * Build an HttpResponseBuilderImpl with the opponent request parser value and a status code */
    public HttpResponseBuilderImpl( HttpRequestParser parser, int status ) {
	this( parser );
	setStatusCode( status );
    }

    private int statusCode = -1;

    /**
     * A statusCode should be >= 100 and <= 505 */
    public void setStatusCode( int statusCode ) {
	this.statusCode = statusCode;
    }

    private boolean hasStatusCode() { return statusCode != -1; }

    /**
     * Build a response from the output stream with the browser */
    public void buildResponse( OutputStream output ) throws IOException {
	PrintWriter pw = new PrintWriter( new OutputStreamWriter( output ) );
	if ( hasStatusCode() )
	    buildStatusLineResponse( pw );
	else
	    buildRequestLineResponse( pw );

	pw.println();
	buildMessageHeaderResponse( pw );
	pw.println();
	pw.flush();
	buildMessageBodyResponse( output );
    }

    /**
     * Build a response with the message-header parser (look at the class header) */
    protected void buildMessageHeaderResponse( PrintWriter pw ) throws IOException {
	for ( Enumeration enum = messageHeader.keys(); enum.hasMoreElements(); ) {
	    String field = (String)enum.nextElement();
	    String value = (String)messageHeader.get( field );
	    pw.println( field + ":" + value );
	}
    }

    /**
     * <pre>Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF</pre><br>
     * Build a response for the Status-Line rule (look at the class header)  */
    protected void buildStatusLineResponse( PrintWriter pw ) throws IOException {
	String toSend = parser.getProtocol() + " " + statusCode;
	pw.print( toSend );
    }

    /**
     * <pre>Request-Line = Method SP Request-URI SP HTTP-Version CRLF</pre><br>
     * Build a response for the Request-Line rule (look at the class header) */
    protected void buildRequestLineResponse( PrintWriter pw ) throws IOException {
	String toSend = parser.getMethod() + " " + parser.getRequestURI() + " " + parser.getProtocol();
	pw.print( toSend );
    }

    /**
     * Add the message body part (look at the class header)*/
    protected void buildMessageBodyResponse( OutputStream output ) throws IOException {
	if ( content != null ) {
	    output.write( content );
	}
    }

    /**
     * Stores a message Header that will be sent with the HTTP response (look at the class header for the rule) */
    public void setMessageHeader( String field, String value ) {
	if ( field != null && value != null )
	    messageHeader.put( field, value );
    }

    /**
     * Reset the content-type of the content message-body part 
     * @param type is a MIME type (text/html for instance) */
    public void setContentType( String type ) {
	setMessageHeader( "Content-Type", type );
    }

    /**
     * Reset the content-length, it should be coherent with the <code>
     * setContent</code> call. Note that this last method resets itself
     * this field, you theorically have no reason to reset it */
    protected void setContentLength( int length ) {
	setMessageHeader( "Content-Length", "" + length );
    }

    /**
     * Set the content of the [message-body] part */
    public void setContent( byte[] content ) {
	setContentLength( content.length );
	this.content = content;
    }

    /**
     * Set the server name for the response */
    public void setServer( String server ) {
	setMessageHeader( "Server", server );
    }

    private byte[] content = null;
    private Hashtable messageHeader = new Hashtable();
}

// HttpResponseBuilderImpl ends here
