// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.service.pool;

/**This class represents a layer or a capsule over the client processing thread and is responsible for
*managing the life-cycle of individual threads. Any interaction with the thread shall have to pass
*through the capsule.<br/>
*
*Author - Sanjay"Ironluca"Debnath <a href="mailto:Ironluca@yahoo.com">Ironluca@yahoo.com</a><br/>
*Last Modified - 9/03/2001
*/

public class ThreadCapsule extends Thread implements ThreadCapsuleInt{

/**The class of the {@link #coreThread coreThread} that is embeded in this threadCapsule.<br/>
This class must implement java.lang.Runnable either explicitely or implicitly by extending java.lang.Thread.<br/>
This class must also have a <b>public visible default constructor</b>*/
        protected Class threadClass;

/**The thread which is embedded in this capsule and is responsible to process a client.*/
        protected Runnable coreThread;

/**The state of the capsule, active or passive - <b>false</b> indicates the thread is not servicing any client right now.
<b>true</b> indicates that it is servicing some client.*/
        protected boolean state;

/**This variable is used to control the thread embedded in the threadCapsule & to keep it alive or terminate it.
<b>true</b> - keep alive, <b>false</b> - marked for termination.*/
        protected boolean keepAlive;

/**Reference of the {@link jason.service.pool.ActivePool ActivePool} for the capsule.*/
        protected ActivePoolInt activePool;

/*Reference of the {@link jason.service.pool.PassivePool PassivePool} for the capsule.*/
        protected PassivePoolInt passivePool;


/** The only visible constructor for the class. It is a critical part of the class and does some important
initializations to the threadCapsule. This constructor is used by {@link jason.service.pool.ThreadPool ThreadPool}
to create individual ThreadCapsules in the pool. 
@param <b>threadClass</b> The Class with which the ThreadCapsule shall initialize it's {@link #coreThread coreThread}.
@throws java.lang.InstantiationException If there is no <b>visible default constructor</b> for the {@link #threadClass threadClass}.
@throws java.lang.IllegalAccessException If there is no <b>public constructor</b> for the {@link #threadClass threadClass}.*/
        public ThreadCapsule(Class threadClass) throws InstantiationException, IllegalAccessException{

                this.threadClass=threadClass;
                coreThread=(Runnable)threadClass.newInstance();
                state=false;
                keepAlive=true;

                this.setPriority(10);//The priority of ThreadCapsule thread must be highest in this section, otherwise
                                                //there is a chance that the thread which is activating the ThreadCapsule thread
                                                //shall run and execute wait() on this object - DO NOT CHANGE THIS PRIORITY.
                                                //THIS CODE WAS ADDED AFTER TESTING.

                this.setDaemon(true);//This code was added after testing - at any point if only ThreadCapsule objects
                                                   //are running in the JVM then the JVM shall exit, as there is no server thread running
                                                 //there is little point in keeping the ThreadCapsule objects alive.
                                                //DO NOT CHANGE THIS CODE - OTHERWISE IN CASE OF MAIN SERVER
                                               //CRASHING THE JVM WILL NEVER EXIT NORMALLY
                this.start();

        }//constructor closing


/**This overridden method is responsible to run the algorithm required to service the client.<br/>
In effect what it does is call the <b>run()</b> method on the {@link #coreThread coreThread}.<br/>
The <b>run()</b> method of the coreThread is then responsible to do whatever is required to service the
client.<br/>
Apart from the above - it also calls <b>wait()</b> and <b>notify()</b> method on itself to suspend and resume
execution.<br/>
When a cycle of execution is over (the coreThread run() method returns), it takes itself off the {@link #activePool activePool}
and pushes back to the {@link #passivePool passivePool}.*/
        public synchronized void run(){

                while(keepAlive){

                        if(!state){
                        //System.out.println("Inside if of ThreadCapsule");
                                try{
                                        //System.out.println(Thread.currentThread().toString());
                                        state=true;
                                        wait();
                                }catch(InterruptedException ie){continue;}

                        }//if closing
                        else {
                                //System.out.println("Inside else of ThreadCapsule");
                                //System.out.println(Thread.currentThread().toString());
                                state=false;
                                notify();
                                return;

                        }//else closing

                        //System.out.println(Thread.currentThread().toString());
                        coreThread.run();


                        activePool.remove((Object)this);
                        passivePool.push((Object)this);
                        //System.out.println(Thread.currentThread().toString());
                        //System.out.println("While looping in ThreadCapsule");

                }//while closing

        }//run closing


/**This method sets the {@link #activePool activePool} reference of the ThreadCapsule.
@param <b>activePool</b> - An instance of a class implementing the {@link jason.service.pool.ActivePoolInt ActivePoolInt} interface.
@see jason.service.pool.ActivePool
@see jason.service.pool.ActivePoolInt*/
        public void setActivePool(ActivePoolInt activePool){this.activePool=activePool;}

/**This method returns the reference to the {@link #activePool activePool} of which th ThreadCapsule is a member.
@return jason.service.pool.ActivePoolInt - The reference to the {@link aws.pool.ActivePool ActivePool}.
@see jason.service.pool.ActivePool
@see jason.service.pool.ActivePoolInt*/
        public ActivePoolInt getActivePool(){return activePool;}

/**This method sets the {@link #passivePool passivePool} reference of the ThreadCapsule.
@param <b>passivePool</b> - An instance of a class implementing the {@link jason.service.pool.PassivePoolInt PassivePoolInt} interface.
@see jason.service.pool.PassivePool
@see jason.service.pool.PassivePoolInt*/
        public void setPassivePool(PassivePoolInt passivePool){this.passivePool=passivePool;}

/**This method returns the reference to the {@link #passivePool passivePool} of which th ThreadCapsule is a member.
@return jason.service.pool.PassivePoolInt - The reference to the {@link aws.pool.PassivePool PassivePool}.
@see jason.service.pool.PassivePool
@see jason.service.pool.PassivePoolInt*/
        public PassivePoolInt getPassivePool(){return passivePool;}

/**This method sets the {@link #state state} variable of the ThreadCapsule.<br/>
This method should only be called by the {@link jason.service.pool.ThreadPool ThreadPool} or the controller of the
thread pool.<br/>
<b>This method should never be called by any application apart from the ones mentioned above.</b>
@param <b>state</b> - The {@link #state state} of the ThreadCapsule.*/
        public void setState(boolean state){this.state=state;}

/**This method merely returns the {@link #state state} of the ThreadCapsule.
@return boolean - the {@link #state state} of the ThreadCapsule.*/
        public boolean getState(){return state;}

/**This method sets the {@link #keepAlive keepAlive} variable of the ThreadCapsule, which is responsible
to control the life-cycle of the ThreadCapsule.<br/>
<b>This method should only be called by the {@link jason.service.pool.ThreadPool ThreadPool} or the controller of the
thread pool.</b>
@param <b>boolean</b> - <b>true</b> to keep the ThreadCapsule alive, <b>false</b> to terminate.*/
        public void setKeepAlive(boolean keepAlive){this.keepAlive=keepAlive;}

/**Returns the value of the {@link #keepAlive keepAlive} variable of the ThreadCapsule.
@return boolean - Indicating the keepAlive status of the ThreadCapsule. It is useful to check for dead ThreadCapsules.*/
        public boolean getKeepAlive(){return keepAlive;}

/**Returns the {@link #threadClass threadClass} reference of the ThreadCapsule.
@return java.lang.Class - The Class for which the pool is being maintained.*/
        public Class getThreadClass(){return threadClass;}

/** This method returns the {@link #coreThread coreThread} stored in this ThreadCapsule. The coreThread
is responsible to service the client.
@return java.lang.Runnable - The thread embedded in the ThreadCapsule.*/
        public Runnable getCoreThread(){return coreThread;}

/**This method is provided for optimization purpose only and should only be called by the {@link jason.service.pool.ThreadPool ThreadPool}
<b>only</b>.
@param coreThread - The thread to be embedded in the ThreadCapsule.*/
public void setCoreThread(Runnable coreThread){this.coreThread=coreThread;}

/**This method must be called by the client to activate the thread capsule as well as the embedded thread.
The generic run() method of the embedded thread should not be called as this may result in destabilization of the entitre pool.*/
        public void activateThreadCapsule(){

                this.run();

        }//activateThreadCapsule closing


}//class closing


