// Copyright (c) 2002 Alexandre Brillant
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



package jason.service.pool;

import java.util.*;

/**
* This class is an generic implementation of a thread pool, which takes the following input<br/>
*a) Size of the pool to be constructed<br/>
*b) Name of the class which implements Runnable (which has a visible default constructor)<br/>
*and constructs a thread pool with active threads that are waiting for activation.
*once the threads have finished processing they come back and wait once again in the pool.<br/><br/>
*
*This thread pool engine can be locked i.e. if some internal operation is performed on the pool
*then it is preferable that the thread engine be locked. Locking ensures that no new threads are issued
*by the engine. However, the currently executing threads are allowed to continue till they come back
*to the passivePool.<br/><br/>
*
*Author - Sanjay "Ironluca" Debnath <a href="mailto:Ironluca@yahoo.com">Ironluca@yahoo.com</a><br/>
*Last Modified - 9/03/2001
*/


public class ThreadPool implements ThreadPoolInt{

/**Initial size of the pool as indicated by the user.*/
        protected int totalPoolSize;

/**Current no of threads in the pool which are not available for servicing.*/
        protected int activePoolSize;

/**Current no of threads in the pool which are available for servicing.*/
        protected int passivePoolSize;

/**The class for which the pool is constructed, it must implement runnable and have a visible default constructor.*/
        protected Class threadClass;

/**The pool where the currently active threads which are servicing the clients are kept.*/
        protected ActivePoolInt activePool;

/**The pool where the currently passive threads which are not servicing the clients are kept.*/
        protected PassivePoolInt passivePool;

/**This represents the status of the pool - unlocked or locked. If pool is locked then no more threads 
are issued by the thread pool, this feature shall become useful when some internal change to the pool 
is being performed - like changing the size of the pool etc.*/
        protected boolean poolLocked;

/**Temporary storage for ThreadCapsule objects.*/
        protected ThreadCapsuleInt tempThreadCapsule;


/**The only visible constructor for the class. It takes as it's argument the java.lang.Class of the java.lang.Thread
implementation and the size of the pool to create. The Class passed as the argument should implement the
java.lang.Runnable interface.

@param <b>threadClass</b> The class for which the pool will be maintained. This class should implement java.lang.Runnable.
Since java.lang.Thread already implements Runnable - this class could extend java.lang.Thread.

@param <b>totalPoolSize</b> This is an integer value which indicates the initial size of the thread pool. This value can
be changed later using provided methods. However, changing pool size frequently is not recommended.

@throws jason.service.pool.IllegalClassException This exception is generated if the class supplied for pooling does not
implement java.lang.Runnable in any form (either explicitely or by extending Thread).
*/
        public ThreadPool(Class threadClass, int totalPoolSize) throws IllegalClassException, IllegalAccessException, InstantiationException{

                poolLocked=false;
                this.initializePools(threadClass, totalPoolSize);

        }//constructor closing


/**This method is called by the {@link #ThreadPool(Class,int) constructor} as well as {@link #setThreadClass(Class,int) setThreadClass}
method to initialize or reinitialize all the pools ({@link jason.service.pool.ActivePool ActivePool} as well as 
{@link jason.service.pool.PassivePool PassivePool}).*/
        protected void initializePools(Class threadClass, int totalPoolSize) throws IllegalClassException, IllegalAccessException, InstantiationException{

                if(!(Runnable.class.isAssignableFrom(threadClass)))throw new IllegalClassException();
                else{

                        this.threadClass=threadClass;
                        this.totalPoolSize=totalPoolSize;

                        activePool=new ActivePool(threadClass);
                        passivePool=new PassivePool(threadClass);

                        for(int i=0;i<totalPoolSize;i++){

                                tempThreadCapsule=new ThreadCapsule(threadClass);
                                tempThreadCapsule.setActivePool(activePool);
                                tempThreadCapsule.setPassivePool(passivePool);
                                passivePool.push(tempThreadCapsule);

                                activePoolSize=activePool.getPresentSize();
                                passivePoolSize=passivePool.getPresentSize();

                        }//for closing

                }//else closing

        }//initializePools closing

/**Returns the class for which the pool is being maintained. This is a convenience method.
@return java.lang.Class - The Class for which the thread pool is being maintained.*/
        public Class getThreadClass(){
                return threadClass;
        }//getThreadClass closing

/**Destroyes the current pools ({@link jason.service.pool.ActivePool ActivePool} as well as {@link aws.pool.PassivePool PassivePool})
and reinitializes the pool. This method should not be called if there are current clients being serviced.<br/>
The working is as follows:<br/>
a)The thread pool is first locked. In the locked state no threads are issued for any request.<br/>
b)The {@link #initializePools(Class,int) initializePools} method is called with the requisit parameters
which reinitializes the ActivePool as well as the PassivePool. Thus effectively destroying the earlier pools.<br/>
c)The thread pool is unlocked to receive request and issuing threads.<br/>
It might be posible to save the earlier objects to introduce optimization.
@param <b>threadClass</b> The java.lang.Class object representing the thread class to be pooled.
@param <b>totalPoolSize</b> The target size of the pool.
@throws jason.service.pool.IllegalClassException This exception is thrown if the Class supplied does not implement
java.lang.Runnable either directly or indirectly by extending java.lang.Thread.
*/
        public void setThreadClass(Class threadClass, int totalPoolSize) throws IllegalClassException, IllegalAccessException, InstantiationException{

                this.setPoolLocked(true);
                this.initializePools(threadClass, totalPoolSize);
                this.setPoolLocked(false);

        }//setClass closing


/**Returns the value of totalPoolSize variable. Which in effect is the total size of the pool being maintained 
<b>(active pool + passive pool)</b>.*/
        public int getTotalPoolSize(){

                return totalPoolSize;

        }//getTotalPoolSize closing


/**This method sets the size of the total pool <b>(active + passive)</b> pool - if the size of the active pool is more than
the size of the passive pool then {@link jason.service.pool.ClientsUnderServiceException ClientUnderServiceException}
is thrown otherwise the size of the passive pool is reduced to <b> (totalPoolSize-newPoolSize)</b>.<br/><br/>
<b>NOTE:</b> In effect, the pools are not destroyed, but the required number of objects are either added to the
pool or removed from the pool ({@link jason.service.pool.PassivePool PassivePool}). In case, the target size of the pool is less than the current
number of thread objects in the {@link jason.service.pool.PassivePool PassivePool}, some thread objects are taken out of the ActivePool. However,
it is not possible to determine which thread objects have been taken out. This method may have detremental
effect on threads that are right now running. However, since a thread is not garbage collected while it is
still active, the servicing of the client/request should proceed without interruption. After the thread finishes
execution, it shall be garbage collected.
@param <b>int</b> The target size of the pool <b>(active pool + passive pool)</b>.*/
        public void setTotalPoolSize(int totalPoolSize) throws ClientsUnderServiceException, IllegalClassException, IllegalAccessException, InstantiationException{

                if(totalPoolSize == this.totalPoolSize)return;

                if(totalPoolSize > this.totalPoolSize){

                        for(int i=0;i<(totalPoolSize-this.totalPoolSize);i++){

                                tempThreadCapsule=new ThreadCapsule(threadClass);
                                tempThreadCapsule.setActivePool(activePool);
                                tempThreadCapsule.setPassivePool(passivePool);
                                passivePool.push(tempThreadCapsule);

                                activePoolSize=activePool.getPresentSize();
                                passivePoolSize=passivePool.getPresentSize();

                        }//for closing

                }//if closing

                else{

                        poolLocked=true;

                        int reduce=this.totalPoolSize-totalPoolSize;

                        if(reduce > passivePoolSize){

                                poolLocked=false; 
                                throw new ClientsUnderServiceException("Clients under service, cannot reduce pool size");

                        }//if closing

                        else if(reduce == passivePoolSize){poolLocked=false; return;}
                                else {passivePool.setSize(this.totalPoolSize-reduce); poolLocked=false;}

                }//else closing

        }//setTotalPoolSize closing



//
/**Returns the value of the activePoolSize variable. In effect the total number of active threads at any instance.<br/>
<b>NOTE: </b> the size of this pool cannot be changed individually.
@return int - The current size of the {@link #activePool activePool}.*/
        public int getActivePoolSize(){

                return activePoolSize;

        }//getActivePoolSize closing


/**Returns the value of the passivePoolSize variable. In effect the total number of waiting threads at any instance.<br/>
<b>NOTE: </b> the size of this pool cannot be changed individually.
@return int - The current size of the {@link #passivePool passivePool}.*/
        public int getPassivePoolSize(){

                return passivePoolSize;

        }//getPassivePoolSize closing


/**Returns the current value of the {@link #poolLocked poolLocked} variable
@return boolean - "true" if the pool is currently locked. "false" if not.*/
public boolean isPoolLocked(){return poolLocked;}


/**Toggles the locked or unlocked state of the pool
@param <b>boolean</b> The desired state of the pool.*/
public void setPoolLocked(boolean lockingState){poolLocked=lockingState;}



/**This method returns a {@link jason.service.pool.ThreadCapsule ThreadCapsule} object with the client processing thread embedded in it.
This object is responsible for controlling the core thread, if there are no passive threads in the pool
this method returns null.
@return jason.service.pool.ThreadCapsule with the client servicing thread embedded in it.*/
        public ThreadCapsuleInt getThreadCapsule(){

                if(poolLocked)return null;
                else{

                        tempThreadCapsule=(ThreadCapsuleInt)passivePool.pop();

                        if(tempThreadCapsule == null) return tempThreadCapsule;
                        else{

                                activePool.add(tempThreadCapsule);
                                activePoolSize=activePool.getPresentSize();
                                passivePoolSize=passivePool.getPresentSize();

                                return tempThreadCapsule;

                        }//else closing

                }//else closing

        }//getThreadCapsule closing


}
